#  tests for vispy-0.7.3-py310h50a3682_0 (this is a generated file);
print('===== testing package: vispy-0.7.3-py310h50a3682_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import vispy.io
import vispy.app
import vispy.gloo

import vispy.glsl
import vispy.glsl.antialias
import vispy.glsl.arrows
import vispy.glsl.collections
import vispy.glsl.colormaps
import vispy.glsl.markers
import vispy.glsl.math
import vispy.glsl.misc
import vispy.glsl.transforms

#  --- run_test.py (end) ---

print('===== vispy-0.7.3-py310h50a3682_0 OK =====');
print("import: 'vispy'")
import vispy

print("import: 'vispy.util.dpi'")
import vispy.util.dpi

print("import: 'vispy.util.fonts'")
import vispy.util.fonts

