define(["@jupyter-widgets/base"],(function(e){return function(e){var t={};function n(r){if(t[r])return t[r].exports;var i=t[r]={i:r,l:!1,exports:{}};return e[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=e,n.c=t,n.d=function(e,t,r){n.o(e,t)||Object.defineProperty(e,t,{enumerable:!0,get:r})},n.r=function(e){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})},n.t=function(e,t){if(1&t&&(e=n(e)),8&t)return e;if(4&t&&"object"==typeof e&&e&&e.__esModule)return e;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:e}),2&t&&"string"!=typeof e)for(var i in e)n.d(r,i,function(t){return e[t]}.bind(null,i));return r},n.n=function(e){var t=e&&e.__esModule?function(){return e.default}:function(){return e};return n.d(t,"a",t),t},n.o=function(e,t){return Object.prototype.hasOwnProperty.call(e,t)},n.p="",n(n.s=4)}([function(e,t,n){var r;
/*!
 * jQuery JavaScript Library v3.6.0
 * https://jquery.com/
 *
 * Includes Sizzle.js
 * https://sizzlejs.com/
 *
 * Copyright OpenJS Foundation and other contributors
 * Released under the MIT license
 * https://jquery.org/license
 *
 * Date: 2021-03-02T17:08Z
 */!function(t,n){"use strict";"object"==typeof e.exports?e.exports=t.document?n(t,!0):function(e){if(!e.document)throw new Error("jQuery requires a window with a document");return n(e)}:n(t)}("undefined"!=typeof window?window:this,(function(n,i){"use strict";var o=[],s=Object.getPrototypeOf,a=o.slice,u=o.flat?function(e){return o.flat.call(e)}:function(e){return o.concat.apply([],e)},l=o.push,c=o.indexOf,f={},h=f.toString,p=f.hasOwnProperty,d=p.toString,g=d.call(Object),v={},m=function(e){return"function"==typeof e&&"number"!=typeof e.nodeType&&"function"!=typeof e.item},y=function(e){return null!=e&&e===e.window},_=n.document,b={type:!0,src:!0,nonce:!0,noModule:!0};function w(e,t,n){var r,i,o=(n=n||_).createElement("script");if(o.text=e,t)for(r in b)(i=t[r]||t.getAttribute&&t.getAttribute(r))&&o.setAttribute(r,i);n.head.appendChild(o).parentNode.removeChild(o)}function x(e){return null==e?e+"":"object"==typeof e||"function"==typeof e?f[h.call(e)]||"object":typeof e}var E=function(e,t){return new E.fn.init(e,t)};function T(e){var t=!!e&&"length"in e&&e.length,n=x(e);return!m(e)&&!y(e)&&("array"===n||0===t||"number"==typeof t&&t>0&&t-1 in e)}E.fn=E.prototype={jquery:"3.6.0",constructor:E,length:0,toArray:function(){return a.call(this)},get:function(e){return null==e?a.call(this):e<0?this[e+this.length]:this[e]},pushStack:function(e){var t=E.merge(this.constructor(),e);return t.prevObject=this,t},each:function(e){return E.each(this,e)},map:function(e){return this.pushStack(E.map(this,(function(t,n){return e.call(t,n,t)})))},slice:function(){return this.pushStack(a.apply(this,arguments))},first:function(){return this.eq(0)},last:function(){return this.eq(-1)},even:function(){return this.pushStack(E.grep(this,(function(e,t){return(t+1)%2})))},odd:function(){return this.pushStack(E.grep(this,(function(e,t){return t%2})))},eq:function(e){var t=this.length,n=+e+(e<0?t:0);return this.pushStack(n>=0&&n<t?[this[n]]:[])},end:function(){return this.prevObject||this.constructor()},push:l,sort:o.sort,splice:o.splice},E.extend=E.fn.extend=function(){var e,t,n,r,i,o,s=arguments[0]||{},a=1,u=arguments.length,l=!1;for("boolean"==typeof s&&(l=s,s=arguments[a]||{},a++),"object"==typeof s||m(s)||(s={}),a===u&&(s=this,a--);a<u;a++)if(null!=(e=arguments[a]))for(t in e)r=e[t],"__proto__"!==t&&s!==r&&(l&&r&&(E.isPlainObject(r)||(i=Array.isArray(r)))?(n=s[t],o=i&&!Array.isArray(n)?[]:i||E.isPlainObject(n)?n:{},i=!1,s[t]=E.extend(l,o,r)):void 0!==r&&(s[t]=r));return s},E.extend({expando:"jQuery"+("3.6.0"+Math.random()).replace(/\D/g,""),isReady:!0,error:function(e){throw new Error(e)},noop:function(){},isPlainObject:function(e){var t,n;return!(!e||"[object Object]"!==h.call(e))&&(!(t=s(e))||"function"==typeof(n=p.call(t,"constructor")&&t.constructor)&&d.call(n)===g)},isEmptyObject:function(e){var t;for(t in e)return!1;return!0},globalEval:function(e,t,n){w(e,{nonce:t&&t.nonce},n)},each:function(e,t){var n,r=0;if(T(e))for(n=e.length;r<n&&!1!==t.call(e[r],r,e[r]);r++);else for(r in e)if(!1===t.call(e[r],r,e[r]))break;return e},makeArray:function(e,t){var n=t||[];return null!=e&&(T(Object(e))?E.merge(n,"string"==typeof e?[e]:e):l.call(n,e)),n},inArray:function(e,t,n){return null==t?-1:c.call(t,e,n)},merge:function(e,t){for(var n=+t.length,r=0,i=e.length;r<n;r++)e[i++]=t[r];return e.length=i,e},grep:function(e,t,n){for(var r=[],i=0,o=e.length,s=!n;i<o;i++)!t(e[i],i)!==s&&r.push(e[i]);return r},map:function(e,t,n){var r,i,o=0,s=[];if(T(e))for(r=e.length;o<r;o++)null!=(i=t(e[o],o,n))&&s.push(i);else for(o in e)null!=(i=t(e[o],o,n))&&s.push(i);return u(s)},guid:1,support:v}),"function"==typeof Symbol&&(E.fn[Symbol.iterator]=o[Symbol.iterator]),E.each("Boolean Number String Function Array Date RegExp Object Error Symbol".split(" "),(function(e,t){f["[object "+t+"]"]=t.toLowerCase()}));var S=
/*!
 * Sizzle CSS Selector Engine v2.3.6
 * https://sizzlejs.com/
 *
 * Copyright JS Foundation and other contributors
 * Released under the MIT license
 * https://js.foundation/
 *
 * Date: 2021-02-16
 */
function(e){var t,n,r,i,o,s,a,u,l,c,f,h,p,d,g,v,m,y,_,b="sizzle"+1*new Date,w=e.document,x=0,E=0,T=ue(),S=ue(),C=ue(),A=ue(),D=function(e,t){return e===t&&(f=!0),0},z={}.hasOwnProperty,k=[],N=k.pop,R=k.push,j=k.push,O=k.slice,F=function(e,t){for(var n=0,r=e.length;n<r;n++)if(e[n]===t)return n;return-1},L="checked|selected|async|autofocus|autoplay|controls|defer|disabled|hidden|ismap|loop|multiple|open|readonly|required|scoped",P="[\\x20\\t\\r\\n\\f]",q="(?:\\\\[\\da-fA-F]{1,6}"+P+"?|\\\\[^\\r\\n\\f]|[\\w-]|[^\0-\\x7f])+",I="\\["+P+"*("+q+")(?:"+P+"*([*^$|!~]?=)"+P+"*(?:'((?:\\\\.|[^\\\\'])*)'|\"((?:\\\\.|[^\\\\\"])*)\"|("+q+"))|)"+P+"*\\]",M=":("+q+")(?:\\((('((?:\\\\.|[^\\\\'])*)'|\"((?:\\\\.|[^\\\\\"])*)\")|((?:\\\\.|[^\\\\()[\\]]|"+I+")*)|.*)\\)|)",W=new RegExp(P+"+","g"),B=new RegExp("^"+P+"+|((?:^|[^\\\\])(?:\\\\.)*)"+P+"+$","g"),H=new RegExp("^"+P+"*,"+P+"*"),U=new RegExp("^"+P+"*([>+~]|"+P+")"+P+"*"),$=new RegExp(P+"|>"),X=new RegExp(M),V=new RegExp("^"+q+"$"),G={ID:new RegExp("^#("+q+")"),CLASS:new RegExp("^\\.("+q+")"),TAG:new RegExp("^("+q+"|[*])"),ATTR:new RegExp("^"+I),PSEUDO:new RegExp("^"+M),CHILD:new RegExp("^:(only|first|last|nth|nth-last)-(child|of-type)(?:\\("+P+"*(even|odd|(([+-]|)(\\d*)n|)"+P+"*(?:([+-]|)"+P+"*(\\d+)|))"+P+"*\\)|)","i"),bool:new RegExp("^(?:"+L+")$","i"),needsContext:new RegExp("^"+P+"*[>+~]|:(even|odd|eq|gt|lt|nth|first|last)(?:\\("+P+"*((?:-\\d)?\\d*)"+P+"*\\)|)(?=[^-]|$)","i")},K=/HTML$/i,Y=/^(?:input|select|textarea|button)$/i,J=/^h\d$/i,Z=/^[^{]+\{\s*\[native \w/,Q=/^(?:#([\w-]+)|(\w+)|\.([\w-]+))$/,ee=/[+~]/,te=new RegExp("\\\\[\\da-fA-F]{1,6}"+P+"?|\\\\([^\\r\\n\\f])","g"),ne=function(e,t){var n="0x"+e.slice(1)-65536;return t||(n<0?String.fromCharCode(n+65536):String.fromCharCode(n>>10|55296,1023&n|56320))},re=/([\0-\x1f\x7f]|^-?\d)|^-$|[^\0-\x1f\x7f-\uFFFF\w-]/g,ie=function(e,t){return t?"\0"===e?"�":e.slice(0,-1)+"\\"+e.charCodeAt(e.length-1).toString(16)+" ":"\\"+e},oe=function(){h()},se=be((function(e){return!0===e.disabled&&"fieldset"===e.nodeName.toLowerCase()}),{dir:"parentNode",next:"legend"});try{j.apply(k=O.call(w.childNodes),w.childNodes),k[w.childNodes.length].nodeType}catch(e){j={apply:k.length?function(e,t){R.apply(e,O.call(t))}:function(e,t){for(var n=e.length,r=0;e[n++]=t[r++];);e.length=n-1}}}function ae(e,t,r,i){var o,a,l,c,f,d,m,y=t&&t.ownerDocument,w=t?t.nodeType:9;if(r=r||[],"string"!=typeof e||!e||1!==w&&9!==w&&11!==w)return r;if(!i&&(h(t),t=t||p,g)){if(11!==w&&(f=Q.exec(e)))if(o=f[1]){if(9===w){if(!(l=t.getElementById(o)))return r;if(l.id===o)return r.push(l),r}else if(y&&(l=y.getElementById(o))&&_(t,l)&&l.id===o)return r.push(l),r}else{if(f[2])return j.apply(r,t.getElementsByTagName(e)),r;if((o=f[3])&&n.getElementsByClassName&&t.getElementsByClassName)return j.apply(r,t.getElementsByClassName(o)),r}if(n.qsa&&!A[e+" "]&&(!v||!v.test(e))&&(1!==w||"object"!==t.nodeName.toLowerCase())){if(m=e,y=t,1===w&&($.test(e)||U.test(e))){for((y=ee.test(e)&&me(t.parentNode)||t)===t&&n.scope||((c=t.getAttribute("id"))?c=c.replace(re,ie):t.setAttribute("id",c=b)),a=(d=s(e)).length;a--;)d[a]=(c?"#"+c:":scope")+" "+_e(d[a]);m=d.join(",")}try{return j.apply(r,y.querySelectorAll(m)),r}catch(t){A(e,!0)}finally{c===b&&t.removeAttribute("id")}}}return u(e.replace(B,"$1"),t,r,i)}function ue(){var e=[];return function t(n,i){return e.push(n+" ")>r.cacheLength&&delete t[e.shift()],t[n+" "]=i}}function le(e){return e[b]=!0,e}function ce(e){var t=p.createElement("fieldset");try{return!!e(t)}catch(e){return!1}finally{t.parentNode&&t.parentNode.removeChild(t),t=null}}function fe(e,t){for(var n=e.split("|"),i=n.length;i--;)r.attrHandle[n[i]]=t}function he(e,t){var n=t&&e,r=n&&1===e.nodeType&&1===t.nodeType&&e.sourceIndex-t.sourceIndex;if(r)return r;if(n)for(;n=n.nextSibling;)if(n===t)return-1;return e?1:-1}function pe(e){return function(t){return"input"===t.nodeName.toLowerCase()&&t.type===e}}function de(e){return function(t){var n=t.nodeName.toLowerCase();return("input"===n||"button"===n)&&t.type===e}}function ge(e){return function(t){return"form"in t?t.parentNode&&!1===t.disabled?"label"in t?"label"in t.parentNode?t.parentNode.disabled===e:t.disabled===e:t.isDisabled===e||t.isDisabled!==!e&&se(t)===e:t.disabled===e:"label"in t&&t.disabled===e}}function ve(e){return le((function(t){return t=+t,le((function(n,r){for(var i,o=e([],n.length,t),s=o.length;s--;)n[i=o[s]]&&(n[i]=!(r[i]=n[i]))}))}))}function me(e){return e&&void 0!==e.getElementsByTagName&&e}for(t in n=ae.support={},o=ae.isXML=function(e){var t=e&&e.namespaceURI,n=e&&(e.ownerDocument||e).documentElement;return!K.test(t||n&&n.nodeName||"HTML")},h=ae.setDocument=function(e){var t,i,s=e?e.ownerDocument||e:w;return s!=p&&9===s.nodeType&&s.documentElement?(d=(p=s).documentElement,g=!o(p),w!=p&&(i=p.defaultView)&&i.top!==i&&(i.addEventListener?i.addEventListener("unload",oe,!1):i.attachEvent&&i.attachEvent("onunload",oe)),n.scope=ce((function(e){return d.appendChild(e).appendChild(p.createElement("div")),void 0!==e.querySelectorAll&&!e.querySelectorAll(":scope fieldset div").length})),n.attributes=ce((function(e){return e.className="i",!e.getAttribute("className")})),n.getElementsByTagName=ce((function(e){return e.appendChild(p.createComment("")),!e.getElementsByTagName("*").length})),n.getElementsByClassName=Z.test(p.getElementsByClassName),n.getById=ce((function(e){return d.appendChild(e).id=b,!p.getElementsByName||!p.getElementsByName(b).length})),n.getById?(r.filter.ID=function(e){var t=e.replace(te,ne);return function(e){return e.getAttribute("id")===t}},r.find.ID=function(e,t){if(void 0!==t.getElementById&&g){var n=t.getElementById(e);return n?[n]:[]}}):(r.filter.ID=function(e){var t=e.replace(te,ne);return function(e){var n=void 0!==e.getAttributeNode&&e.getAttributeNode("id");return n&&n.value===t}},r.find.ID=function(e,t){if(void 0!==t.getElementById&&g){var n,r,i,o=t.getElementById(e);if(o){if((n=o.getAttributeNode("id"))&&n.value===e)return[o];for(i=t.getElementsByName(e),r=0;o=i[r++];)if((n=o.getAttributeNode("id"))&&n.value===e)return[o]}return[]}}),r.find.TAG=n.getElementsByTagName?function(e,t){return void 0!==t.getElementsByTagName?t.getElementsByTagName(e):n.qsa?t.querySelectorAll(e):void 0}:function(e,t){var n,r=[],i=0,o=t.getElementsByTagName(e);if("*"===e){for(;n=o[i++];)1===n.nodeType&&r.push(n);return r}return o},r.find.CLASS=n.getElementsByClassName&&function(e,t){if(void 0!==t.getElementsByClassName&&g)return t.getElementsByClassName(e)},m=[],v=[],(n.qsa=Z.test(p.querySelectorAll))&&(ce((function(e){var t;d.appendChild(e).innerHTML="<a id='"+b+"'></a><select id='"+b+"-\r\\' msallowcapture=''><option selected=''></option></select>",e.querySelectorAll("[msallowcapture^='']").length&&v.push("[*^$]="+P+"*(?:''|\"\")"),e.querySelectorAll("[selected]").length||v.push("\\["+P+"*(?:value|"+L+")"),e.querySelectorAll("[id~="+b+"-]").length||v.push("~="),(t=p.createElement("input")).setAttribute("name",""),e.appendChild(t),e.querySelectorAll("[name='']").length||v.push("\\["+P+"*name"+P+"*="+P+"*(?:''|\"\")"),e.querySelectorAll(":checked").length||v.push(":checked"),e.querySelectorAll("a#"+b+"+*").length||v.push(".#.+[+~]"),e.querySelectorAll("\\\f"),v.push("[\\r\\n\\f]")})),ce((function(e){e.innerHTML="<a href='' disabled='disabled'></a><select disabled='disabled'><option/></select>";var t=p.createElement("input");t.setAttribute("type","hidden"),e.appendChild(t).setAttribute("name","D"),e.querySelectorAll("[name=d]").length&&v.push("name"+P+"*[*^$|!~]?="),2!==e.querySelectorAll(":enabled").length&&v.push(":enabled",":disabled"),d.appendChild(e).disabled=!0,2!==e.querySelectorAll(":disabled").length&&v.push(":enabled",":disabled"),e.querySelectorAll("*,:x"),v.push(",.*:")}))),(n.matchesSelector=Z.test(y=d.matches||d.webkitMatchesSelector||d.mozMatchesSelector||d.oMatchesSelector||d.msMatchesSelector))&&ce((function(e){n.disconnectedMatch=y.call(e,"*"),y.call(e,"[s!='']:x"),m.push("!=",M)})),v=v.length&&new RegExp(v.join("|")),m=m.length&&new RegExp(m.join("|")),t=Z.test(d.compareDocumentPosition),_=t||Z.test(d.contains)?function(e,t){var n=9===e.nodeType?e.documentElement:e,r=t&&t.parentNode;return e===r||!(!r||1!==r.nodeType||!(n.contains?n.contains(r):e.compareDocumentPosition&&16&e.compareDocumentPosition(r)))}:function(e,t){if(t)for(;t=t.parentNode;)if(t===e)return!0;return!1},D=t?function(e,t){if(e===t)return f=!0,0;var r=!e.compareDocumentPosition-!t.compareDocumentPosition;return r||(1&(r=(e.ownerDocument||e)==(t.ownerDocument||t)?e.compareDocumentPosition(t):1)||!n.sortDetached&&t.compareDocumentPosition(e)===r?e==p||e.ownerDocument==w&&_(w,e)?-1:t==p||t.ownerDocument==w&&_(w,t)?1:c?F(c,e)-F(c,t):0:4&r?-1:1)}:function(e,t){if(e===t)return f=!0,0;var n,r=0,i=e.parentNode,o=t.parentNode,s=[e],a=[t];if(!i||!o)return e==p?-1:t==p?1:i?-1:o?1:c?F(c,e)-F(c,t):0;if(i===o)return he(e,t);for(n=e;n=n.parentNode;)s.unshift(n);for(n=t;n=n.parentNode;)a.unshift(n);for(;s[r]===a[r];)r++;return r?he(s[r],a[r]):s[r]==w?-1:a[r]==w?1:0},p):p},ae.matches=function(e,t){return ae(e,null,null,t)},ae.matchesSelector=function(e,t){if(h(e),n.matchesSelector&&g&&!A[t+" "]&&(!m||!m.test(t))&&(!v||!v.test(t)))try{var r=y.call(e,t);if(r||n.disconnectedMatch||e.document&&11!==e.document.nodeType)return r}catch(e){A(t,!0)}return ae(t,p,null,[e]).length>0},ae.contains=function(e,t){return(e.ownerDocument||e)!=p&&h(e),_(e,t)},ae.attr=function(e,t){(e.ownerDocument||e)!=p&&h(e);var i=r.attrHandle[t.toLowerCase()],o=i&&z.call(r.attrHandle,t.toLowerCase())?i(e,t,!g):void 0;return void 0!==o?o:n.attributes||!g?e.getAttribute(t):(o=e.getAttributeNode(t))&&o.specified?o.value:null},ae.escape=function(e){return(e+"").replace(re,ie)},ae.error=function(e){throw new Error("Syntax error, unrecognized expression: "+e)},ae.uniqueSort=function(e){var t,r=[],i=0,o=0;if(f=!n.detectDuplicates,c=!n.sortStable&&e.slice(0),e.sort(D),f){for(;t=e[o++];)t===e[o]&&(i=r.push(o));for(;i--;)e.splice(r[i],1)}return c=null,e},i=ae.getText=function(e){var t,n="",r=0,o=e.nodeType;if(o){if(1===o||9===o||11===o){if("string"==typeof e.textContent)return e.textContent;for(e=e.firstChild;e;e=e.nextSibling)n+=i(e)}else if(3===o||4===o)return e.nodeValue}else for(;t=e[r++];)n+=i(t);return n},(r=ae.selectors={cacheLength:50,createPseudo:le,match:G,attrHandle:{},find:{},relative:{">":{dir:"parentNode",first:!0}," ":{dir:"parentNode"},"+":{dir:"previousSibling",first:!0},"~":{dir:"previousSibling"}},preFilter:{ATTR:function(e){return e[1]=e[1].replace(te,ne),e[3]=(e[3]||e[4]||e[5]||"").replace(te,ne),"~="===e[2]&&(e[3]=" "+e[3]+" "),e.slice(0,4)},CHILD:function(e){return e[1]=e[1].toLowerCase(),"nth"===e[1].slice(0,3)?(e[3]||ae.error(e[0]),e[4]=+(e[4]?e[5]+(e[6]||1):2*("even"===e[3]||"odd"===e[3])),e[5]=+(e[7]+e[8]||"odd"===e[3])):e[3]&&ae.error(e[0]),e},PSEUDO:function(e){var t,n=!e[6]&&e[2];return G.CHILD.test(e[0])?null:(e[3]?e[2]=e[4]||e[5]||"":n&&X.test(n)&&(t=s(n,!0))&&(t=n.indexOf(")",n.length-t)-n.length)&&(e[0]=e[0].slice(0,t),e[2]=n.slice(0,t)),e.slice(0,3))}},filter:{TAG:function(e){var t=e.replace(te,ne).toLowerCase();return"*"===e?function(){return!0}:function(e){return e.nodeName&&e.nodeName.toLowerCase()===t}},CLASS:function(e){var t=T[e+" "];return t||(t=new RegExp("(^|"+P+")"+e+"("+P+"|$)"))&&T(e,(function(e){return t.test("string"==typeof e.className&&e.className||void 0!==e.getAttribute&&e.getAttribute("class")||"")}))},ATTR:function(e,t,n){return function(r){var i=ae.attr(r,e);return null==i?"!="===t:!t||(i+="","="===t?i===n:"!="===t?i!==n:"^="===t?n&&0===i.indexOf(n):"*="===t?n&&i.indexOf(n)>-1:"$="===t?n&&i.slice(-n.length)===n:"~="===t?(" "+i.replace(W," ")+" ").indexOf(n)>-1:"|="===t&&(i===n||i.slice(0,n.length+1)===n+"-"))}},CHILD:function(e,t,n,r,i){var o="nth"!==e.slice(0,3),s="last"!==e.slice(-4),a="of-type"===t;return 1===r&&0===i?function(e){return!!e.parentNode}:function(t,n,u){var l,c,f,h,p,d,g=o!==s?"nextSibling":"previousSibling",v=t.parentNode,m=a&&t.nodeName.toLowerCase(),y=!u&&!a,_=!1;if(v){if(o){for(;g;){for(h=t;h=h[g];)if(a?h.nodeName.toLowerCase()===m:1===h.nodeType)return!1;d=g="only"===e&&!d&&"nextSibling"}return!0}if(d=[s?v.firstChild:v.lastChild],s&&y){for(_=(p=(l=(c=(f=(h=v)[b]||(h[b]={}))[h.uniqueID]||(f[h.uniqueID]={}))[e]||[])[0]===x&&l[1])&&l[2],h=p&&v.childNodes[p];h=++p&&h&&h[g]||(_=p=0)||d.pop();)if(1===h.nodeType&&++_&&h===t){c[e]=[x,p,_];break}}else if(y&&(_=p=(l=(c=(f=(h=t)[b]||(h[b]={}))[h.uniqueID]||(f[h.uniqueID]={}))[e]||[])[0]===x&&l[1]),!1===_)for(;(h=++p&&h&&h[g]||(_=p=0)||d.pop())&&((a?h.nodeName.toLowerCase()!==m:1!==h.nodeType)||!++_||(y&&((c=(f=h[b]||(h[b]={}))[h.uniqueID]||(f[h.uniqueID]={}))[e]=[x,_]),h!==t)););return(_-=i)===r||_%r==0&&_/r>=0}}},PSEUDO:function(e,t){var n,i=r.pseudos[e]||r.setFilters[e.toLowerCase()]||ae.error("unsupported pseudo: "+e);return i[b]?i(t):i.length>1?(n=[e,e,"",t],r.setFilters.hasOwnProperty(e.toLowerCase())?le((function(e,n){for(var r,o=i(e,t),s=o.length;s--;)e[r=F(e,o[s])]=!(n[r]=o[s])})):function(e){return i(e,0,n)}):i}},pseudos:{not:le((function(e){var t=[],n=[],r=a(e.replace(B,"$1"));return r[b]?le((function(e,t,n,i){for(var o,s=r(e,null,i,[]),a=e.length;a--;)(o=s[a])&&(e[a]=!(t[a]=o))})):function(e,i,o){return t[0]=e,r(t,null,o,n),t[0]=null,!n.pop()}})),has:le((function(e){return function(t){return ae(e,t).length>0}})),contains:le((function(e){return e=e.replace(te,ne),function(t){return(t.textContent||i(t)).indexOf(e)>-1}})),lang:le((function(e){return V.test(e||"")||ae.error("unsupported lang: "+e),e=e.replace(te,ne).toLowerCase(),function(t){var n;do{if(n=g?t.lang:t.getAttribute("xml:lang")||t.getAttribute("lang"))return(n=n.toLowerCase())===e||0===n.indexOf(e+"-")}while((t=t.parentNode)&&1===t.nodeType);return!1}})),target:function(t){var n=e.location&&e.location.hash;return n&&n.slice(1)===t.id},root:function(e){return e===d},focus:function(e){return e===p.activeElement&&(!p.hasFocus||p.hasFocus())&&!!(e.type||e.href||~e.tabIndex)},enabled:ge(!1),disabled:ge(!0),checked:function(e){var t=e.nodeName.toLowerCase();return"input"===t&&!!e.checked||"option"===t&&!!e.selected},selected:function(e){return e.parentNode&&e.parentNode.selectedIndex,!0===e.selected},empty:function(e){for(e=e.firstChild;e;e=e.nextSibling)if(e.nodeType<6)return!1;return!0},parent:function(e){return!r.pseudos.empty(e)},header:function(e){return J.test(e.nodeName)},input:function(e){return Y.test(e.nodeName)},button:function(e){var t=e.nodeName.toLowerCase();return"input"===t&&"button"===e.type||"button"===t},text:function(e){var t;return"input"===e.nodeName.toLowerCase()&&"text"===e.type&&(null==(t=e.getAttribute("type"))||"text"===t.toLowerCase())},first:ve((function(){return[0]})),last:ve((function(e,t){return[t-1]})),eq:ve((function(e,t,n){return[n<0?n+t:n]})),even:ve((function(e,t){for(var n=0;n<t;n+=2)e.push(n);return e})),odd:ve((function(e,t){for(var n=1;n<t;n+=2)e.push(n);return e})),lt:ve((function(e,t,n){for(var r=n<0?n+t:n>t?t:n;--r>=0;)e.push(r);return e})),gt:ve((function(e,t,n){for(var r=n<0?n+t:n;++r<t;)e.push(r);return e}))}}).pseudos.nth=r.pseudos.eq,{radio:!0,checkbox:!0,file:!0,password:!0,image:!0})r.pseudos[t]=pe(t);for(t in{submit:!0,reset:!0})r.pseudos[t]=de(t);function ye(){}function _e(e){for(var t=0,n=e.length,r="";t<n;t++)r+=e[t].value;return r}function be(e,t,n){var r=t.dir,i=t.next,o=i||r,s=n&&"parentNode"===o,a=E++;return t.first?function(t,n,i){for(;t=t[r];)if(1===t.nodeType||s)return e(t,n,i);return!1}:function(t,n,u){var l,c,f,h=[x,a];if(u){for(;t=t[r];)if((1===t.nodeType||s)&&e(t,n,u))return!0}else for(;t=t[r];)if(1===t.nodeType||s)if(c=(f=t[b]||(t[b]={}))[t.uniqueID]||(f[t.uniqueID]={}),i&&i===t.nodeName.toLowerCase())t=t[r]||t;else{if((l=c[o])&&l[0]===x&&l[1]===a)return h[2]=l[2];if(c[o]=h,h[2]=e(t,n,u))return!0}return!1}}function we(e){return e.length>1?function(t,n,r){for(var i=e.length;i--;)if(!e[i](t,n,r))return!1;return!0}:e[0]}function xe(e,t,n,r,i){for(var o,s=[],a=0,u=e.length,l=null!=t;a<u;a++)(o=e[a])&&(n&&!n(o,r,i)||(s.push(o),l&&t.push(a)));return s}function Ee(e,t,n,r,i,o){return r&&!r[b]&&(r=Ee(r)),i&&!i[b]&&(i=Ee(i,o)),le((function(o,s,a,u){var l,c,f,h=[],p=[],d=s.length,g=o||function(e,t,n){for(var r=0,i=t.length;r<i;r++)ae(e,t[r],n);return n}(t||"*",a.nodeType?[a]:a,[]),v=!e||!o&&t?g:xe(g,h,e,a,u),m=n?i||(o?e:d||r)?[]:s:v;if(n&&n(v,m,a,u),r)for(l=xe(m,p),r(l,[],a,u),c=l.length;c--;)(f=l[c])&&(m[p[c]]=!(v[p[c]]=f));if(o){if(i||e){if(i){for(l=[],c=m.length;c--;)(f=m[c])&&l.push(v[c]=f);i(null,m=[],l,u)}for(c=m.length;c--;)(f=m[c])&&(l=i?F(o,f):h[c])>-1&&(o[l]=!(s[l]=f))}}else m=xe(m===s?m.splice(d,m.length):m),i?i(null,s,m,u):j.apply(s,m)}))}function Te(e){for(var t,n,i,o=e.length,s=r.relative[e[0].type],a=s||r.relative[" "],u=s?1:0,c=be((function(e){return e===t}),a,!0),f=be((function(e){return F(t,e)>-1}),a,!0),h=[function(e,n,r){var i=!s&&(r||n!==l)||((t=n).nodeType?c(e,n,r):f(e,n,r));return t=null,i}];u<o;u++)if(n=r.relative[e[u].type])h=[be(we(h),n)];else{if((n=r.filter[e[u].type].apply(null,e[u].matches))[b]){for(i=++u;i<o&&!r.relative[e[i].type];i++);return Ee(u>1&&we(h),u>1&&_e(e.slice(0,u-1).concat({value:" "===e[u-2].type?"*":""})).replace(B,"$1"),n,u<i&&Te(e.slice(u,i)),i<o&&Te(e=e.slice(i)),i<o&&_e(e))}h.push(n)}return we(h)}return ye.prototype=r.filters=r.pseudos,r.setFilters=new ye,s=ae.tokenize=function(e,t){var n,i,o,s,a,u,l,c=S[e+" "];if(c)return t?0:c.slice(0);for(a=e,u=[],l=r.preFilter;a;){for(s in n&&!(i=H.exec(a))||(i&&(a=a.slice(i[0].length)||a),u.push(o=[])),n=!1,(i=U.exec(a))&&(n=i.shift(),o.push({value:n,type:i[0].replace(B," ")}),a=a.slice(n.length)),r.filter)!(i=G[s].exec(a))||l[s]&&!(i=l[s](i))||(n=i.shift(),o.push({value:n,type:s,matches:i}),a=a.slice(n.length));if(!n)break}return t?a.length:a?ae.error(e):S(e,u).slice(0)},a=ae.compile=function(e,t){var n,i=[],o=[],a=C[e+" "];if(!a){for(t||(t=s(e)),n=t.length;n--;)(a=Te(t[n]))[b]?i.push(a):o.push(a);(a=C(e,function(e,t){var n=t.length>0,i=e.length>0,o=function(o,s,a,u,c){var f,d,v,m=0,y="0",_=o&&[],b=[],w=l,E=o||i&&r.find.TAG("*",c),T=x+=null==w?1:Math.random()||.1,S=E.length;for(c&&(l=s==p||s||c);y!==S&&null!=(f=E[y]);y++){if(i&&f){for(d=0,s||f.ownerDocument==p||(h(f),a=!g);v=e[d++];)if(v(f,s||p,a)){u.push(f);break}c&&(x=T)}n&&((f=!v&&f)&&m--,o&&_.push(f))}if(m+=y,n&&y!==m){for(d=0;v=t[d++];)v(_,b,s,a);if(o){if(m>0)for(;y--;)_[y]||b[y]||(b[y]=N.call(u));b=xe(b)}j.apply(u,b),c&&!o&&b.length>0&&m+t.length>1&&ae.uniqueSort(u)}return c&&(x=T,l=w),_};return n?le(o):o}(o,i))).selector=e}return a},u=ae.select=function(e,t,n,i){var o,u,l,c,f,h="function"==typeof e&&e,p=!i&&s(e=h.selector||e);if(n=n||[],1===p.length){if((u=p[0]=p[0].slice(0)).length>2&&"ID"===(l=u[0]).type&&9===t.nodeType&&g&&r.relative[u[1].type]){if(!(t=(r.find.ID(l.matches[0].replace(te,ne),t)||[])[0]))return n;h&&(t=t.parentNode),e=e.slice(u.shift().value.length)}for(o=G.needsContext.test(e)?0:u.length;o--&&(l=u[o],!r.relative[c=l.type]);)if((f=r.find[c])&&(i=f(l.matches[0].replace(te,ne),ee.test(u[0].type)&&me(t.parentNode)||t))){if(u.splice(o,1),!(e=i.length&&_e(u)))return j.apply(n,i),n;break}}return(h||a(e,p))(i,t,!g,n,!t||ee.test(e)&&me(t.parentNode)||t),n},n.sortStable=b.split("").sort(D).join("")===b,n.detectDuplicates=!!f,h(),n.sortDetached=ce((function(e){return 1&e.compareDocumentPosition(p.createElement("fieldset"))})),ce((function(e){return e.innerHTML="<a href='#'></a>","#"===e.firstChild.getAttribute("href")}))||fe("type|href|height|width",(function(e,t,n){if(!n)return e.getAttribute(t,"type"===t.toLowerCase()?1:2)})),n.attributes&&ce((function(e){return e.innerHTML="<input/>",e.firstChild.setAttribute("value",""),""===e.firstChild.getAttribute("value")}))||fe("value",(function(e,t,n){if(!n&&"input"===e.nodeName.toLowerCase())return e.defaultValue})),ce((function(e){return null==e.getAttribute("disabled")}))||fe(L,(function(e,t,n){var r;if(!n)return!0===e[t]?t.toLowerCase():(r=e.getAttributeNode(t))&&r.specified?r.value:null})),ae}(n);E.find=S,E.expr=S.selectors,E.expr[":"]=E.expr.pseudos,E.uniqueSort=E.unique=S.uniqueSort,E.text=S.getText,E.isXMLDoc=S.isXML,E.contains=S.contains,E.escapeSelector=S.escape;var C=function(e,t,n){for(var r=[],i=void 0!==n;(e=e[t])&&9!==e.nodeType;)if(1===e.nodeType){if(i&&E(e).is(n))break;r.push(e)}return r},A=function(e,t){for(var n=[];e;e=e.nextSibling)1===e.nodeType&&e!==t&&n.push(e);return n},D=E.expr.match.needsContext;function z(e,t){return e.nodeName&&e.nodeName.toLowerCase()===t.toLowerCase()}var k=/^<([a-z][^\/\0>:\x20\t\r\n\f]*)[\x20\t\r\n\f]*\/?>(?:<\/\1>|)$/i;function N(e,t,n){return m(t)?E.grep(e,(function(e,r){return!!t.call(e,r,e)!==n})):t.nodeType?E.grep(e,(function(e){return e===t!==n})):"string"!=typeof t?E.grep(e,(function(e){return c.call(t,e)>-1!==n})):E.filter(t,e,n)}E.filter=function(e,t,n){var r=t[0];return n&&(e=":not("+e+")"),1===t.length&&1===r.nodeType?E.find.matchesSelector(r,e)?[r]:[]:E.find.matches(e,E.grep(t,(function(e){return 1===e.nodeType})))},E.fn.extend({find:function(e){var t,n,r=this.length,i=this;if("string"!=typeof e)return this.pushStack(E(e).filter((function(){for(t=0;t<r;t++)if(E.contains(i[t],this))return!0})));for(n=this.pushStack([]),t=0;t<r;t++)E.find(e,i[t],n);return r>1?E.uniqueSort(n):n},filter:function(e){return this.pushStack(N(this,e||[],!1))},not:function(e){return this.pushStack(N(this,e||[],!0))},is:function(e){return!!N(this,"string"==typeof e&&D.test(e)?E(e):e||[],!1).length}});var R,j=/^(?:\s*(<[\w\W]+>)[^>]*|#([\w-]+))$/;(E.fn.init=function(e,t,n){var r,i;if(!e)return this;if(n=n||R,"string"==typeof e){if(!(r="<"===e[0]&&">"===e[e.length-1]&&e.length>=3?[null,e,null]:j.exec(e))||!r[1]&&t)return!t||t.jquery?(t||n).find(e):this.constructor(t).find(e);if(r[1]){if(t=t instanceof E?t[0]:t,E.merge(this,E.parseHTML(r[1],t&&t.nodeType?t.ownerDocument||t:_,!0)),k.test(r[1])&&E.isPlainObject(t))for(r in t)m(this[r])?this[r](t[r]):this.attr(r,t[r]);return this}return(i=_.getElementById(r[2]))&&(this[0]=i,this.length=1),this}return e.nodeType?(this[0]=e,this.length=1,this):m(e)?void 0!==n.ready?n.ready(e):e(E):E.makeArray(e,this)}).prototype=E.fn,R=E(_);var O=/^(?:parents|prev(?:Until|All))/,F={children:!0,contents:!0,next:!0,prev:!0};function L(e,t){for(;(e=e[t])&&1!==e.nodeType;);return e}E.fn.extend({has:function(e){var t=E(e,this),n=t.length;return this.filter((function(){for(var e=0;e<n;e++)if(E.contains(this,t[e]))return!0}))},closest:function(e,t){var n,r=0,i=this.length,o=[],s="string"!=typeof e&&E(e);if(!D.test(e))for(;r<i;r++)for(n=this[r];n&&n!==t;n=n.parentNode)if(n.nodeType<11&&(s?s.index(n)>-1:1===n.nodeType&&E.find.matchesSelector(n,e))){o.push(n);break}return this.pushStack(o.length>1?E.uniqueSort(o):o)},index:function(e){return e?"string"==typeof e?c.call(E(e),this[0]):c.call(this,e.jquery?e[0]:e):this[0]&&this[0].parentNode?this.first().prevAll().length:-1},add:function(e,t){return this.pushStack(E.uniqueSort(E.merge(this.get(),E(e,t))))},addBack:function(e){return this.add(null==e?this.prevObject:this.prevObject.filter(e))}}),E.each({parent:function(e){var t=e.parentNode;return t&&11!==t.nodeType?t:null},parents:function(e){return C(e,"parentNode")},parentsUntil:function(e,t,n){return C(e,"parentNode",n)},next:function(e){return L(e,"nextSibling")},prev:function(e){return L(e,"previousSibling")},nextAll:function(e){return C(e,"nextSibling")},prevAll:function(e){return C(e,"previousSibling")},nextUntil:function(e,t,n){return C(e,"nextSibling",n)},prevUntil:function(e,t,n){return C(e,"previousSibling",n)},siblings:function(e){return A((e.parentNode||{}).firstChild,e)},children:function(e){return A(e.firstChild)},contents:function(e){return null!=e.contentDocument&&s(e.contentDocument)?e.contentDocument:(z(e,"template")&&(e=e.content||e),E.merge([],e.childNodes))}},(function(e,t){E.fn[e]=function(n,r){var i=E.map(this,t,n);return"Until"!==e.slice(-5)&&(r=n),r&&"string"==typeof r&&(i=E.filter(r,i)),this.length>1&&(F[e]||E.uniqueSort(i),O.test(e)&&i.reverse()),this.pushStack(i)}}));var P=/[^\x20\t\r\n\f]+/g;function q(e){return e}function I(e){throw e}function M(e,t,n,r){var i;try{e&&m(i=e.promise)?i.call(e).done(t).fail(n):e&&m(i=e.then)?i.call(e,t,n):t.apply(void 0,[e].slice(r))}catch(e){n.apply(void 0,[e])}}E.Callbacks=function(e){e="string"==typeof e?function(e){var t={};return E.each(e.match(P)||[],(function(e,n){t[n]=!0})),t}(e):E.extend({},e);var t,n,r,i,o=[],s=[],a=-1,u=function(){for(i=i||e.once,r=t=!0;s.length;a=-1)for(n=s.shift();++a<o.length;)!1===o[a].apply(n[0],n[1])&&e.stopOnFalse&&(a=o.length,n=!1);e.memory||(n=!1),t=!1,i&&(o=n?[]:"")},l={add:function(){return o&&(n&&!t&&(a=o.length-1,s.push(n)),function t(n){E.each(n,(function(n,r){m(r)?e.unique&&l.has(r)||o.push(r):r&&r.length&&"string"!==x(r)&&t(r)}))}(arguments),n&&!t&&u()),this},remove:function(){return E.each(arguments,(function(e,t){for(var n;(n=E.inArray(t,o,n))>-1;)o.splice(n,1),n<=a&&a--})),this},has:function(e){return e?E.inArray(e,o)>-1:o.length>0},empty:function(){return o&&(o=[]),this},disable:function(){return i=s=[],o=n="",this},disabled:function(){return!o},lock:function(){return i=s=[],n||t||(o=n=""),this},locked:function(){return!!i},fireWith:function(e,n){return i||(n=[e,(n=n||[]).slice?n.slice():n],s.push(n),t||u()),this},fire:function(){return l.fireWith(this,arguments),this},fired:function(){return!!r}};return l},E.extend({Deferred:function(e){var t=[["notify","progress",E.Callbacks("memory"),E.Callbacks("memory"),2],["resolve","done",E.Callbacks("once memory"),E.Callbacks("once memory"),0,"resolved"],["reject","fail",E.Callbacks("once memory"),E.Callbacks("once memory"),1,"rejected"]],r="pending",i={state:function(){return r},always:function(){return o.done(arguments).fail(arguments),this},catch:function(e){return i.then(null,e)},pipe:function(){var e=arguments;return E.Deferred((function(n){E.each(t,(function(t,r){var i=m(e[r[4]])&&e[r[4]];o[r[1]]((function(){var e=i&&i.apply(this,arguments);e&&m(e.promise)?e.promise().progress(n.notify).done(n.resolve).fail(n.reject):n[r[0]+"With"](this,i?[e]:arguments)}))})),e=null})).promise()},then:function(e,r,i){var o=0;function s(e,t,r,i){return function(){var a=this,u=arguments,l=function(){var n,l;if(!(e<o)){if((n=r.apply(a,u))===t.promise())throw new TypeError("Thenable self-resolution");l=n&&("object"==typeof n||"function"==typeof n)&&n.then,m(l)?i?l.call(n,s(o,t,q,i),s(o,t,I,i)):(o++,l.call(n,s(o,t,q,i),s(o,t,I,i),s(o,t,q,t.notifyWith))):(r!==q&&(a=void 0,u=[n]),(i||t.resolveWith)(a,u))}},c=i?l:function(){try{l()}catch(n){E.Deferred.exceptionHook&&E.Deferred.exceptionHook(n,c.stackTrace),e+1>=o&&(r!==I&&(a=void 0,u=[n]),t.rejectWith(a,u))}};e?c():(E.Deferred.getStackHook&&(c.stackTrace=E.Deferred.getStackHook()),n.setTimeout(c))}}return E.Deferred((function(n){t[0][3].add(s(0,n,m(i)?i:q,n.notifyWith)),t[1][3].add(s(0,n,m(e)?e:q)),t[2][3].add(s(0,n,m(r)?r:I))})).promise()},promise:function(e){return null!=e?E.extend(e,i):i}},o={};return E.each(t,(function(e,n){var s=n[2],a=n[5];i[n[1]]=s.add,a&&s.add((function(){r=a}),t[3-e][2].disable,t[3-e][3].disable,t[0][2].lock,t[0][3].lock),s.add(n[3].fire),o[n[0]]=function(){return o[n[0]+"With"](this===o?void 0:this,arguments),this},o[n[0]+"With"]=s.fireWith})),i.promise(o),e&&e.call(o,o),o},when:function(e){var t=arguments.length,n=t,r=Array(n),i=a.call(arguments),o=E.Deferred(),s=function(e){return function(n){r[e]=this,i[e]=arguments.length>1?a.call(arguments):n,--t||o.resolveWith(r,i)}};if(t<=1&&(M(e,o.done(s(n)).resolve,o.reject,!t),"pending"===o.state()||m(i[n]&&i[n].then)))return o.then();for(;n--;)M(i[n],s(n),o.reject);return o.promise()}});var W=/^(Eval|Internal|Range|Reference|Syntax|Type|URI)Error$/;E.Deferred.exceptionHook=function(e,t){n.console&&n.console.warn&&e&&W.test(e.name)&&n.console.warn("jQuery.Deferred exception: "+e.message,e.stack,t)},E.readyException=function(e){n.setTimeout((function(){throw e}))};var B=E.Deferred();function H(){_.removeEventListener("DOMContentLoaded",H),n.removeEventListener("load",H),E.ready()}E.fn.ready=function(e){return B.then(e).catch((function(e){E.readyException(e)})),this},E.extend({isReady:!1,readyWait:1,ready:function(e){(!0===e?--E.readyWait:E.isReady)||(E.isReady=!0,!0!==e&&--E.readyWait>0||B.resolveWith(_,[E]))}}),E.ready.then=B.then,"complete"===_.readyState||"loading"!==_.readyState&&!_.documentElement.doScroll?n.setTimeout(E.ready):(_.addEventListener("DOMContentLoaded",H),n.addEventListener("load",H));var U=function(e,t,n,r,i,o,s){var a=0,u=e.length,l=null==n;if("object"===x(n))for(a in i=!0,n)U(e,t,a,n[a],!0,o,s);else if(void 0!==r&&(i=!0,m(r)||(s=!0),l&&(s?(t.call(e,r),t=null):(l=t,t=function(e,t,n){return l.call(E(e),n)})),t))for(;a<u;a++)t(e[a],n,s?r:r.call(e[a],a,t(e[a],n)));return i?e:l?t.call(e):u?t(e[0],n):o},$=/^-ms-/,X=/-([a-z])/g;function V(e,t){return t.toUpperCase()}function G(e){return e.replace($,"ms-").replace(X,V)}var K=function(e){return 1===e.nodeType||9===e.nodeType||!+e.nodeType};function Y(){this.expando=E.expando+Y.uid++}Y.uid=1,Y.prototype={cache:function(e){var t=e[this.expando];return t||(t={},K(e)&&(e.nodeType?e[this.expando]=t:Object.defineProperty(e,this.expando,{value:t,configurable:!0}))),t},set:function(e,t,n){var r,i=this.cache(e);if("string"==typeof t)i[G(t)]=n;else for(r in t)i[G(r)]=t[r];return i},get:function(e,t){return void 0===t?this.cache(e):e[this.expando]&&e[this.expando][G(t)]},access:function(e,t,n){return void 0===t||t&&"string"==typeof t&&void 0===n?this.get(e,t):(this.set(e,t,n),void 0!==n?n:t)},remove:function(e,t){var n,r=e[this.expando];if(void 0!==r){if(void 0!==t){n=(t=Array.isArray(t)?t.map(G):(t=G(t))in r?[t]:t.match(P)||[]).length;for(;n--;)delete r[t[n]]}(void 0===t||E.isEmptyObject(r))&&(e.nodeType?e[this.expando]=void 0:delete e[this.expando])}},hasData:function(e){var t=e[this.expando];return void 0!==t&&!E.isEmptyObject(t)}};var J=new Y,Z=new Y,Q=/^(?:\{[\w\W]*\}|\[[\w\W]*\])$/,ee=/[A-Z]/g;function te(e,t,n){var r;if(void 0===n&&1===e.nodeType)if(r="data-"+t.replace(ee,"-$&").toLowerCase(),"string"==typeof(n=e.getAttribute(r))){try{n=function(e){return"true"===e||"false"!==e&&("null"===e?null:e===+e+""?+e:Q.test(e)?JSON.parse(e):e)}(n)}catch(e){}Z.set(e,t,n)}else n=void 0;return n}E.extend({hasData:function(e){return Z.hasData(e)||J.hasData(e)},data:function(e,t,n){return Z.access(e,t,n)},removeData:function(e,t){Z.remove(e,t)},_data:function(e,t,n){return J.access(e,t,n)},_removeData:function(e,t){J.remove(e,t)}}),E.fn.extend({data:function(e,t){var n,r,i,o=this[0],s=o&&o.attributes;if(void 0===e){if(this.length&&(i=Z.get(o),1===o.nodeType&&!J.get(o,"hasDataAttrs"))){for(n=s.length;n--;)s[n]&&0===(r=s[n].name).indexOf("data-")&&(r=G(r.slice(5)),te(o,r,i[r]));J.set(o,"hasDataAttrs",!0)}return i}return"object"==typeof e?this.each((function(){Z.set(this,e)})):U(this,(function(t){var n;if(o&&void 0===t)return void 0!==(n=Z.get(o,e))||void 0!==(n=te(o,e))?n:void 0;this.each((function(){Z.set(this,e,t)}))}),null,t,arguments.length>1,null,!0)},removeData:function(e){return this.each((function(){Z.remove(this,e)}))}}),E.extend({queue:function(e,t,n){var r;if(e)return t=(t||"fx")+"queue",r=J.get(e,t),n&&(!r||Array.isArray(n)?r=J.access(e,t,E.makeArray(n)):r.push(n)),r||[]},dequeue:function(e,t){t=t||"fx";var n=E.queue(e,t),r=n.length,i=n.shift(),o=E._queueHooks(e,t);"inprogress"===i&&(i=n.shift(),r--),i&&("fx"===t&&n.unshift("inprogress"),delete o.stop,i.call(e,(function(){E.dequeue(e,t)}),o)),!r&&o&&o.empty.fire()},_queueHooks:function(e,t){var n=t+"queueHooks";return J.get(e,n)||J.access(e,n,{empty:E.Callbacks("once memory").add((function(){J.remove(e,[t+"queue",n])}))})}}),E.fn.extend({queue:function(e,t){var n=2;return"string"!=typeof e&&(t=e,e="fx",n--),arguments.length<n?E.queue(this[0],e):void 0===t?this:this.each((function(){var n=E.queue(this,e,t);E._queueHooks(this,e),"fx"===e&&"inprogress"!==n[0]&&E.dequeue(this,e)}))},dequeue:function(e){return this.each((function(){E.dequeue(this,e)}))},clearQueue:function(e){return this.queue(e||"fx",[])},promise:function(e,t){var n,r=1,i=E.Deferred(),o=this,s=this.length,a=function(){--r||i.resolveWith(o,[o])};for("string"!=typeof e&&(t=e,e=void 0),e=e||"fx";s--;)(n=J.get(o[s],e+"queueHooks"))&&n.empty&&(r++,n.empty.add(a));return a(),i.promise(t)}});var ne=/[+-]?(?:\d*\.|)\d+(?:[eE][+-]?\d+|)/.source,re=new RegExp("^(?:([+-])=|)("+ne+")([a-z%]*)$","i"),ie=["Top","Right","Bottom","Left"],oe=_.documentElement,se=function(e){return E.contains(e.ownerDocument,e)},ae={composed:!0};oe.getRootNode&&(se=function(e){return E.contains(e.ownerDocument,e)||e.getRootNode(ae)===e.ownerDocument});var ue=function(e,t){return"none"===(e=t||e).style.display||""===e.style.display&&se(e)&&"none"===E.css(e,"display")};function le(e,t,n,r){var i,o,s=20,a=r?function(){return r.cur()}:function(){return E.css(e,t,"")},u=a(),l=n&&n[3]||(E.cssNumber[t]?"":"px"),c=e.nodeType&&(E.cssNumber[t]||"px"!==l&&+u)&&re.exec(E.css(e,t));if(c&&c[3]!==l){for(u/=2,l=l||c[3],c=+u||1;s--;)E.style(e,t,c+l),(1-o)*(1-(o=a()/u||.5))<=0&&(s=0),c/=o;c*=2,E.style(e,t,c+l),n=n||[]}return n&&(c=+c||+u||0,i=n[1]?c+(n[1]+1)*n[2]:+n[2],r&&(r.unit=l,r.start=c,r.end=i)),i}var ce={};function fe(e){var t,n=e.ownerDocument,r=e.nodeName,i=ce[r];return i||(t=n.body.appendChild(n.createElement(r)),i=E.css(t,"display"),t.parentNode.removeChild(t),"none"===i&&(i="block"),ce[r]=i,i)}function he(e,t){for(var n,r,i=[],o=0,s=e.length;o<s;o++)(r=e[o]).style&&(n=r.style.display,t?("none"===n&&(i[o]=J.get(r,"display")||null,i[o]||(r.style.display="")),""===r.style.display&&ue(r)&&(i[o]=fe(r))):"none"!==n&&(i[o]="none",J.set(r,"display",n)));for(o=0;o<s;o++)null!=i[o]&&(e[o].style.display=i[o]);return e}E.fn.extend({show:function(){return he(this,!0)},hide:function(){return he(this)},toggle:function(e){return"boolean"==typeof e?e?this.show():this.hide():this.each((function(){ue(this)?E(this).show():E(this).hide()}))}});var pe,de,ge=/^(?:checkbox|radio)$/i,ve=/<([a-z][^\/\0>\x20\t\r\n\f]*)/i,me=/^$|^module$|\/(?:java|ecma)script/i;pe=_.createDocumentFragment().appendChild(_.createElement("div")),(de=_.createElement("input")).setAttribute("type","radio"),de.setAttribute("checked","checked"),de.setAttribute("name","t"),pe.appendChild(de),v.checkClone=pe.cloneNode(!0).cloneNode(!0).lastChild.checked,pe.innerHTML="<textarea>x</textarea>",v.noCloneChecked=!!pe.cloneNode(!0).lastChild.defaultValue,pe.innerHTML="<option></option>",v.option=!!pe.lastChild;var ye={thead:[1,"<table>","</table>"],col:[2,"<table><colgroup>","</colgroup></table>"],tr:[2,"<table><tbody>","</tbody></table>"],td:[3,"<table><tbody><tr>","</tr></tbody></table>"],_default:[0,"",""]};function _e(e,t){var n;return n=void 0!==e.getElementsByTagName?e.getElementsByTagName(t||"*"):void 0!==e.querySelectorAll?e.querySelectorAll(t||"*"):[],void 0===t||t&&z(e,t)?E.merge([e],n):n}function be(e,t){for(var n=0,r=e.length;n<r;n++)J.set(e[n],"globalEval",!t||J.get(t[n],"globalEval"))}ye.tbody=ye.tfoot=ye.colgroup=ye.caption=ye.thead,ye.th=ye.td,v.option||(ye.optgroup=ye.option=[1,"<select multiple='multiple'>","</select>"]);var we=/<|&#?\w+;/;function xe(e,t,n,r,i){for(var o,s,a,u,l,c,f=t.createDocumentFragment(),h=[],p=0,d=e.length;p<d;p++)if((o=e[p])||0===o)if("object"===x(o))E.merge(h,o.nodeType?[o]:o);else if(we.test(o)){for(s=s||f.appendChild(t.createElement("div")),a=(ve.exec(o)||["",""])[1].toLowerCase(),u=ye[a]||ye._default,s.innerHTML=u[1]+E.htmlPrefilter(o)+u[2],c=u[0];c--;)s=s.lastChild;E.merge(h,s.childNodes),(s=f.firstChild).textContent=""}else h.push(t.createTextNode(o));for(f.textContent="",p=0;o=h[p++];)if(r&&E.inArray(o,r)>-1)i&&i.push(o);else if(l=se(o),s=_e(f.appendChild(o),"script"),l&&be(s),n)for(c=0;o=s[c++];)me.test(o.type||"")&&n.push(o);return f}var Ee=/^([^.]*)(?:\.(.+)|)/;function Te(){return!0}function Se(){return!1}function Ce(e,t){return e===function(){try{return _.activeElement}catch(e){}}()==("focus"===t)}function Ae(e,t,n,r,i,o){var s,a;if("object"==typeof t){for(a in"string"!=typeof n&&(r=r||n,n=void 0),t)Ae(e,a,n,r,t[a],o);return e}if(null==r&&null==i?(i=n,r=n=void 0):null==i&&("string"==typeof n?(i=r,r=void 0):(i=r,r=n,n=void 0)),!1===i)i=Se;else if(!i)return e;return 1===o&&(s=i,(i=function(e){return E().off(e),s.apply(this,arguments)}).guid=s.guid||(s.guid=E.guid++)),e.each((function(){E.event.add(this,t,i,r,n)}))}function De(e,t,n){n?(J.set(e,t,!1),E.event.add(e,t,{namespace:!1,handler:function(e){var r,i,o=J.get(this,t);if(1&e.isTrigger&&this[t]){if(o.length)(E.event.special[t]||{}).delegateType&&e.stopPropagation();else if(o=a.call(arguments),J.set(this,t,o),r=n(this,t),this[t](),o!==(i=J.get(this,t))||r?J.set(this,t,!1):i={},o!==i)return e.stopImmediatePropagation(),e.preventDefault(),i&&i.value}else o.length&&(J.set(this,t,{value:E.event.trigger(E.extend(o[0],E.Event.prototype),o.slice(1),this)}),e.stopImmediatePropagation())}})):void 0===J.get(e,t)&&E.event.add(e,t,Te)}E.event={global:{},add:function(e,t,n,r,i){var o,s,a,u,l,c,f,h,p,d,g,v=J.get(e);if(K(e))for(n.handler&&(n=(o=n).handler,i=o.selector),i&&E.find.matchesSelector(oe,i),n.guid||(n.guid=E.guid++),(u=v.events)||(u=v.events=Object.create(null)),(s=v.handle)||(s=v.handle=function(t){return void 0!==E&&E.event.triggered!==t.type?E.event.dispatch.apply(e,arguments):void 0}),l=(t=(t||"").match(P)||[""]).length;l--;)p=g=(a=Ee.exec(t[l])||[])[1],d=(a[2]||"").split(".").sort(),p&&(f=E.event.special[p]||{},p=(i?f.delegateType:f.bindType)||p,f=E.event.special[p]||{},c=E.extend({type:p,origType:g,data:r,handler:n,guid:n.guid,selector:i,needsContext:i&&E.expr.match.needsContext.test(i),namespace:d.join(".")},o),(h=u[p])||((h=u[p]=[]).delegateCount=0,f.setup&&!1!==f.setup.call(e,r,d,s)||e.addEventListener&&e.addEventListener(p,s)),f.add&&(f.add.call(e,c),c.handler.guid||(c.handler.guid=n.guid)),i?h.splice(h.delegateCount++,0,c):h.push(c),E.event.global[p]=!0)},remove:function(e,t,n,r,i){var o,s,a,u,l,c,f,h,p,d,g,v=J.hasData(e)&&J.get(e);if(v&&(u=v.events)){for(l=(t=(t||"").match(P)||[""]).length;l--;)if(p=g=(a=Ee.exec(t[l])||[])[1],d=(a[2]||"").split(".").sort(),p){for(f=E.event.special[p]||{},h=u[p=(r?f.delegateType:f.bindType)||p]||[],a=a[2]&&new RegExp("(^|\\.)"+d.join("\\.(?:.*\\.|)")+"(\\.|$)"),s=o=h.length;o--;)c=h[o],!i&&g!==c.origType||n&&n.guid!==c.guid||a&&!a.test(c.namespace)||r&&r!==c.selector&&("**"!==r||!c.selector)||(h.splice(o,1),c.selector&&h.delegateCount--,f.remove&&f.remove.call(e,c));s&&!h.length&&(f.teardown&&!1!==f.teardown.call(e,d,v.handle)||E.removeEvent(e,p,v.handle),delete u[p])}else for(p in u)E.event.remove(e,p+t[l],n,r,!0);E.isEmptyObject(u)&&J.remove(e,"handle events")}},dispatch:function(e){var t,n,r,i,o,s,a=new Array(arguments.length),u=E.event.fix(e),l=(J.get(this,"events")||Object.create(null))[u.type]||[],c=E.event.special[u.type]||{};for(a[0]=u,t=1;t<arguments.length;t++)a[t]=arguments[t];if(u.delegateTarget=this,!c.preDispatch||!1!==c.preDispatch.call(this,u)){for(s=E.event.handlers.call(this,u,l),t=0;(i=s[t++])&&!u.isPropagationStopped();)for(u.currentTarget=i.elem,n=0;(o=i.handlers[n++])&&!u.isImmediatePropagationStopped();)u.rnamespace&&!1!==o.namespace&&!u.rnamespace.test(o.namespace)||(u.handleObj=o,u.data=o.data,void 0!==(r=((E.event.special[o.origType]||{}).handle||o.handler).apply(i.elem,a))&&!1===(u.result=r)&&(u.preventDefault(),u.stopPropagation()));return c.postDispatch&&c.postDispatch.call(this,u),u.result}},handlers:function(e,t){var n,r,i,o,s,a=[],u=t.delegateCount,l=e.target;if(u&&l.nodeType&&!("click"===e.type&&e.button>=1))for(;l!==this;l=l.parentNode||this)if(1===l.nodeType&&("click"!==e.type||!0!==l.disabled)){for(o=[],s={},n=0;n<u;n++)void 0===s[i=(r=t[n]).selector+" "]&&(s[i]=r.needsContext?E(i,this).index(l)>-1:E.find(i,this,null,[l]).length),s[i]&&o.push(r);o.length&&a.push({elem:l,handlers:o})}return l=this,u<t.length&&a.push({elem:l,handlers:t.slice(u)}),a},addProp:function(e,t){Object.defineProperty(E.Event.prototype,e,{enumerable:!0,configurable:!0,get:m(t)?function(){if(this.originalEvent)return t(this.originalEvent)}:function(){if(this.originalEvent)return this.originalEvent[e]},set:function(t){Object.defineProperty(this,e,{enumerable:!0,configurable:!0,writable:!0,value:t})}})},fix:function(e){return e[E.expando]?e:new E.Event(e)},special:{load:{noBubble:!0},click:{setup:function(e){var t=this||e;return ge.test(t.type)&&t.click&&z(t,"input")&&De(t,"click",Te),!1},trigger:function(e){var t=this||e;return ge.test(t.type)&&t.click&&z(t,"input")&&De(t,"click"),!0},_default:function(e){var t=e.target;return ge.test(t.type)&&t.click&&z(t,"input")&&J.get(t,"click")||z(t,"a")}},beforeunload:{postDispatch:function(e){void 0!==e.result&&e.originalEvent&&(e.originalEvent.returnValue=e.result)}}}},E.removeEvent=function(e,t,n){e.removeEventListener&&e.removeEventListener(t,n)},E.Event=function(e,t){if(!(this instanceof E.Event))return new E.Event(e,t);e&&e.type?(this.originalEvent=e,this.type=e.type,this.isDefaultPrevented=e.defaultPrevented||void 0===e.defaultPrevented&&!1===e.returnValue?Te:Se,this.target=e.target&&3===e.target.nodeType?e.target.parentNode:e.target,this.currentTarget=e.currentTarget,this.relatedTarget=e.relatedTarget):this.type=e,t&&E.extend(this,t),this.timeStamp=e&&e.timeStamp||Date.now(),this[E.expando]=!0},E.Event.prototype={constructor:E.Event,isDefaultPrevented:Se,isPropagationStopped:Se,isImmediatePropagationStopped:Se,isSimulated:!1,preventDefault:function(){var e=this.originalEvent;this.isDefaultPrevented=Te,e&&!this.isSimulated&&e.preventDefault()},stopPropagation:function(){var e=this.originalEvent;this.isPropagationStopped=Te,e&&!this.isSimulated&&e.stopPropagation()},stopImmediatePropagation:function(){var e=this.originalEvent;this.isImmediatePropagationStopped=Te,e&&!this.isSimulated&&e.stopImmediatePropagation(),this.stopPropagation()}},E.each({altKey:!0,bubbles:!0,cancelable:!0,changedTouches:!0,ctrlKey:!0,detail:!0,eventPhase:!0,metaKey:!0,pageX:!0,pageY:!0,shiftKey:!0,view:!0,char:!0,code:!0,charCode:!0,key:!0,keyCode:!0,button:!0,buttons:!0,clientX:!0,clientY:!0,offsetX:!0,offsetY:!0,pointerId:!0,pointerType:!0,screenX:!0,screenY:!0,targetTouches:!0,toElement:!0,touches:!0,which:!0},E.event.addProp),E.each({focus:"focusin",blur:"focusout"},(function(e,t){E.event.special[e]={setup:function(){return De(this,e,Ce),!1},trigger:function(){return De(this,e),!0},_default:function(){return!0},delegateType:t}})),E.each({mouseenter:"mouseover",mouseleave:"mouseout",pointerenter:"pointerover",pointerleave:"pointerout"},(function(e,t){E.event.special[e]={delegateType:t,bindType:t,handle:function(e){var n,r=this,i=e.relatedTarget,o=e.handleObj;return i&&(i===r||E.contains(r,i))||(e.type=o.origType,n=o.handler.apply(this,arguments),e.type=t),n}}})),E.fn.extend({on:function(e,t,n,r){return Ae(this,e,t,n,r)},one:function(e,t,n,r){return Ae(this,e,t,n,r,1)},off:function(e,t,n){var r,i;if(e&&e.preventDefault&&e.handleObj)return r=e.handleObj,E(e.delegateTarget).off(r.namespace?r.origType+"."+r.namespace:r.origType,r.selector,r.handler),this;if("object"==typeof e){for(i in e)this.off(i,t,e[i]);return this}return!1!==t&&"function"!=typeof t||(n=t,t=void 0),!1===n&&(n=Se),this.each((function(){E.event.remove(this,e,n,t)}))}});var ze=/<script|<style|<link/i,ke=/checked\s*(?:[^=]|=\s*.checked.)/i,Ne=/^\s*<!(?:\[CDATA\[|--)|(?:\]\]|--)>\s*$/g;function Re(e,t){return z(e,"table")&&z(11!==t.nodeType?t:t.firstChild,"tr")&&E(e).children("tbody")[0]||e}function je(e){return e.type=(null!==e.getAttribute("type"))+"/"+e.type,e}function Oe(e){return"true/"===(e.type||"").slice(0,5)?e.type=e.type.slice(5):e.removeAttribute("type"),e}function Fe(e,t){var n,r,i,o,s,a;if(1===t.nodeType){if(J.hasData(e)&&(a=J.get(e).events))for(i in J.remove(t,"handle events"),a)for(n=0,r=a[i].length;n<r;n++)E.event.add(t,i,a[i][n]);Z.hasData(e)&&(o=Z.access(e),s=E.extend({},o),Z.set(t,s))}}function Le(e,t){var n=t.nodeName.toLowerCase();"input"===n&&ge.test(e.type)?t.checked=e.checked:"input"!==n&&"textarea"!==n||(t.defaultValue=e.defaultValue)}function Pe(e,t,n,r){t=u(t);var i,o,s,a,l,c,f=0,h=e.length,p=h-1,d=t[0],g=m(d);if(g||h>1&&"string"==typeof d&&!v.checkClone&&ke.test(d))return e.each((function(i){var o=e.eq(i);g&&(t[0]=d.call(this,i,o.html())),Pe(o,t,n,r)}));if(h&&(o=(i=xe(t,e[0].ownerDocument,!1,e,r)).firstChild,1===i.childNodes.length&&(i=o),o||r)){for(a=(s=E.map(_e(i,"script"),je)).length;f<h;f++)l=i,f!==p&&(l=E.clone(l,!0,!0),a&&E.merge(s,_e(l,"script"))),n.call(e[f],l,f);if(a)for(c=s[s.length-1].ownerDocument,E.map(s,Oe),f=0;f<a;f++)l=s[f],me.test(l.type||"")&&!J.access(l,"globalEval")&&E.contains(c,l)&&(l.src&&"module"!==(l.type||"").toLowerCase()?E._evalUrl&&!l.noModule&&E._evalUrl(l.src,{nonce:l.nonce||l.getAttribute("nonce")},c):w(l.textContent.replace(Ne,""),l,c))}return e}function qe(e,t,n){for(var r,i=t?E.filter(t,e):e,o=0;null!=(r=i[o]);o++)n||1!==r.nodeType||E.cleanData(_e(r)),r.parentNode&&(n&&se(r)&&be(_e(r,"script")),r.parentNode.removeChild(r));return e}E.extend({htmlPrefilter:function(e){return e},clone:function(e,t,n){var r,i,o,s,a=e.cloneNode(!0),u=se(e);if(!(v.noCloneChecked||1!==e.nodeType&&11!==e.nodeType||E.isXMLDoc(e)))for(s=_e(a),r=0,i=(o=_e(e)).length;r<i;r++)Le(o[r],s[r]);if(t)if(n)for(o=o||_e(e),s=s||_e(a),r=0,i=o.length;r<i;r++)Fe(o[r],s[r]);else Fe(e,a);return(s=_e(a,"script")).length>0&&be(s,!u&&_e(e,"script")),a},cleanData:function(e){for(var t,n,r,i=E.event.special,o=0;void 0!==(n=e[o]);o++)if(K(n)){if(t=n[J.expando]){if(t.events)for(r in t.events)i[r]?E.event.remove(n,r):E.removeEvent(n,r,t.handle);n[J.expando]=void 0}n[Z.expando]&&(n[Z.expando]=void 0)}}}),E.fn.extend({detach:function(e){return qe(this,e,!0)},remove:function(e){return qe(this,e)},text:function(e){return U(this,(function(e){return void 0===e?E.text(this):this.empty().each((function(){1!==this.nodeType&&11!==this.nodeType&&9!==this.nodeType||(this.textContent=e)}))}),null,e,arguments.length)},append:function(){return Pe(this,arguments,(function(e){1!==this.nodeType&&11!==this.nodeType&&9!==this.nodeType||Re(this,e).appendChild(e)}))},prepend:function(){return Pe(this,arguments,(function(e){if(1===this.nodeType||11===this.nodeType||9===this.nodeType){var t=Re(this,e);t.insertBefore(e,t.firstChild)}}))},before:function(){return Pe(this,arguments,(function(e){this.parentNode&&this.parentNode.insertBefore(e,this)}))},after:function(){return Pe(this,arguments,(function(e){this.parentNode&&this.parentNode.insertBefore(e,this.nextSibling)}))},empty:function(){for(var e,t=0;null!=(e=this[t]);t++)1===e.nodeType&&(E.cleanData(_e(e,!1)),e.textContent="");return this},clone:function(e,t){return e=null!=e&&e,t=null==t?e:t,this.map((function(){return E.clone(this,e,t)}))},html:function(e){return U(this,(function(e){var t=this[0]||{},n=0,r=this.length;if(void 0===e&&1===t.nodeType)return t.innerHTML;if("string"==typeof e&&!ze.test(e)&&!ye[(ve.exec(e)||["",""])[1].toLowerCase()]){e=E.htmlPrefilter(e);try{for(;n<r;n++)1===(t=this[n]||{}).nodeType&&(E.cleanData(_e(t,!1)),t.innerHTML=e);t=0}catch(e){}}t&&this.empty().append(e)}),null,e,arguments.length)},replaceWith:function(){var e=[];return Pe(this,arguments,(function(t){var n=this.parentNode;E.inArray(this,e)<0&&(E.cleanData(_e(this)),n&&n.replaceChild(t,this))}),e)}}),E.each({appendTo:"append",prependTo:"prepend",insertBefore:"before",insertAfter:"after",replaceAll:"replaceWith"},(function(e,t){E.fn[e]=function(e){for(var n,r=[],i=E(e),o=i.length-1,s=0;s<=o;s++)n=s===o?this:this.clone(!0),E(i[s])[t](n),l.apply(r,n.get());return this.pushStack(r)}}));var Ie=new RegExp("^("+ne+")(?!px)[a-z%]+$","i"),Me=function(e){var t=e.ownerDocument.defaultView;return t&&t.opener||(t=n),t.getComputedStyle(e)},We=function(e,t,n){var r,i,o={};for(i in t)o[i]=e.style[i],e.style[i]=t[i];for(i in r=n.call(e),t)e.style[i]=o[i];return r},Be=new RegExp(ie.join("|"),"i");function He(e,t,n){var r,i,o,s,a=e.style;return(n=n||Me(e))&&(""!==(s=n.getPropertyValue(t)||n[t])||se(e)||(s=E.style(e,t)),!v.pixelBoxStyles()&&Ie.test(s)&&Be.test(t)&&(r=a.width,i=a.minWidth,o=a.maxWidth,a.minWidth=a.maxWidth=a.width=s,s=n.width,a.width=r,a.minWidth=i,a.maxWidth=o)),void 0!==s?s+"":s}function Ue(e,t){return{get:function(){if(!e())return(this.get=t).apply(this,arguments);delete this.get}}}!function(){function e(){if(c){l.style.cssText="position:absolute;left:-11111px;width:60px;margin-top:1px;padding:0;border:0",c.style.cssText="position:relative;display:block;box-sizing:border-box;overflow:scroll;margin:auto;border:1px;padding:1px;width:60%;top:1%",oe.appendChild(l).appendChild(c);var e=n.getComputedStyle(c);r="1%"!==e.top,u=12===t(e.marginLeft),c.style.right="60%",s=36===t(e.right),i=36===t(e.width),c.style.position="absolute",o=12===t(c.offsetWidth/3),oe.removeChild(l),c=null}}function t(e){return Math.round(parseFloat(e))}var r,i,o,s,a,u,l=_.createElement("div"),c=_.createElement("div");c.style&&(c.style.backgroundClip="content-box",c.cloneNode(!0).style.backgroundClip="",v.clearCloneStyle="content-box"===c.style.backgroundClip,E.extend(v,{boxSizingReliable:function(){return e(),i},pixelBoxStyles:function(){return e(),s},pixelPosition:function(){return e(),r},reliableMarginLeft:function(){return e(),u},scrollboxSize:function(){return e(),o},reliableTrDimensions:function(){var e,t,r,i;return null==a&&(e=_.createElement("table"),t=_.createElement("tr"),r=_.createElement("div"),e.style.cssText="position:absolute;left:-11111px;border-collapse:separate",t.style.cssText="border:1px solid",t.style.height="1px",r.style.height="9px",r.style.display="block",oe.appendChild(e).appendChild(t).appendChild(r),i=n.getComputedStyle(t),a=parseInt(i.height,10)+parseInt(i.borderTopWidth,10)+parseInt(i.borderBottomWidth,10)===t.offsetHeight,oe.removeChild(e)),a}}))}();var $e=["Webkit","Moz","ms"],Xe=_.createElement("div").style,Ve={};function Ge(e){var t=E.cssProps[e]||Ve[e];return t||(e in Xe?e:Ve[e]=function(e){for(var t=e[0].toUpperCase()+e.slice(1),n=$e.length;n--;)if((e=$e[n]+t)in Xe)return e}(e)||e)}var Ke=/^(none|table(?!-c[ea]).+)/,Ye=/^--/,Je={position:"absolute",visibility:"hidden",display:"block"},Ze={letterSpacing:"0",fontWeight:"400"};function Qe(e,t,n){var r=re.exec(t);return r?Math.max(0,r[2]-(n||0))+(r[3]||"px"):t}function et(e,t,n,r,i,o){var s="width"===t?1:0,a=0,u=0;if(n===(r?"border":"content"))return 0;for(;s<4;s+=2)"margin"===n&&(u+=E.css(e,n+ie[s],!0,i)),r?("content"===n&&(u-=E.css(e,"padding"+ie[s],!0,i)),"margin"!==n&&(u-=E.css(e,"border"+ie[s]+"Width",!0,i))):(u+=E.css(e,"padding"+ie[s],!0,i),"padding"!==n?u+=E.css(e,"border"+ie[s]+"Width",!0,i):a+=E.css(e,"border"+ie[s]+"Width",!0,i));return!r&&o>=0&&(u+=Math.max(0,Math.ceil(e["offset"+t[0].toUpperCase()+t.slice(1)]-o-u-a-.5))||0),u}function tt(e,t,n){var r=Me(e),i=(!v.boxSizingReliable()||n)&&"border-box"===E.css(e,"boxSizing",!1,r),o=i,s=He(e,t,r),a="offset"+t[0].toUpperCase()+t.slice(1);if(Ie.test(s)){if(!n)return s;s="auto"}return(!v.boxSizingReliable()&&i||!v.reliableTrDimensions()&&z(e,"tr")||"auto"===s||!parseFloat(s)&&"inline"===E.css(e,"display",!1,r))&&e.getClientRects().length&&(i="border-box"===E.css(e,"boxSizing",!1,r),(o=a in e)&&(s=e[a])),(s=parseFloat(s)||0)+et(e,t,n||(i?"border":"content"),o,r,s)+"px"}function nt(e,t,n,r,i){return new nt.prototype.init(e,t,n,r,i)}E.extend({cssHooks:{opacity:{get:function(e,t){if(t){var n=He(e,"opacity");return""===n?"1":n}}}},cssNumber:{animationIterationCount:!0,columnCount:!0,fillOpacity:!0,flexGrow:!0,flexShrink:!0,fontWeight:!0,gridArea:!0,gridColumn:!0,gridColumnEnd:!0,gridColumnStart:!0,gridRow:!0,gridRowEnd:!0,gridRowStart:!0,lineHeight:!0,opacity:!0,order:!0,orphans:!0,widows:!0,zIndex:!0,zoom:!0},cssProps:{},style:function(e,t,n,r){if(e&&3!==e.nodeType&&8!==e.nodeType&&e.style){var i,o,s,a=G(t),u=Ye.test(t),l=e.style;if(u||(t=Ge(a)),s=E.cssHooks[t]||E.cssHooks[a],void 0===n)return s&&"get"in s&&void 0!==(i=s.get(e,!1,r))?i:l[t];"string"===(o=typeof n)&&(i=re.exec(n))&&i[1]&&(n=le(e,t,i),o="number"),null!=n&&n==n&&("number"!==o||u||(n+=i&&i[3]||(E.cssNumber[a]?"":"px")),v.clearCloneStyle||""!==n||0!==t.indexOf("background")||(l[t]="inherit"),s&&"set"in s&&void 0===(n=s.set(e,n,r))||(u?l.setProperty(t,n):l[t]=n))}},css:function(e,t,n,r){var i,o,s,a=G(t);return Ye.test(t)||(t=Ge(a)),(s=E.cssHooks[t]||E.cssHooks[a])&&"get"in s&&(i=s.get(e,!0,n)),void 0===i&&(i=He(e,t,r)),"normal"===i&&t in Ze&&(i=Ze[t]),""===n||n?(o=parseFloat(i),!0===n||isFinite(o)?o||0:i):i}}),E.each(["height","width"],(function(e,t){E.cssHooks[t]={get:function(e,n,r){if(n)return!Ke.test(E.css(e,"display"))||e.getClientRects().length&&e.getBoundingClientRect().width?tt(e,t,r):We(e,Je,(function(){return tt(e,t,r)}))},set:function(e,n,r){var i,o=Me(e),s=!v.scrollboxSize()&&"absolute"===o.position,a=(s||r)&&"border-box"===E.css(e,"boxSizing",!1,o),u=r?et(e,t,r,a,o):0;return a&&s&&(u-=Math.ceil(e["offset"+t[0].toUpperCase()+t.slice(1)]-parseFloat(o[t])-et(e,t,"border",!1,o)-.5)),u&&(i=re.exec(n))&&"px"!==(i[3]||"px")&&(e.style[t]=n,n=E.css(e,t)),Qe(0,n,u)}}})),E.cssHooks.marginLeft=Ue(v.reliableMarginLeft,(function(e,t){if(t)return(parseFloat(He(e,"marginLeft"))||e.getBoundingClientRect().left-We(e,{marginLeft:0},(function(){return e.getBoundingClientRect().left})))+"px"})),E.each({margin:"",padding:"",border:"Width"},(function(e,t){E.cssHooks[e+t]={expand:function(n){for(var r=0,i={},o="string"==typeof n?n.split(" "):[n];r<4;r++)i[e+ie[r]+t]=o[r]||o[r-2]||o[0];return i}},"margin"!==e&&(E.cssHooks[e+t].set=Qe)})),E.fn.extend({css:function(e,t){return U(this,(function(e,t,n){var r,i,o={},s=0;if(Array.isArray(t)){for(r=Me(e),i=t.length;s<i;s++)o[t[s]]=E.css(e,t[s],!1,r);return o}return void 0!==n?E.style(e,t,n):E.css(e,t)}),e,t,arguments.length>1)}}),E.Tween=nt,nt.prototype={constructor:nt,init:function(e,t,n,r,i,o){this.elem=e,this.prop=n,this.easing=i||E.easing._default,this.options=t,this.start=this.now=this.cur(),this.end=r,this.unit=o||(E.cssNumber[n]?"":"px")},cur:function(){var e=nt.propHooks[this.prop];return e&&e.get?e.get(this):nt.propHooks._default.get(this)},run:function(e){var t,n=nt.propHooks[this.prop];return this.options.duration?this.pos=t=E.easing[this.easing](e,this.options.duration*e,0,1,this.options.duration):this.pos=t=e,this.now=(this.end-this.start)*t+this.start,this.options.step&&this.options.step.call(this.elem,this.now,this),n&&n.set?n.set(this):nt.propHooks._default.set(this),this}},nt.prototype.init.prototype=nt.prototype,nt.propHooks={_default:{get:function(e){var t;return 1!==e.elem.nodeType||null!=e.elem[e.prop]&&null==e.elem.style[e.prop]?e.elem[e.prop]:(t=E.css(e.elem,e.prop,""))&&"auto"!==t?t:0},set:function(e){E.fx.step[e.prop]?E.fx.step[e.prop](e):1!==e.elem.nodeType||!E.cssHooks[e.prop]&&null==e.elem.style[Ge(e.prop)]?e.elem[e.prop]=e.now:E.style(e.elem,e.prop,e.now+e.unit)}}},nt.propHooks.scrollTop=nt.propHooks.scrollLeft={set:function(e){e.elem.nodeType&&e.elem.parentNode&&(e.elem[e.prop]=e.now)}},E.easing={linear:function(e){return e},swing:function(e){return.5-Math.cos(e*Math.PI)/2},_default:"swing"},E.fx=nt.prototype.init,E.fx.step={};var rt,it,ot=/^(?:toggle|show|hide)$/,st=/queueHooks$/;function at(){it&&(!1===_.hidden&&n.requestAnimationFrame?n.requestAnimationFrame(at):n.setTimeout(at,E.fx.interval),E.fx.tick())}function ut(){return n.setTimeout((function(){rt=void 0})),rt=Date.now()}function lt(e,t){var n,r=0,i={height:e};for(t=t?1:0;r<4;r+=2-t)i["margin"+(n=ie[r])]=i["padding"+n]=e;return t&&(i.opacity=i.width=e),i}function ct(e,t,n){for(var r,i=(ft.tweeners[t]||[]).concat(ft.tweeners["*"]),o=0,s=i.length;o<s;o++)if(r=i[o].call(n,t,e))return r}function ft(e,t,n){var r,i,o=0,s=ft.prefilters.length,a=E.Deferred().always((function(){delete u.elem})),u=function(){if(i)return!1;for(var t=rt||ut(),n=Math.max(0,l.startTime+l.duration-t),r=1-(n/l.duration||0),o=0,s=l.tweens.length;o<s;o++)l.tweens[o].run(r);return a.notifyWith(e,[l,r,n]),r<1&&s?n:(s||a.notifyWith(e,[l,1,0]),a.resolveWith(e,[l]),!1)},l=a.promise({elem:e,props:E.extend({},t),opts:E.extend(!0,{specialEasing:{},easing:E.easing._default},n),originalProperties:t,originalOptions:n,startTime:rt||ut(),duration:n.duration,tweens:[],createTween:function(t,n){var r=E.Tween(e,l.opts,t,n,l.opts.specialEasing[t]||l.opts.easing);return l.tweens.push(r),r},stop:function(t){var n=0,r=t?l.tweens.length:0;if(i)return this;for(i=!0;n<r;n++)l.tweens[n].run(1);return t?(a.notifyWith(e,[l,1,0]),a.resolveWith(e,[l,t])):a.rejectWith(e,[l,t]),this}}),c=l.props;for(!function(e,t){var n,r,i,o,s;for(n in e)if(i=t[r=G(n)],o=e[n],Array.isArray(o)&&(i=o[1],o=e[n]=o[0]),n!==r&&(e[r]=o,delete e[n]),(s=E.cssHooks[r])&&"expand"in s)for(n in o=s.expand(o),delete e[r],o)n in e||(e[n]=o[n],t[n]=i);else t[r]=i}(c,l.opts.specialEasing);o<s;o++)if(r=ft.prefilters[o].call(l,e,c,l.opts))return m(r.stop)&&(E._queueHooks(l.elem,l.opts.queue).stop=r.stop.bind(r)),r;return E.map(c,ct,l),m(l.opts.start)&&l.opts.start.call(e,l),l.progress(l.opts.progress).done(l.opts.done,l.opts.complete).fail(l.opts.fail).always(l.opts.always),E.fx.timer(E.extend(u,{elem:e,anim:l,queue:l.opts.queue})),l}E.Animation=E.extend(ft,{tweeners:{"*":[function(e,t){var n=this.createTween(e,t);return le(n.elem,e,re.exec(t),n),n}]},tweener:function(e,t){m(e)?(t=e,e=["*"]):e=e.match(P);for(var n,r=0,i=e.length;r<i;r++)n=e[r],ft.tweeners[n]=ft.tweeners[n]||[],ft.tweeners[n].unshift(t)},prefilters:[function(e,t,n){var r,i,o,s,a,u,l,c,f="width"in t||"height"in t,h=this,p={},d=e.style,g=e.nodeType&&ue(e),v=J.get(e,"fxshow");for(r in n.queue||(null==(s=E._queueHooks(e,"fx")).unqueued&&(s.unqueued=0,a=s.empty.fire,s.empty.fire=function(){s.unqueued||a()}),s.unqueued++,h.always((function(){h.always((function(){s.unqueued--,E.queue(e,"fx").length||s.empty.fire()}))}))),t)if(i=t[r],ot.test(i)){if(delete t[r],o=o||"toggle"===i,i===(g?"hide":"show")){if("show"!==i||!v||void 0===v[r])continue;g=!0}p[r]=v&&v[r]||E.style(e,r)}if((u=!E.isEmptyObject(t))||!E.isEmptyObject(p))for(r in f&&1===e.nodeType&&(n.overflow=[d.overflow,d.overflowX,d.overflowY],null==(l=v&&v.display)&&(l=J.get(e,"display")),"none"===(c=E.css(e,"display"))&&(l?c=l:(he([e],!0),l=e.style.display||l,c=E.css(e,"display"),he([e]))),("inline"===c||"inline-block"===c&&null!=l)&&"none"===E.css(e,"float")&&(u||(h.done((function(){d.display=l})),null==l&&(c=d.display,l="none"===c?"":c)),d.display="inline-block")),n.overflow&&(d.overflow="hidden",h.always((function(){d.overflow=n.overflow[0],d.overflowX=n.overflow[1],d.overflowY=n.overflow[2]}))),u=!1,p)u||(v?"hidden"in v&&(g=v.hidden):v=J.access(e,"fxshow",{display:l}),o&&(v.hidden=!g),g&&he([e],!0),h.done((function(){for(r in g||he([e]),J.remove(e,"fxshow"),p)E.style(e,r,p[r])}))),u=ct(g?v[r]:0,r,h),r in v||(v[r]=u.start,g&&(u.end=u.start,u.start=0))}],prefilter:function(e,t){t?ft.prefilters.unshift(e):ft.prefilters.push(e)}}),E.speed=function(e,t,n){var r=e&&"object"==typeof e?E.extend({},e):{complete:n||!n&&t||m(e)&&e,duration:e,easing:n&&t||t&&!m(t)&&t};return E.fx.off?r.duration=0:"number"!=typeof r.duration&&(r.duration in E.fx.speeds?r.duration=E.fx.speeds[r.duration]:r.duration=E.fx.speeds._default),null!=r.queue&&!0!==r.queue||(r.queue="fx"),r.old=r.complete,r.complete=function(){m(r.old)&&r.old.call(this),r.queue&&E.dequeue(this,r.queue)},r},E.fn.extend({fadeTo:function(e,t,n,r){return this.filter(ue).css("opacity",0).show().end().animate({opacity:t},e,n,r)},animate:function(e,t,n,r){var i=E.isEmptyObject(e),o=E.speed(t,n,r),s=function(){var t=ft(this,E.extend({},e),o);(i||J.get(this,"finish"))&&t.stop(!0)};return s.finish=s,i||!1===o.queue?this.each(s):this.queue(o.queue,s)},stop:function(e,t,n){var r=function(e){var t=e.stop;delete e.stop,t(n)};return"string"!=typeof e&&(n=t,t=e,e=void 0),t&&this.queue(e||"fx",[]),this.each((function(){var t=!0,i=null!=e&&e+"queueHooks",o=E.timers,s=J.get(this);if(i)s[i]&&s[i].stop&&r(s[i]);else for(i in s)s[i]&&s[i].stop&&st.test(i)&&r(s[i]);for(i=o.length;i--;)o[i].elem!==this||null!=e&&o[i].queue!==e||(o[i].anim.stop(n),t=!1,o.splice(i,1));!t&&n||E.dequeue(this,e)}))},finish:function(e){return!1!==e&&(e=e||"fx"),this.each((function(){var t,n=J.get(this),r=n[e+"queue"],i=n[e+"queueHooks"],o=E.timers,s=r?r.length:0;for(n.finish=!0,E.queue(this,e,[]),i&&i.stop&&i.stop.call(this,!0),t=o.length;t--;)o[t].elem===this&&o[t].queue===e&&(o[t].anim.stop(!0),o.splice(t,1));for(t=0;t<s;t++)r[t]&&r[t].finish&&r[t].finish.call(this);delete n.finish}))}}),E.each(["toggle","show","hide"],(function(e,t){var n=E.fn[t];E.fn[t]=function(e,r,i){return null==e||"boolean"==typeof e?n.apply(this,arguments):this.animate(lt(t,!0),e,r,i)}})),E.each({slideDown:lt("show"),slideUp:lt("hide"),slideToggle:lt("toggle"),fadeIn:{opacity:"show"},fadeOut:{opacity:"hide"},fadeToggle:{opacity:"toggle"}},(function(e,t){E.fn[e]=function(e,n,r){return this.animate(t,e,n,r)}})),E.timers=[],E.fx.tick=function(){var e,t=0,n=E.timers;for(rt=Date.now();t<n.length;t++)(e=n[t])()||n[t]!==e||n.splice(t--,1);n.length||E.fx.stop(),rt=void 0},E.fx.timer=function(e){E.timers.push(e),E.fx.start()},E.fx.interval=13,E.fx.start=function(){it||(it=!0,at())},E.fx.stop=function(){it=null},E.fx.speeds={slow:600,fast:200,_default:400},E.fn.delay=function(e,t){return e=E.fx&&E.fx.speeds[e]||e,t=t||"fx",this.queue(t,(function(t,r){var i=n.setTimeout(t,e);r.stop=function(){n.clearTimeout(i)}}))},function(){var e=_.createElement("input"),t=_.createElement("select").appendChild(_.createElement("option"));e.type="checkbox",v.checkOn=""!==e.value,v.optSelected=t.selected,(e=_.createElement("input")).value="t",e.type="radio",v.radioValue="t"===e.value}();var ht,pt=E.expr.attrHandle;E.fn.extend({attr:function(e,t){return U(this,E.attr,e,t,arguments.length>1)},removeAttr:function(e){return this.each((function(){E.removeAttr(this,e)}))}}),E.extend({attr:function(e,t,n){var r,i,o=e.nodeType;if(3!==o&&8!==o&&2!==o)return void 0===e.getAttribute?E.prop(e,t,n):(1===o&&E.isXMLDoc(e)||(i=E.attrHooks[t.toLowerCase()]||(E.expr.match.bool.test(t)?ht:void 0)),void 0!==n?null===n?void E.removeAttr(e,t):i&&"set"in i&&void 0!==(r=i.set(e,n,t))?r:(e.setAttribute(t,n+""),n):i&&"get"in i&&null!==(r=i.get(e,t))?r:null==(r=E.find.attr(e,t))?void 0:r)},attrHooks:{type:{set:function(e,t){if(!v.radioValue&&"radio"===t&&z(e,"input")){var n=e.value;return e.setAttribute("type",t),n&&(e.value=n),t}}}},removeAttr:function(e,t){var n,r=0,i=t&&t.match(P);if(i&&1===e.nodeType)for(;n=i[r++];)e.removeAttribute(n)}}),ht={set:function(e,t,n){return!1===t?E.removeAttr(e,n):e.setAttribute(n,n),n}},E.each(E.expr.match.bool.source.match(/\w+/g),(function(e,t){var n=pt[t]||E.find.attr;pt[t]=function(e,t,r){var i,o,s=t.toLowerCase();return r||(o=pt[s],pt[s]=i,i=null!=n(e,t,r)?s:null,pt[s]=o),i}}));var dt=/^(?:input|select|textarea|button)$/i,gt=/^(?:a|area)$/i;function vt(e){return(e.match(P)||[]).join(" ")}function mt(e){return e.getAttribute&&e.getAttribute("class")||""}function yt(e){return Array.isArray(e)?e:"string"==typeof e&&e.match(P)||[]}E.fn.extend({prop:function(e,t){return U(this,E.prop,e,t,arguments.length>1)},removeProp:function(e){return this.each((function(){delete this[E.propFix[e]||e]}))}}),E.extend({prop:function(e,t,n){var r,i,o=e.nodeType;if(3!==o&&8!==o&&2!==o)return 1===o&&E.isXMLDoc(e)||(t=E.propFix[t]||t,i=E.propHooks[t]),void 0!==n?i&&"set"in i&&void 0!==(r=i.set(e,n,t))?r:e[t]=n:i&&"get"in i&&null!==(r=i.get(e,t))?r:e[t]},propHooks:{tabIndex:{get:function(e){var t=E.find.attr(e,"tabindex");return t?parseInt(t,10):dt.test(e.nodeName)||gt.test(e.nodeName)&&e.href?0:-1}}},propFix:{for:"htmlFor",class:"className"}}),v.optSelected||(E.propHooks.selected={get:function(e){var t=e.parentNode;return t&&t.parentNode&&t.parentNode.selectedIndex,null},set:function(e){var t=e.parentNode;t&&(t.selectedIndex,t.parentNode&&t.parentNode.selectedIndex)}}),E.each(["tabIndex","readOnly","maxLength","cellSpacing","cellPadding","rowSpan","colSpan","useMap","frameBorder","contentEditable"],(function(){E.propFix[this.toLowerCase()]=this})),E.fn.extend({addClass:function(e){var t,n,r,i,o,s,a,u=0;if(m(e))return this.each((function(t){E(this).addClass(e.call(this,t,mt(this)))}));if((t=yt(e)).length)for(;n=this[u++];)if(i=mt(n),r=1===n.nodeType&&" "+vt(i)+" "){for(s=0;o=t[s++];)r.indexOf(" "+o+" ")<0&&(r+=o+" ");i!==(a=vt(r))&&n.setAttribute("class",a)}return this},removeClass:function(e){var t,n,r,i,o,s,a,u=0;if(m(e))return this.each((function(t){E(this).removeClass(e.call(this,t,mt(this)))}));if(!arguments.length)return this.attr("class","");if((t=yt(e)).length)for(;n=this[u++];)if(i=mt(n),r=1===n.nodeType&&" "+vt(i)+" "){for(s=0;o=t[s++];)for(;r.indexOf(" "+o+" ")>-1;)r=r.replace(" "+o+" "," ");i!==(a=vt(r))&&n.setAttribute("class",a)}return this},toggleClass:function(e,t){var n=typeof e,r="string"===n||Array.isArray(e);return"boolean"==typeof t&&r?t?this.addClass(e):this.removeClass(e):m(e)?this.each((function(n){E(this).toggleClass(e.call(this,n,mt(this),t),t)})):this.each((function(){var t,i,o,s;if(r)for(i=0,o=E(this),s=yt(e);t=s[i++];)o.hasClass(t)?o.removeClass(t):o.addClass(t);else void 0!==e&&"boolean"!==n||((t=mt(this))&&J.set(this,"__className__",t),this.setAttribute&&this.setAttribute("class",t||!1===e?"":J.get(this,"__className__")||""))}))},hasClass:function(e){var t,n,r=0;for(t=" "+e+" ";n=this[r++];)if(1===n.nodeType&&(" "+vt(mt(n))+" ").indexOf(t)>-1)return!0;return!1}});var _t=/\r/g;E.fn.extend({val:function(e){var t,n,r,i=this[0];return arguments.length?(r=m(e),this.each((function(n){var i;1===this.nodeType&&(null==(i=r?e.call(this,n,E(this).val()):e)?i="":"number"==typeof i?i+="":Array.isArray(i)&&(i=E.map(i,(function(e){return null==e?"":e+""}))),(t=E.valHooks[this.type]||E.valHooks[this.nodeName.toLowerCase()])&&"set"in t&&void 0!==t.set(this,i,"value")||(this.value=i))}))):i?(t=E.valHooks[i.type]||E.valHooks[i.nodeName.toLowerCase()])&&"get"in t&&void 0!==(n=t.get(i,"value"))?n:"string"==typeof(n=i.value)?n.replace(_t,""):null==n?"":n:void 0}}),E.extend({valHooks:{option:{get:function(e){var t=E.find.attr(e,"value");return null!=t?t:vt(E.text(e))}},select:{get:function(e){var t,n,r,i=e.options,o=e.selectedIndex,s="select-one"===e.type,a=s?null:[],u=s?o+1:i.length;for(r=o<0?u:s?o:0;r<u;r++)if(((n=i[r]).selected||r===o)&&!n.disabled&&(!n.parentNode.disabled||!z(n.parentNode,"optgroup"))){if(t=E(n).val(),s)return t;a.push(t)}return a},set:function(e,t){for(var n,r,i=e.options,o=E.makeArray(t),s=i.length;s--;)((r=i[s]).selected=E.inArray(E.valHooks.option.get(r),o)>-1)&&(n=!0);return n||(e.selectedIndex=-1),o}}}}),E.each(["radio","checkbox"],(function(){E.valHooks[this]={set:function(e,t){if(Array.isArray(t))return e.checked=E.inArray(E(e).val(),t)>-1}},v.checkOn||(E.valHooks[this].get=function(e){return null===e.getAttribute("value")?"on":e.value})})),v.focusin="onfocusin"in n;var bt=/^(?:focusinfocus|focusoutblur)$/,wt=function(e){e.stopPropagation()};E.extend(E.event,{trigger:function(e,t,r,i){var o,s,a,u,l,c,f,h,d=[r||_],g=p.call(e,"type")?e.type:e,v=p.call(e,"namespace")?e.namespace.split("."):[];if(s=h=a=r=r||_,3!==r.nodeType&&8!==r.nodeType&&!bt.test(g+E.event.triggered)&&(g.indexOf(".")>-1&&(v=g.split("."),g=v.shift(),v.sort()),l=g.indexOf(":")<0&&"on"+g,(e=e[E.expando]?e:new E.Event(g,"object"==typeof e&&e)).isTrigger=i?2:3,e.namespace=v.join("."),e.rnamespace=e.namespace?new RegExp("(^|\\.)"+v.join("\\.(?:.*\\.|)")+"(\\.|$)"):null,e.result=void 0,e.target||(e.target=r),t=null==t?[e]:E.makeArray(t,[e]),f=E.event.special[g]||{},i||!f.trigger||!1!==f.trigger.apply(r,t))){if(!i&&!f.noBubble&&!y(r)){for(u=f.delegateType||g,bt.test(u+g)||(s=s.parentNode);s;s=s.parentNode)d.push(s),a=s;a===(r.ownerDocument||_)&&d.push(a.defaultView||a.parentWindow||n)}for(o=0;(s=d[o++])&&!e.isPropagationStopped();)h=s,e.type=o>1?u:f.bindType||g,(c=(J.get(s,"events")||Object.create(null))[e.type]&&J.get(s,"handle"))&&c.apply(s,t),(c=l&&s[l])&&c.apply&&K(s)&&(e.result=c.apply(s,t),!1===e.result&&e.preventDefault());return e.type=g,i||e.isDefaultPrevented()||f._default&&!1!==f._default.apply(d.pop(),t)||!K(r)||l&&m(r[g])&&!y(r)&&((a=r[l])&&(r[l]=null),E.event.triggered=g,e.isPropagationStopped()&&h.addEventListener(g,wt),r[g](),e.isPropagationStopped()&&h.removeEventListener(g,wt),E.event.triggered=void 0,a&&(r[l]=a)),e.result}},simulate:function(e,t,n){var r=E.extend(new E.Event,n,{type:e,isSimulated:!0});E.event.trigger(r,null,t)}}),E.fn.extend({trigger:function(e,t){return this.each((function(){E.event.trigger(e,t,this)}))},triggerHandler:function(e,t){var n=this[0];if(n)return E.event.trigger(e,t,n,!0)}}),v.focusin||E.each({focus:"focusin",blur:"focusout"},(function(e,t){var n=function(e){E.event.simulate(t,e.target,E.event.fix(e))};E.event.special[t]={setup:function(){var r=this.ownerDocument||this.document||this,i=J.access(r,t);i||r.addEventListener(e,n,!0),J.access(r,t,(i||0)+1)},teardown:function(){var r=this.ownerDocument||this.document||this,i=J.access(r,t)-1;i?J.access(r,t,i):(r.removeEventListener(e,n,!0),J.remove(r,t))}}}));var xt=n.location,Et={guid:Date.now()},Tt=/\?/;E.parseXML=function(e){var t,r;if(!e||"string"!=typeof e)return null;try{t=(new n.DOMParser).parseFromString(e,"text/xml")}catch(e){}return r=t&&t.getElementsByTagName("parsererror")[0],t&&!r||E.error("Invalid XML: "+(r?E.map(r.childNodes,(function(e){return e.textContent})).join("\n"):e)),t};var St=/\[\]$/,Ct=/\r?\n/g,At=/^(?:submit|button|image|reset|file)$/i,Dt=/^(?:input|select|textarea|keygen)/i;function zt(e,t,n,r){var i;if(Array.isArray(t))E.each(t,(function(t,i){n||St.test(e)?r(e,i):zt(e+"["+("object"==typeof i&&null!=i?t:"")+"]",i,n,r)}));else if(n||"object"!==x(t))r(e,t);else for(i in t)zt(e+"["+i+"]",t[i],n,r)}E.param=function(e,t){var n,r=[],i=function(e,t){var n=m(t)?t():t;r[r.length]=encodeURIComponent(e)+"="+encodeURIComponent(null==n?"":n)};if(null==e)return"";if(Array.isArray(e)||e.jquery&&!E.isPlainObject(e))E.each(e,(function(){i(this.name,this.value)}));else for(n in e)zt(n,e[n],t,i);return r.join("&")},E.fn.extend({serialize:function(){return E.param(this.serializeArray())},serializeArray:function(){return this.map((function(){var e=E.prop(this,"elements");return e?E.makeArray(e):this})).filter((function(){var e=this.type;return this.name&&!E(this).is(":disabled")&&Dt.test(this.nodeName)&&!At.test(e)&&(this.checked||!ge.test(e))})).map((function(e,t){var n=E(this).val();return null==n?null:Array.isArray(n)?E.map(n,(function(e){return{name:t.name,value:e.replace(Ct,"\r\n")}})):{name:t.name,value:n.replace(Ct,"\r\n")}})).get()}});var kt=/%20/g,Nt=/#.*$/,Rt=/([?&])_=[^&]*/,jt=/^(.*?):[ \t]*([^\r\n]*)$/gm,Ot=/^(?:GET|HEAD)$/,Ft=/^\/\//,Lt={},Pt={},qt="*/".concat("*"),It=_.createElement("a");function Mt(e){return function(t,n){"string"!=typeof t&&(n=t,t="*");var r,i=0,o=t.toLowerCase().match(P)||[];if(m(n))for(;r=o[i++];)"+"===r[0]?(r=r.slice(1)||"*",(e[r]=e[r]||[]).unshift(n)):(e[r]=e[r]||[]).push(n)}}function Wt(e,t,n,r){var i={},o=e===Pt;function s(a){var u;return i[a]=!0,E.each(e[a]||[],(function(e,a){var l=a(t,n,r);return"string"!=typeof l||o||i[l]?o?!(u=l):void 0:(t.dataTypes.unshift(l),s(l),!1)})),u}return s(t.dataTypes[0])||!i["*"]&&s("*")}function Bt(e,t){var n,r,i=E.ajaxSettings.flatOptions||{};for(n in t)void 0!==t[n]&&((i[n]?e:r||(r={}))[n]=t[n]);return r&&E.extend(!0,e,r),e}It.href=xt.href,E.extend({active:0,lastModified:{},etag:{},ajaxSettings:{url:xt.href,type:"GET",isLocal:/^(?:about|app|app-storage|.+-extension|file|res|widget):$/.test(xt.protocol),global:!0,processData:!0,async:!0,contentType:"application/x-www-form-urlencoded; charset=UTF-8",accepts:{"*":qt,text:"text/plain",html:"text/html",xml:"application/xml, text/xml",json:"application/json, text/javascript"},contents:{xml:/\bxml\b/,html:/\bhtml/,json:/\bjson\b/},responseFields:{xml:"responseXML",text:"responseText",json:"responseJSON"},converters:{"* text":String,"text html":!0,"text json":JSON.parse,"text xml":E.parseXML},flatOptions:{url:!0,context:!0}},ajaxSetup:function(e,t){return t?Bt(Bt(e,E.ajaxSettings),t):Bt(E.ajaxSettings,e)},ajaxPrefilter:Mt(Lt),ajaxTransport:Mt(Pt),ajax:function(e,t){"object"==typeof e&&(t=e,e=void 0),t=t||{};var r,i,o,s,a,u,l,c,f,h,p=E.ajaxSetup({},t),d=p.context||p,g=p.context&&(d.nodeType||d.jquery)?E(d):E.event,v=E.Deferred(),m=E.Callbacks("once memory"),y=p.statusCode||{},b={},w={},x="canceled",T={readyState:0,getResponseHeader:function(e){var t;if(l){if(!s)for(s={};t=jt.exec(o);)s[t[1].toLowerCase()+" "]=(s[t[1].toLowerCase()+" "]||[]).concat(t[2]);t=s[e.toLowerCase()+" "]}return null==t?null:t.join(", ")},getAllResponseHeaders:function(){return l?o:null},setRequestHeader:function(e,t){return null==l&&(e=w[e.toLowerCase()]=w[e.toLowerCase()]||e,b[e]=t),this},overrideMimeType:function(e){return null==l&&(p.mimeType=e),this},statusCode:function(e){var t;if(e)if(l)T.always(e[T.status]);else for(t in e)y[t]=[y[t],e[t]];return this},abort:function(e){var t=e||x;return r&&r.abort(t),S(0,t),this}};if(v.promise(T),p.url=((e||p.url||xt.href)+"").replace(Ft,xt.protocol+"//"),p.type=t.method||t.type||p.method||p.type,p.dataTypes=(p.dataType||"*").toLowerCase().match(P)||[""],null==p.crossDomain){u=_.createElement("a");try{u.href=p.url,u.href=u.href,p.crossDomain=It.protocol+"//"+It.host!=u.protocol+"//"+u.host}catch(e){p.crossDomain=!0}}if(p.data&&p.processData&&"string"!=typeof p.data&&(p.data=E.param(p.data,p.traditional)),Wt(Lt,p,t,T),l)return T;for(f in(c=E.event&&p.global)&&0==E.active++&&E.event.trigger("ajaxStart"),p.type=p.type.toUpperCase(),p.hasContent=!Ot.test(p.type),i=p.url.replace(Nt,""),p.hasContent?p.data&&p.processData&&0===(p.contentType||"").indexOf("application/x-www-form-urlencoded")&&(p.data=p.data.replace(kt,"+")):(h=p.url.slice(i.length),p.data&&(p.processData||"string"==typeof p.data)&&(i+=(Tt.test(i)?"&":"?")+p.data,delete p.data),!1===p.cache&&(i=i.replace(Rt,"$1"),h=(Tt.test(i)?"&":"?")+"_="+Et.guid+++h),p.url=i+h),p.ifModified&&(E.lastModified[i]&&T.setRequestHeader("If-Modified-Since",E.lastModified[i]),E.etag[i]&&T.setRequestHeader("If-None-Match",E.etag[i])),(p.data&&p.hasContent&&!1!==p.contentType||t.contentType)&&T.setRequestHeader("Content-Type",p.contentType),T.setRequestHeader("Accept",p.dataTypes[0]&&p.accepts[p.dataTypes[0]]?p.accepts[p.dataTypes[0]]+("*"!==p.dataTypes[0]?", "+qt+"; q=0.01":""):p.accepts["*"]),p.headers)T.setRequestHeader(f,p.headers[f]);if(p.beforeSend&&(!1===p.beforeSend.call(d,T,p)||l))return T.abort();if(x="abort",m.add(p.complete),T.done(p.success),T.fail(p.error),r=Wt(Pt,p,t,T)){if(T.readyState=1,c&&g.trigger("ajaxSend",[T,p]),l)return T;p.async&&p.timeout>0&&(a=n.setTimeout((function(){T.abort("timeout")}),p.timeout));try{l=!1,r.send(b,S)}catch(e){if(l)throw e;S(-1,e)}}else S(-1,"No Transport");function S(e,t,s,u){var f,h,_,b,w,x=t;l||(l=!0,a&&n.clearTimeout(a),r=void 0,o=u||"",T.readyState=e>0?4:0,f=e>=200&&e<300||304===e,s&&(b=function(e,t,n){for(var r,i,o,s,a=e.contents,u=e.dataTypes;"*"===u[0];)u.shift(),void 0===r&&(r=e.mimeType||t.getResponseHeader("Content-Type"));if(r)for(i in a)if(a[i]&&a[i].test(r)){u.unshift(i);break}if(u[0]in n)o=u[0];else{for(i in n){if(!u[0]||e.converters[i+" "+u[0]]){o=i;break}s||(s=i)}o=o||s}if(o)return o!==u[0]&&u.unshift(o),n[o]}(p,T,s)),!f&&E.inArray("script",p.dataTypes)>-1&&E.inArray("json",p.dataTypes)<0&&(p.converters["text script"]=function(){}),b=function(e,t,n,r){var i,o,s,a,u,l={},c=e.dataTypes.slice();if(c[1])for(s in e.converters)l[s.toLowerCase()]=e.converters[s];for(o=c.shift();o;)if(e.responseFields[o]&&(n[e.responseFields[o]]=t),!u&&r&&e.dataFilter&&(t=e.dataFilter(t,e.dataType)),u=o,o=c.shift())if("*"===o)o=u;else if("*"!==u&&u!==o){if(!(s=l[u+" "+o]||l["* "+o]))for(i in l)if((a=i.split(" "))[1]===o&&(s=l[u+" "+a[0]]||l["* "+a[0]])){!0===s?s=l[i]:!0!==l[i]&&(o=a[0],c.unshift(a[1]));break}if(!0!==s)if(s&&e.throws)t=s(t);else try{t=s(t)}catch(e){return{state:"parsererror",error:s?e:"No conversion from "+u+" to "+o}}}return{state:"success",data:t}}(p,b,T,f),f?(p.ifModified&&((w=T.getResponseHeader("Last-Modified"))&&(E.lastModified[i]=w),(w=T.getResponseHeader("etag"))&&(E.etag[i]=w)),204===e||"HEAD"===p.type?x="nocontent":304===e?x="notmodified":(x=b.state,h=b.data,f=!(_=b.error))):(_=x,!e&&x||(x="error",e<0&&(e=0))),T.status=e,T.statusText=(t||x)+"",f?v.resolveWith(d,[h,x,T]):v.rejectWith(d,[T,x,_]),T.statusCode(y),y=void 0,c&&g.trigger(f?"ajaxSuccess":"ajaxError",[T,p,f?h:_]),m.fireWith(d,[T,x]),c&&(g.trigger("ajaxComplete",[T,p]),--E.active||E.event.trigger("ajaxStop")))}return T},getJSON:function(e,t,n){return E.get(e,t,n,"json")},getScript:function(e,t){return E.get(e,void 0,t,"script")}}),E.each(["get","post"],(function(e,t){E[t]=function(e,n,r,i){return m(n)&&(i=i||r,r=n,n=void 0),E.ajax(E.extend({url:e,type:t,dataType:i,data:n,success:r},E.isPlainObject(e)&&e))}})),E.ajaxPrefilter((function(e){var t;for(t in e.headers)"content-type"===t.toLowerCase()&&(e.contentType=e.headers[t]||"")})),E._evalUrl=function(e,t,n){return E.ajax({url:e,type:"GET",dataType:"script",cache:!0,async:!1,global:!1,converters:{"text script":function(){}},dataFilter:function(e){E.globalEval(e,t,n)}})},E.fn.extend({wrapAll:function(e){var t;return this[0]&&(m(e)&&(e=e.call(this[0])),t=E(e,this[0].ownerDocument).eq(0).clone(!0),this[0].parentNode&&t.insertBefore(this[0]),t.map((function(){for(var e=this;e.firstElementChild;)e=e.firstElementChild;return e})).append(this)),this},wrapInner:function(e){return m(e)?this.each((function(t){E(this).wrapInner(e.call(this,t))})):this.each((function(){var t=E(this),n=t.contents();n.length?n.wrapAll(e):t.append(e)}))},wrap:function(e){var t=m(e);return this.each((function(n){E(this).wrapAll(t?e.call(this,n):e)}))},unwrap:function(e){return this.parent(e).not("body").each((function(){E(this).replaceWith(this.childNodes)})),this}}),E.expr.pseudos.hidden=function(e){return!E.expr.pseudos.visible(e)},E.expr.pseudos.visible=function(e){return!!(e.offsetWidth||e.offsetHeight||e.getClientRects().length)},E.ajaxSettings.xhr=function(){try{return new n.XMLHttpRequest}catch(e){}};var Ht={0:200,1223:204},Ut=E.ajaxSettings.xhr();v.cors=!!Ut&&"withCredentials"in Ut,v.ajax=Ut=!!Ut,E.ajaxTransport((function(e){var t,r;if(v.cors||Ut&&!e.crossDomain)return{send:function(i,o){var s,a=e.xhr();if(a.open(e.type,e.url,e.async,e.username,e.password),e.xhrFields)for(s in e.xhrFields)a[s]=e.xhrFields[s];for(s in e.mimeType&&a.overrideMimeType&&a.overrideMimeType(e.mimeType),e.crossDomain||i["X-Requested-With"]||(i["X-Requested-With"]="XMLHttpRequest"),i)a.setRequestHeader(s,i[s]);t=function(e){return function(){t&&(t=r=a.onload=a.onerror=a.onabort=a.ontimeout=a.onreadystatechange=null,"abort"===e?a.abort():"error"===e?"number"!=typeof a.status?o(0,"error"):o(a.status,a.statusText):o(Ht[a.status]||a.status,a.statusText,"text"!==(a.responseType||"text")||"string"!=typeof a.responseText?{binary:a.response}:{text:a.responseText},a.getAllResponseHeaders()))}},a.onload=t(),r=a.onerror=a.ontimeout=t("error"),void 0!==a.onabort?a.onabort=r:a.onreadystatechange=function(){4===a.readyState&&n.setTimeout((function(){t&&r()}))},t=t("abort");try{a.send(e.hasContent&&e.data||null)}catch(e){if(t)throw e}},abort:function(){t&&t()}}})),E.ajaxPrefilter((function(e){e.crossDomain&&(e.contents.script=!1)})),E.ajaxSetup({accepts:{script:"text/javascript, application/javascript, application/ecmascript, application/x-ecmascript"},contents:{script:/\b(?:java|ecma)script\b/},converters:{"text script":function(e){return E.globalEval(e),e}}}),E.ajaxPrefilter("script",(function(e){void 0===e.cache&&(e.cache=!1),e.crossDomain&&(e.type="GET")})),E.ajaxTransport("script",(function(e){var t,n;if(e.crossDomain||e.scriptAttrs)return{send:function(r,i){t=E("<script>").attr(e.scriptAttrs||{}).prop({charset:e.scriptCharset,src:e.url}).on("load error",n=function(e){t.remove(),n=null,e&&i("error"===e.type?404:200,e.type)}),_.head.appendChild(t[0])},abort:function(){n&&n()}}}));var $t,Xt=[],Vt=/(=)\?(?=&|$)|\?\?/;E.ajaxSetup({jsonp:"callback",jsonpCallback:function(){var e=Xt.pop()||E.expando+"_"+Et.guid++;return this[e]=!0,e}}),E.ajaxPrefilter("json jsonp",(function(e,t,r){var i,o,s,a=!1!==e.jsonp&&(Vt.test(e.url)?"url":"string"==typeof e.data&&0===(e.contentType||"").indexOf("application/x-www-form-urlencoded")&&Vt.test(e.data)&&"data");if(a||"jsonp"===e.dataTypes[0])return i=e.jsonpCallback=m(e.jsonpCallback)?e.jsonpCallback():e.jsonpCallback,a?e[a]=e[a].replace(Vt,"$1"+i):!1!==e.jsonp&&(e.url+=(Tt.test(e.url)?"&":"?")+e.jsonp+"="+i),e.converters["script json"]=function(){return s||E.error(i+" was not called"),s[0]},e.dataTypes[0]="json",o=n[i],n[i]=function(){s=arguments},r.always((function(){void 0===o?E(n).removeProp(i):n[i]=o,e[i]&&(e.jsonpCallback=t.jsonpCallback,Xt.push(i)),s&&m(o)&&o(s[0]),s=o=void 0})),"script"})),v.createHTMLDocument=(($t=_.implementation.createHTMLDocument("").body).innerHTML="<form></form><form></form>",2===$t.childNodes.length),E.parseHTML=function(e,t,n){return"string"!=typeof e?[]:("boolean"==typeof t&&(n=t,t=!1),t||(v.createHTMLDocument?((r=(t=_.implementation.createHTMLDocument("")).createElement("base")).href=_.location.href,t.head.appendChild(r)):t=_),o=!n&&[],(i=k.exec(e))?[t.createElement(i[1])]:(i=xe([e],t,o),o&&o.length&&E(o).remove(),E.merge([],i.childNodes)));var r,i,o},E.fn.load=function(e,t,n){var r,i,o,s=this,a=e.indexOf(" ");return a>-1&&(r=vt(e.slice(a)),e=e.slice(0,a)),m(t)?(n=t,t=void 0):t&&"object"==typeof t&&(i="POST"),s.length>0&&E.ajax({url:e,type:i||"GET",dataType:"html",data:t}).done((function(e){o=arguments,s.html(r?E("<div>").append(E.parseHTML(e)).find(r):e)})).always(n&&function(e,t){s.each((function(){n.apply(this,o||[e.responseText,t,e])}))}),this},E.expr.pseudos.animated=function(e){return E.grep(E.timers,(function(t){return e===t.elem})).length},E.offset={setOffset:function(e,t,n){var r,i,o,s,a,u,l=E.css(e,"position"),c=E(e),f={};"static"===l&&(e.style.position="relative"),a=c.offset(),o=E.css(e,"top"),u=E.css(e,"left"),("absolute"===l||"fixed"===l)&&(o+u).indexOf("auto")>-1?(s=(r=c.position()).top,i=r.left):(s=parseFloat(o)||0,i=parseFloat(u)||0),m(t)&&(t=t.call(e,n,E.extend({},a))),null!=t.top&&(f.top=t.top-a.top+s),null!=t.left&&(f.left=t.left-a.left+i),"using"in t?t.using.call(e,f):c.css(f)}},E.fn.extend({offset:function(e){if(arguments.length)return void 0===e?this:this.each((function(t){E.offset.setOffset(this,e,t)}));var t,n,r=this[0];return r?r.getClientRects().length?(t=r.getBoundingClientRect(),n=r.ownerDocument.defaultView,{top:t.top+n.pageYOffset,left:t.left+n.pageXOffset}):{top:0,left:0}:void 0},position:function(){if(this[0]){var e,t,n,r=this[0],i={top:0,left:0};if("fixed"===E.css(r,"position"))t=r.getBoundingClientRect();else{for(t=this.offset(),n=r.ownerDocument,e=r.offsetParent||n.documentElement;e&&(e===n.body||e===n.documentElement)&&"static"===E.css(e,"position");)e=e.parentNode;e&&e!==r&&1===e.nodeType&&((i=E(e).offset()).top+=E.css(e,"borderTopWidth",!0),i.left+=E.css(e,"borderLeftWidth",!0))}return{top:t.top-i.top-E.css(r,"marginTop",!0),left:t.left-i.left-E.css(r,"marginLeft",!0)}}},offsetParent:function(){return this.map((function(){for(var e=this.offsetParent;e&&"static"===E.css(e,"position");)e=e.offsetParent;return e||oe}))}}),E.each({scrollLeft:"pageXOffset",scrollTop:"pageYOffset"},(function(e,t){var n="pageYOffset"===t;E.fn[e]=function(r){return U(this,(function(e,r,i){var o;if(y(e)?o=e:9===e.nodeType&&(o=e.defaultView),void 0===i)return o?o[t]:e[r];o?o.scrollTo(n?o.pageXOffset:i,n?i:o.pageYOffset):e[r]=i}),e,r,arguments.length)}})),E.each(["top","left"],(function(e,t){E.cssHooks[t]=Ue(v.pixelPosition,(function(e,n){if(n)return n=He(e,t),Ie.test(n)?E(e).position()[t]+"px":n}))})),E.each({Height:"height",Width:"width"},(function(e,t){E.each({padding:"inner"+e,content:t,"":"outer"+e},(function(n,r){E.fn[r]=function(i,o){var s=arguments.length&&(n||"boolean"!=typeof i),a=n||(!0===i||!0===o?"margin":"border");return U(this,(function(t,n,i){var o;return y(t)?0===r.indexOf("outer")?t["inner"+e]:t.document.documentElement["client"+e]:9===t.nodeType?(o=t.documentElement,Math.max(t.body["scroll"+e],o["scroll"+e],t.body["offset"+e],o["offset"+e],o["client"+e])):void 0===i?E.css(t,n,a):E.style(t,n,i,a)}),t,s?i:void 0,s)}}))})),E.each(["ajaxStart","ajaxStop","ajaxComplete","ajaxError","ajaxSuccess","ajaxSend"],(function(e,t){E.fn[t]=function(e){return this.on(t,e)}})),E.fn.extend({bind:function(e,t,n){return this.on(e,null,t,n)},unbind:function(e,t){return this.off(e,null,t)},delegate:function(e,t,n,r){return this.on(t,e,n,r)},undelegate:function(e,t,n){return 1===arguments.length?this.off(e,"**"):this.off(t,e||"**",n)},hover:function(e,t){return this.mouseenter(e).mouseleave(t||e)}}),E.each("blur focus focusin focusout resize scroll click dblclick mousedown mouseup mousemove mouseover mouseout mouseenter mouseleave change select submit keydown keypress keyup contextmenu".split(" "),(function(e,t){E.fn[t]=function(e,n){return arguments.length>0?this.on(t,null,e,n):this.trigger(t)}}));var Gt=/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g;E.proxy=function(e,t){var n,r,i;if("string"==typeof t&&(n=e[t],t=e,e=n),m(e))return r=a.call(arguments,2),(i=function(){return e.apply(t||this,r.concat(a.call(arguments)))}).guid=e.guid=e.guid||E.guid++,i},E.holdReady=function(e){e?E.readyWait++:E.ready(!0)},E.isArray=Array.isArray,E.parseJSON=JSON.parse,E.nodeName=z,E.isFunction=m,E.isWindow=y,E.camelCase=G,E.type=x,E.now=Date.now,E.isNumeric=function(e){var t=E.type(e);return("number"===t||"string"===t)&&!isNaN(e-parseFloat(e))},E.trim=function(e){return null==e?"":(e+"").replace(Gt,"")},void 0===(r=function(){return E}.apply(t,[]))||(e.exports=r);var Kt=n.jQuery,Yt=n.$;return E.noConflict=function(e){return n.$===E&&(n.$=Yt),e&&n.jQuery===E&&(n.jQuery=Kt),E},void 0===i&&(n.jQuery=n.$=E),E}))},function(e,t,n){var r,i,o;i=[n(0)],void 0===(o="function"==typeof(r=function(e){return e.ui=e.ui||{},e.ui.version="1.12.1"})?r.apply(t,i):r)||(e.exports=o)},function(e,t){e.exports=function(e){this.$el=e}},function(e,t,n){var r,i,o;
/*!
 * jQuery UI Widget 1.12.1
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */i=[n(0),n(1)],void 0===(o="function"==typeof(r=function(e){var t,n=0,r=Array.prototype.slice;return e.cleanData=(t=e.cleanData,function(n){var r,i,o;for(o=0;null!=(i=n[o]);o++)try{(r=e._data(i,"events"))&&r.remove&&e(i).triggerHandler("remove")}catch(e){}t(n)}),e.widget=function(t,n,r){var i,o,s,a={},u=t.split(".")[0],l=u+"-"+(t=t.split(".")[1]);return r||(r=n,n=e.Widget),e.isArray(r)&&(r=e.extend.apply(null,[{}].concat(r))),e.expr[":"][l.toLowerCase()]=function(t){return!!e.data(t,l)},e[u]=e[u]||{},i=e[u][t],o=e[u][t]=function(e,t){if(!this._createWidget)return new o(e,t);arguments.length&&this._createWidget(e,t)},e.extend(o,i,{version:r.version,_proto:e.extend({},r),_childConstructors:[]}),(s=new n).options=e.widget.extend({},s.options),e.each(r,(function(t,r){e.isFunction(r)?a[t]=function(){function e(){return n.prototype[t].apply(this,arguments)}function i(e){return n.prototype[t].apply(this,e)}return function(){var t,n=this._super,o=this._superApply;return this._super=e,this._superApply=i,t=r.apply(this,arguments),this._super=n,this._superApply=o,t}}():a[t]=r})),o.prototype=e.widget.extend(s,{widgetEventPrefix:i&&s.widgetEventPrefix||t},a,{constructor:o,namespace:u,widgetName:t,widgetFullName:l}),i?(e.each(i._childConstructors,(function(t,n){var r=n.prototype;e.widget(r.namespace+"."+r.widgetName,o,n._proto)})),delete i._childConstructors):n._childConstructors.push(o),e.widget.bridge(t,o),o},e.widget.extend=function(t){for(var n,i,o=r.call(arguments,1),s=0,a=o.length;s<a;s++)for(n in o[s])i=o[s][n],o[s].hasOwnProperty(n)&&void 0!==i&&(e.isPlainObject(i)?t[n]=e.isPlainObject(t[n])?e.widget.extend({},t[n],i):e.widget.extend({},i):t[n]=i);return t},e.widget.bridge=function(t,n){var i=n.prototype.widgetFullName||t;e.fn[t]=function(o){var s="string"==typeof o,a=r.call(arguments,1),u=this;return s?this.length||"instance"!==o?this.each((function(){var n,r=e.data(this,i);return"instance"===o?(u=r,!1):r?e.isFunction(r[o])&&"_"!==o.charAt(0)?(n=r[o].apply(r,a))!==r&&void 0!==n?(u=n&&n.jquery?u.pushStack(n.get()):n,!1):void 0:e.error("no such method '"+o+"' for "+t+" widget instance"):e.error("cannot call methods on "+t+" prior to initialization; attempted to call method '"+o+"'")})):u=void 0:(a.length&&(o=e.widget.extend.apply(null,[o].concat(a))),this.each((function(){var t=e.data(this,i);t?(t.option(o||{}),t._init&&t._init()):e.data(this,i,new n(o,this))}))),u}},e.Widget=function(){},e.Widget._childConstructors=[],e.Widget.prototype={widgetName:"widget",widgetEventPrefix:"",defaultElement:"<div>",options:{classes:{},disabled:!1,create:null},_createWidget:function(t,r){r=e(r||this.defaultElement||this)[0],this.element=e(r),this.uuid=n++,this.eventNamespace="."+this.widgetName+this.uuid,this.bindings=e(),this.hoverable=e(),this.focusable=e(),this.classesElementLookup={},r!==this&&(e.data(r,this.widgetFullName,this),this._on(!0,this.element,{remove:function(e){e.target===r&&this.destroy()}}),this.document=e(r.style?r.ownerDocument:r.document||r),this.window=e(this.document[0].defaultView||this.document[0].parentWindow)),this.options=e.widget.extend({},this.options,this._getCreateOptions(),t),this._create(),this.options.disabled&&this._setOptionDisabled(this.options.disabled),this._trigger("create",null,this._getCreateEventData()),this._init()},_getCreateOptions:function(){return{}},_getCreateEventData:e.noop,_create:e.noop,_init:e.noop,destroy:function(){var t=this;this._destroy(),e.each(this.classesElementLookup,(function(e,n){t._removeClass(n,e)})),this.element.off(this.eventNamespace).removeData(this.widgetFullName),this.widget().off(this.eventNamespace).removeAttr("aria-disabled"),this.bindings.off(this.eventNamespace)},_destroy:e.noop,widget:function(){return this.element},option:function(t,n){var r,i,o,s=t;if(0===arguments.length)return e.widget.extend({},this.options);if("string"==typeof t)if(s={},r=t.split("."),t=r.shift(),r.length){for(i=s[t]=e.widget.extend({},this.options[t]),o=0;o<r.length-1;o++)i[r[o]]=i[r[o]]||{},i=i[r[o]];if(t=r.pop(),1===arguments.length)return void 0===i[t]?null:i[t];i[t]=n}else{if(1===arguments.length)return void 0===this.options[t]?null:this.options[t];s[t]=n}return this._setOptions(s),this},_setOptions:function(e){var t;for(t in e)this._setOption(t,e[t]);return this},_setOption:function(e,t){return"classes"===e&&this._setOptionClasses(t),this.options[e]=t,"disabled"===e&&this._setOptionDisabled(t),this},_setOptionClasses:function(t){var n,r,i;for(n in t)i=this.classesElementLookup[n],t[n]!==this.options.classes[n]&&i&&i.length&&(r=e(i.get()),this._removeClass(i,n),r.addClass(this._classes({element:r,keys:n,classes:t,add:!0})))},_setOptionDisabled:function(e){this._toggleClass(this.widget(),this.widgetFullName+"-disabled",null,!!e),e&&(this._removeClass(this.hoverable,null,"ui-state-hover"),this._removeClass(this.focusable,null,"ui-state-focus"))},enable:function(){return this._setOptions({disabled:!1})},disable:function(){return this._setOptions({disabled:!0})},_classes:function(t){var n=[],r=this;function i(i,o){var s,a;for(a=0;a<i.length;a++)s=r.classesElementLookup[i[a]]||e(),s=t.add?e(e.unique(s.get().concat(t.element.get()))):e(s.not(t.element).get()),r.classesElementLookup[i[a]]=s,n.push(i[a]),o&&t.classes[i[a]]&&n.push(t.classes[i[a]])}return t=e.extend({element:this.element,classes:this.options.classes||{}},t),this._on(t.element,{remove:"_untrackClassesElement"}),t.keys&&i(t.keys.match(/\S+/g)||[],!0),t.extra&&i(t.extra.match(/\S+/g)||[]),n.join(" ")},_untrackClassesElement:function(t){var n=this;e.each(n.classesElementLookup,(function(r,i){-1!==e.inArray(t.target,i)&&(n.classesElementLookup[r]=e(i.not(t.target).get()))}))},_removeClass:function(e,t,n){return this._toggleClass(e,t,n,!1)},_addClass:function(e,t,n){return this._toggleClass(e,t,n,!0)},_toggleClass:function(e,t,n,r){r="boolean"==typeof r?r:n;var i="string"==typeof e||null===e,o={extra:i?t:n,keys:i?e:t,element:i?this.element:e,add:r};return o.element.toggleClass(this._classes(o),r),this},_on:function(t,n,r){var i,o=this;"boolean"!=typeof t&&(r=n,n=t,t=!1),r?(n=i=e(n),this.bindings=this.bindings.add(n)):(r=n,n=this.element,i=this.widget()),e.each(r,(function(r,s){function a(){if(t||!0!==o.options.disabled&&!e(this).hasClass("ui-state-disabled"))return("string"==typeof s?o[s]:s).apply(o,arguments)}"string"!=typeof s&&(a.guid=s.guid=s.guid||a.guid||e.guid++);var u=r.match(/^([\w:-]*)\s*(.*)$/),l=u[1]+o.eventNamespace,c=u[2];c?i.on(l,c,a):n.on(l,a)}))},_off:function(t,n){n=(n||"").split(" ").join(this.eventNamespace+" ")+this.eventNamespace,t.off(n).off(n),this.bindings=e(this.bindings.not(t).get()),this.focusable=e(this.focusable.not(t).get()),this.hoverable=e(this.hoverable.not(t).get())},_delay:function(e,t){var n=this;return setTimeout((function(){return("string"==typeof e?n[e]:e).apply(n,arguments)}),t||0)},_hoverable:function(t){this.hoverable=this.hoverable.add(t),this._on(t,{mouseenter:function(t){this._addClass(e(t.currentTarget),null,"ui-state-hover")},mouseleave:function(t){this._removeClass(e(t.currentTarget),null,"ui-state-hover")}})},_focusable:function(t){this.focusable=this.focusable.add(t),this._on(t,{focusin:function(t){this._addClass(e(t.currentTarget),null,"ui-state-focus")},focusout:function(t){this._removeClass(e(t.currentTarget),null,"ui-state-focus")}})},_trigger:function(t,n,r){var i,o,s=this.options[t];if(r=r||{},(n=e.Event(n)).type=(t===this.widgetEventPrefix?t:this.widgetEventPrefix+t).toLowerCase(),n.target=this.element[0],o=n.originalEvent)for(i in o)i in n||(n[i]=o[i]);return this.element.trigger(n,r),!(e.isFunction(s)&&!1===s.apply(this.element[0],[n].concat(r))||n.isDefaultPrevented())}},e.each({show:"fadeIn",hide:"fadeOut"},(function(t,n){e.Widget.prototype["_"+t]=function(r,i,o){var s;"string"==typeof i&&(i={effect:i});var a=i?!0===i||"number"==typeof i?n:i.effect||n:t;"number"==typeof(i=i||{})&&(i={duration:i}),s=!e.isEmptyObject(i),i.complete=o,i.delay&&r.delay(i.delay),s&&e.effects&&e.effects.effect[a]?r[t](i):a!==t&&r[a]?r[a](i.duration,i.easing,o):r.queue((function(n){e(this)[t](),o&&o.call(r[0]),n()}))}})),e.widget})?r.apply(t,i):r)||(e.exports=o)},function(e,t,n){e.exports=n(5),e.exports.version=n(22).version},function(e,t,n){var r=n(6),i=n(7),o=n(0),s=n(10);var a=r.DOMWidgetView.extend({initialize:function(e){a.__super__.initialize.apply(this,[e]),this.model.on("msg:custom",this.on_msg,this),this.model.on("change:width",this.size_changed,this),this.model.on("change:height",this.size_changed,this)},render:function(){var e=this,t=o("<canvas></canvas>");t.css("background-color","#000"),t.attr("tabindex","1"),this.$el.append(t),this.$canvas=t,this.c=s.init(t,this.model.get("webgl_config")),this.c.on_resize((function(t){e.model.set("width",t.size[0]),e.model.set("height",t.size[1]),e.touch()})),this.c.on_event_tick((function(){var t=e.c.event_queue.get();if(e.c.event_queue.clear(),t.length>0){var n={msg_type:"events",contents:t};e.send(n)}})),s.start_event_loop();this.send({msg_type:"init"}),this.size_changed(),this.c.resize(),this.c.resizable()},on_msg:function(e,t){if(null!=e&&"glir_commands"==e.msg_type){var n=e.commands;if("base64"==e.array_serialization)var r=e.buffers;else if("binary"==e.array_serialization){r=[];for(var i=0;i<t.length;i++)r[i]={storage_type:"binary",buffer:t[i]}}var o=function(e,t){for(var n=0;n<e.length;n++){var r=e[n];if("DATA"==r[0]){var i=r[3].buffer_index,o=r[3].buffer_shape,s=r[3].buffer_dtype;r[3]=t[i],r[3].shape=o,r[3].dtype=s}}return e}(n,r);for(i=0;i<o.length;i++){var s=n[i];this.c.command(s)}}},size_changed:function(){var e=[this.model.get("width"),this.model.get("height")];this.$canvas.css("width",e[0]+"px"),this.$canvas.css("height",e[1]+"px")},remove:function(){s.unregister(this.c),this.send({msg_type:"status",contents:"removed"})}}),u=r.DOMWidgetModel.extend({defaults:i.extend(r.DOMWidgetModel.prototype.defaults(),{_model_name:"VispyModel",_view_name:"VispyView",_model_module:"vispy",_view_module:"vispy",_model_module_version:"~0.2.0",_view_module_version:"~0.2.0"})});e.exports={VispyView:a,VispyModel:u}},function(t,n){t.exports=e},function(e,t,n){(function(e,r){var i;
/**
 * @license
 * Lodash <https://lodash.com/>
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */(function(){var o="Expected a function",s="__lodash_placeholder__",a=[["ary",128],["bind",1],["bindKey",2],["curry",8],["curryRight",16],["flip",512],["partial",32],["partialRight",64],["rearg",256]],u="[object Arguments]",l="[object Array]",c="[object Boolean]",f="[object Date]",h="[object Error]",p="[object Function]",d="[object GeneratorFunction]",g="[object Map]",v="[object Number]",m="[object Object]",y="[object RegExp]",_="[object Set]",b="[object String]",w="[object Symbol]",x="[object WeakMap]",E="[object ArrayBuffer]",T="[object DataView]",S="[object Float32Array]",C="[object Float64Array]",A="[object Int8Array]",D="[object Int16Array]",z="[object Int32Array]",k="[object Uint8Array]",N="[object Uint16Array]",R="[object Uint32Array]",j=/\b__p \+= '';/g,O=/\b(__p \+=) '' \+/g,F=/(__e\(.*?\)|\b__t\)) \+\n'';/g,L=/&(?:amp|lt|gt|quot|#39);/g,P=/[&<>"']/g,q=RegExp(L.source),I=RegExp(P.source),M=/<%-([\s\S]+?)%>/g,W=/<%([\s\S]+?)%>/g,B=/<%=([\s\S]+?)%>/g,H=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,U=/^\w*$/,$=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,X=/[\\^$.*+?()[\]{}|]/g,V=RegExp(X.source),G=/^\s+/,K=/\s/,Y=/\{(?:\n\/\* \[wrapped with .+\] \*\/)?\n?/,J=/\{\n\/\* \[wrapped with (.+)\] \*/,Z=/,? & /,Q=/[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g,ee=/[()=,{}\[\]\/\s]/,te=/\\(\\)?/g,ne=/\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g,re=/\w*$/,ie=/^[-+]0x[0-9a-f]+$/i,oe=/^0b[01]+$/i,se=/^\[object .+?Constructor\]$/,ae=/^0o[0-7]+$/i,ue=/^(?:0|[1-9]\d*)$/,le=/[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g,ce=/($^)/,fe=/['\n\r\u2028\u2029\\]/g,he="\\u0300-\\u036f\\ufe20-\\ufe2f\\u20d0-\\u20ff",pe="\\xac\\xb1\\xd7\\xf7\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf\\u2000-\\u206f \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000",de="[\\ud800-\\udfff]",ge="["+pe+"]",ve="["+he+"]",me="\\d+",ye="[\\u2700-\\u27bf]",_e="[a-z\\xdf-\\xf6\\xf8-\\xff]",be="[^\\ud800-\\udfff"+pe+me+"\\u2700-\\u27bfa-z\\xdf-\\xf6\\xf8-\\xffA-Z\\xc0-\\xd6\\xd8-\\xde]",we="\\ud83c[\\udffb-\\udfff]",xe="[^\\ud800-\\udfff]",Ee="(?:\\ud83c[\\udde6-\\uddff]){2}",Te="[\\ud800-\\udbff][\\udc00-\\udfff]",Se="[A-Z\\xc0-\\xd6\\xd8-\\xde]",Ce="(?:"+_e+"|"+be+")",Ae="(?:"+Se+"|"+be+")",De="(?:"+ve+"|"+we+")"+"?",ze="[\\ufe0e\\ufe0f]?"+De+("(?:\\u200d(?:"+[xe,Ee,Te].join("|")+")[\\ufe0e\\ufe0f]?"+De+")*"),ke="(?:"+[ye,Ee,Te].join("|")+")"+ze,Ne="(?:"+[xe+ve+"?",ve,Ee,Te,de].join("|")+")",Re=RegExp("['’]","g"),je=RegExp(ve,"g"),Oe=RegExp(we+"(?="+we+")|"+Ne+ze,"g"),Fe=RegExp([Se+"?"+_e+"+(?:['’](?:d|ll|m|re|s|t|ve))?(?="+[ge,Se,"$"].join("|")+")",Ae+"+(?:['’](?:D|LL|M|RE|S|T|VE))?(?="+[ge,Se+Ce,"$"].join("|")+")",Se+"?"+Ce+"+(?:['’](?:d|ll|m|re|s|t|ve))?",Se+"+(?:['’](?:D|LL|M|RE|S|T|VE))?","\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])","\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])",me,ke].join("|"),"g"),Le=RegExp("[\\u200d\\ud800-\\udfff"+he+"\\ufe0e\\ufe0f]"),Pe=/[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/,qe=["Array","Buffer","DataView","Date","Error","Float32Array","Float64Array","Function","Int8Array","Int16Array","Int32Array","Map","Math","Object","Promise","RegExp","Set","String","Symbol","TypeError","Uint8Array","Uint8ClampedArray","Uint16Array","Uint32Array","WeakMap","_","clearTimeout","isFinite","parseInt","setTimeout"],Ie=-1,Me={};Me[S]=Me[C]=Me[A]=Me[D]=Me[z]=Me[k]=Me["[object Uint8ClampedArray]"]=Me[N]=Me[R]=!0,Me[u]=Me[l]=Me[E]=Me[c]=Me[T]=Me[f]=Me[h]=Me[p]=Me[g]=Me[v]=Me[m]=Me[y]=Me[_]=Me[b]=Me[x]=!1;var We={};We[u]=We[l]=We[E]=We[T]=We[c]=We[f]=We[S]=We[C]=We[A]=We[D]=We[z]=We[g]=We[v]=We[m]=We[y]=We[_]=We[b]=We[w]=We[k]=We["[object Uint8ClampedArray]"]=We[N]=We[R]=!0,We[h]=We[p]=We[x]=!1;var Be={"\\":"\\","'":"'","\n":"n","\r":"r","\u2028":"u2028","\u2029":"u2029"},He=parseFloat,Ue=parseInt,$e="object"==typeof e&&e&&e.Object===Object&&e,Xe="object"==typeof self&&self&&self.Object===Object&&self,Ve=$e||Xe||Function("return this")(),Ge=t&&!t.nodeType&&t,Ke=Ge&&"object"==typeof r&&r&&!r.nodeType&&r,Ye=Ke&&Ke.exports===Ge,Je=Ye&&$e.process,Ze=function(){try{var e=Ke&&Ke.require&&Ke.require("util").types;return e||Je&&Je.binding&&Je.binding("util")}catch(e){}}(),Qe=Ze&&Ze.isArrayBuffer,et=Ze&&Ze.isDate,tt=Ze&&Ze.isMap,nt=Ze&&Ze.isRegExp,rt=Ze&&Ze.isSet,it=Ze&&Ze.isTypedArray;function ot(e,t,n){switch(n.length){case 0:return e.call(t);case 1:return e.call(t,n[0]);case 2:return e.call(t,n[0],n[1]);case 3:return e.call(t,n[0],n[1],n[2])}return e.apply(t,n)}function st(e,t,n,r){for(var i=-1,o=null==e?0:e.length;++i<o;){var s=e[i];t(r,s,n(s),e)}return r}function at(e,t){for(var n=-1,r=null==e?0:e.length;++n<r&&!1!==t(e[n],n,e););return e}function ut(e,t){for(var n=null==e?0:e.length;n--&&!1!==t(e[n],n,e););return e}function lt(e,t){for(var n=-1,r=null==e?0:e.length;++n<r;)if(!t(e[n],n,e))return!1;return!0}function ct(e,t){for(var n=-1,r=null==e?0:e.length,i=0,o=[];++n<r;){var s=e[n];t(s,n,e)&&(o[i++]=s)}return o}function ft(e,t){return!!(null==e?0:e.length)&&wt(e,t,0)>-1}function ht(e,t,n){for(var r=-1,i=null==e?0:e.length;++r<i;)if(n(t,e[r]))return!0;return!1}function pt(e,t){for(var n=-1,r=null==e?0:e.length,i=Array(r);++n<r;)i[n]=t(e[n],n,e);return i}function dt(e,t){for(var n=-1,r=t.length,i=e.length;++n<r;)e[i+n]=t[n];return e}function gt(e,t,n,r){var i=-1,o=null==e?0:e.length;for(r&&o&&(n=e[++i]);++i<o;)n=t(n,e[i],i,e);return n}function vt(e,t,n,r){var i=null==e?0:e.length;for(r&&i&&(n=e[--i]);i--;)n=t(n,e[i],i,e);return n}function mt(e,t){for(var n=-1,r=null==e?0:e.length;++n<r;)if(t(e[n],n,e))return!0;return!1}var yt=St("length");function _t(e,t,n){var r;return n(e,(function(e,n,i){if(t(e,n,i))return r=n,!1})),r}function bt(e,t,n,r){for(var i=e.length,o=n+(r?1:-1);r?o--:++o<i;)if(t(e[o],o,e))return o;return-1}function wt(e,t,n){return t==t?function(e,t,n){var r=n-1,i=e.length;for(;++r<i;)if(e[r]===t)return r;return-1}(e,t,n):bt(e,Et,n)}function xt(e,t,n,r){for(var i=n-1,o=e.length;++i<o;)if(r(e[i],t))return i;return-1}function Et(e){return e!=e}function Tt(e,t){var n=null==e?0:e.length;return n?Dt(e,t)/n:NaN}function St(e){return function(t){return null==t?void 0:t[e]}}function Ct(e){return function(t){return null==e?void 0:e[t]}}function At(e,t,n,r,i){return i(e,(function(e,i,o){n=r?(r=!1,e):t(n,e,i,o)})),n}function Dt(e,t){for(var n,r=-1,i=e.length;++r<i;){var o=t(e[r]);void 0!==o&&(n=void 0===n?o:n+o)}return n}function zt(e,t){for(var n=-1,r=Array(e);++n<e;)r[n]=t(n);return r}function kt(e){return e?e.slice(0,Gt(e)+1).replace(G,""):e}function Nt(e){return function(t){return e(t)}}function Rt(e,t){return pt(t,(function(t){return e[t]}))}function jt(e,t){return e.has(t)}function Ot(e,t){for(var n=-1,r=e.length;++n<r&&wt(t,e[n],0)>-1;);return n}function Ft(e,t){for(var n=e.length;n--&&wt(t,e[n],0)>-1;);return n}function Lt(e,t){for(var n=e.length,r=0;n--;)e[n]===t&&++r;return r}var Pt=Ct({"À":"A","Á":"A","Â":"A","Ã":"A","Ä":"A","Å":"A","à":"a","á":"a","â":"a","ã":"a","ä":"a","å":"a","Ç":"C","ç":"c","Ð":"D","ð":"d","È":"E","É":"E","Ê":"E","Ë":"E","è":"e","é":"e","ê":"e","ë":"e","Ì":"I","Í":"I","Î":"I","Ï":"I","ì":"i","í":"i","î":"i","ï":"i","Ñ":"N","ñ":"n","Ò":"O","Ó":"O","Ô":"O","Õ":"O","Ö":"O","Ø":"O","ò":"o","ó":"o","ô":"o","õ":"o","ö":"o","ø":"o","Ù":"U","Ú":"U","Û":"U","Ü":"U","ù":"u","ú":"u","û":"u","ü":"u","Ý":"Y","ý":"y","ÿ":"y","Æ":"Ae","æ":"ae","Þ":"Th","þ":"th","ß":"ss","Ā":"A","Ă":"A","Ą":"A","ā":"a","ă":"a","ą":"a","Ć":"C","Ĉ":"C","Ċ":"C","Č":"C","ć":"c","ĉ":"c","ċ":"c","č":"c","Ď":"D","Đ":"D","ď":"d","đ":"d","Ē":"E","Ĕ":"E","Ė":"E","Ę":"E","Ě":"E","ē":"e","ĕ":"e","ė":"e","ę":"e","ě":"e","Ĝ":"G","Ğ":"G","Ġ":"G","Ģ":"G","ĝ":"g","ğ":"g","ġ":"g","ģ":"g","Ĥ":"H","Ħ":"H","ĥ":"h","ħ":"h","Ĩ":"I","Ī":"I","Ĭ":"I","Į":"I","İ":"I","ĩ":"i","ī":"i","ĭ":"i","į":"i","ı":"i","Ĵ":"J","ĵ":"j","Ķ":"K","ķ":"k","ĸ":"k","Ĺ":"L","Ļ":"L","Ľ":"L","Ŀ":"L","Ł":"L","ĺ":"l","ļ":"l","ľ":"l","ŀ":"l","ł":"l","Ń":"N","Ņ":"N","Ň":"N","Ŋ":"N","ń":"n","ņ":"n","ň":"n","ŋ":"n","Ō":"O","Ŏ":"O","Ő":"O","ō":"o","ŏ":"o","ő":"o","Ŕ":"R","Ŗ":"R","Ř":"R","ŕ":"r","ŗ":"r","ř":"r","Ś":"S","Ŝ":"S","Ş":"S","Š":"S","ś":"s","ŝ":"s","ş":"s","š":"s","Ţ":"T","Ť":"T","Ŧ":"T","ţ":"t","ť":"t","ŧ":"t","Ũ":"U","Ū":"U","Ŭ":"U","Ů":"U","Ű":"U","Ų":"U","ũ":"u","ū":"u","ŭ":"u","ů":"u","ű":"u","ų":"u","Ŵ":"W","ŵ":"w","Ŷ":"Y","ŷ":"y","Ÿ":"Y","Ź":"Z","Ż":"Z","Ž":"Z","ź":"z","ż":"z","ž":"z","Ĳ":"IJ","ĳ":"ij","Œ":"Oe","œ":"oe","ŉ":"'n","ſ":"s"}),qt=Ct({"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"});function It(e){return"\\"+Be[e]}function Mt(e){return Le.test(e)}function Wt(e){var t=-1,n=Array(e.size);return e.forEach((function(e,r){n[++t]=[r,e]})),n}function Bt(e,t){return function(n){return e(t(n))}}function Ht(e,t){for(var n=-1,r=e.length,i=0,o=[];++n<r;){var a=e[n];a!==t&&a!==s||(e[n]=s,o[i++]=n)}return o}function Ut(e){var t=-1,n=Array(e.size);return e.forEach((function(e){n[++t]=e})),n}function $t(e){var t=-1,n=Array(e.size);return e.forEach((function(e){n[++t]=[e,e]})),n}function Xt(e){return Mt(e)?function(e){var t=Oe.lastIndex=0;for(;Oe.test(e);)++t;return t}(e):yt(e)}function Vt(e){return Mt(e)?function(e){return e.match(Oe)||[]}(e):function(e){return e.split("")}(e)}function Gt(e){for(var t=e.length;t--&&K.test(e.charAt(t)););return t}var Kt=Ct({"&amp;":"&","&lt;":"<","&gt;":">","&quot;":'"',"&#39;":"'"});var Yt=function e(t){var n,r=(t=null==t?Ve:Yt.defaults(Ve.Object(),t,Yt.pick(Ve,qe))).Array,i=t.Date,K=t.Error,he=t.Function,pe=t.Math,de=t.Object,ge=t.RegExp,ve=t.String,me=t.TypeError,ye=r.prototype,_e=he.prototype,be=de.prototype,we=t["__core-js_shared__"],xe=_e.toString,Ee=be.hasOwnProperty,Te=0,Se=(n=/[^.]+$/.exec(we&&we.keys&&we.keys.IE_PROTO||""))?"Symbol(src)_1."+n:"",Ce=be.toString,Ae=xe.call(de),De=Ve._,ze=ge("^"+xe.call(Ee).replace(X,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$"),ke=Ye?t.Buffer:void 0,Ne=t.Symbol,Oe=t.Uint8Array,Le=ke?ke.allocUnsafe:void 0,Be=Bt(de.getPrototypeOf,de),$e=de.create,Xe=be.propertyIsEnumerable,Ge=ye.splice,Ke=Ne?Ne.isConcatSpreadable:void 0,Je=Ne?Ne.iterator:void 0,Ze=Ne?Ne.toStringTag:void 0,yt=function(){try{var e=eo(de,"defineProperty");return e({},"",{}),e}catch(e){}}(),Ct=t.clearTimeout!==Ve.clearTimeout&&t.clearTimeout,Jt=i&&i.now!==Ve.Date.now&&i.now,Zt=t.setTimeout!==Ve.setTimeout&&t.setTimeout,Qt=pe.ceil,en=pe.floor,tn=de.getOwnPropertySymbols,nn=ke?ke.isBuffer:void 0,rn=t.isFinite,on=ye.join,sn=Bt(de.keys,de),an=pe.max,un=pe.min,ln=i.now,cn=t.parseInt,fn=pe.random,hn=ye.reverse,pn=eo(t,"DataView"),dn=eo(t,"Map"),gn=eo(t,"Promise"),vn=eo(t,"Set"),mn=eo(t,"WeakMap"),yn=eo(de,"create"),_n=mn&&new mn,bn={},wn=Do(pn),xn=Do(dn),En=Do(gn),Tn=Do(vn),Sn=Do(mn),Cn=Ne?Ne.prototype:void 0,An=Cn?Cn.valueOf:void 0,Dn=Cn?Cn.toString:void 0;function zn(e){if($s(e)&&!Os(e)&&!(e instanceof jn)){if(e instanceof Rn)return e;if(Ee.call(e,"__wrapped__"))return zo(e)}return new Rn(e)}var kn=function(){function e(){}return function(t){if(!Us(t))return{};if($e)return $e(t);e.prototype=t;var n=new e;return e.prototype=void 0,n}}();function Nn(){}function Rn(e,t){this.__wrapped__=e,this.__actions__=[],this.__chain__=!!t,this.__index__=0,this.__values__=void 0}function jn(e){this.__wrapped__=e,this.__actions__=[],this.__dir__=1,this.__filtered__=!1,this.__iteratees__=[],this.__takeCount__=4294967295,this.__views__=[]}function On(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var r=e[t];this.set(r[0],r[1])}}function Fn(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var r=e[t];this.set(r[0],r[1])}}function Ln(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var r=e[t];this.set(r[0],r[1])}}function Pn(e){var t=-1,n=null==e?0:e.length;for(this.__data__=new Ln;++t<n;)this.add(e[t])}function qn(e){var t=this.__data__=new Fn(e);this.size=t.size}function In(e,t){var n=Os(e),r=!n&&js(e),i=!n&&!r&&qs(e),o=!n&&!r&&!i&&Qs(e),s=n||r||i||o,a=s?zt(e.length,ve):[],u=a.length;for(var l in e)!t&&!Ee.call(e,l)||s&&("length"==l||i&&("offset"==l||"parent"==l)||o&&("buffer"==l||"byteLength"==l||"byteOffset"==l)||ao(l,u))||a.push(l);return a}function Mn(e){var t=e.length;return t?e[Pr(0,t-1)]:void 0}function Wn(e,t){return So(yi(e),Yn(t,0,e.length))}function Bn(e){return So(yi(e))}function Hn(e,t,n){(void 0!==n&&!ks(e[t],n)||void 0===n&&!(t in e))&&Gn(e,t,n)}function Un(e,t,n){var r=e[t];Ee.call(e,t)&&ks(r,n)&&(void 0!==n||t in e)||Gn(e,t,n)}function $n(e,t){for(var n=e.length;n--;)if(ks(e[n][0],t))return n;return-1}function Xn(e,t,n,r){return tr(e,(function(e,i,o){t(r,e,n(e),o)})),r}function Vn(e,t){return e&&_i(t,wa(t),e)}function Gn(e,t,n){"__proto__"==t&&yt?yt(e,t,{configurable:!0,enumerable:!0,value:n,writable:!0}):e[t]=n}function Kn(e,t){for(var n=-1,i=t.length,o=r(i),s=null==e;++n<i;)o[n]=s?void 0:va(e,t[n]);return o}function Yn(e,t,n){return e==e&&(void 0!==n&&(e=e<=n?e:n),void 0!==t&&(e=e>=t?e:t)),e}function Jn(e,t,n,r,i,o){var s,a=1&t,l=2&t,h=4&t;if(n&&(s=i?n(e,r,i,o):n(e)),void 0!==s)return s;if(!Us(e))return e;var x=Os(e);if(x){if(s=function(e){var t=e.length,n=new e.constructor(t);t&&"string"==typeof e[0]&&Ee.call(e,"index")&&(n.index=e.index,n.input=e.input);return n}(e),!a)return yi(e,s)}else{var j=ro(e),O=j==p||j==d;if(qs(e))return hi(e,a);if(j==m||j==u||O&&!i){if(s=l||O?{}:oo(e),!a)return l?function(e,t){return _i(e,no(e),t)}(e,function(e,t){return e&&_i(t,xa(t),e)}(s,e)):function(e,t){return _i(e,to(e),t)}(e,Vn(s,e))}else{if(!We[j])return i?e:{};s=function(e,t,n){var r=e.constructor;switch(t){case E:return pi(e);case c:case f:return new r(+e);case T:return function(e,t){var n=t?pi(e.buffer):e.buffer;return new e.constructor(n,e.byteOffset,e.byteLength)}(e,n);case S:case C:case A:case D:case z:case k:case"[object Uint8ClampedArray]":case N:case R:return di(e,n);case g:return new r;case v:case b:return new r(e);case y:return function(e){var t=new e.constructor(e.source,re.exec(e));return t.lastIndex=e.lastIndex,t}(e);case _:return new r;case w:return i=e,An?de(An.call(i)):{}}var i}(e,j,a)}}o||(o=new qn);var F=o.get(e);if(F)return F;o.set(e,s),Ys(e)?e.forEach((function(r){s.add(Jn(r,t,n,r,e,o))})):Xs(e)&&e.forEach((function(r,i){s.set(i,Jn(r,t,n,i,e,o))}));var L=x?void 0:(h?l?Vi:Xi:l?xa:wa)(e);return at(L||e,(function(r,i){L&&(r=e[i=r]),Un(s,i,Jn(r,t,n,i,e,o))})),s}function Zn(e,t,n){var r=n.length;if(null==e)return!r;for(e=de(e);r--;){var i=n[r],o=t[i],s=e[i];if(void 0===s&&!(i in e)||!o(s))return!1}return!0}function Qn(e,t,n){if("function"!=typeof e)throw new me(o);return wo((function(){e.apply(void 0,n)}),t)}function er(e,t,n,r){var i=-1,o=ft,s=!0,a=e.length,u=[],l=t.length;if(!a)return u;n&&(t=pt(t,Nt(n))),r?(o=ht,s=!1):t.length>=200&&(o=jt,s=!1,t=new Pn(t));e:for(;++i<a;){var c=e[i],f=null==n?c:n(c);if(c=r||0!==c?c:0,s&&f==f){for(var h=l;h--;)if(t[h]===f)continue e;u.push(c)}else o(t,f,r)||u.push(c)}return u}zn.templateSettings={escape:M,evaluate:W,interpolate:B,variable:"",imports:{_:zn}},zn.prototype=Nn.prototype,zn.prototype.constructor=zn,Rn.prototype=kn(Nn.prototype),Rn.prototype.constructor=Rn,jn.prototype=kn(Nn.prototype),jn.prototype.constructor=jn,On.prototype.clear=function(){this.__data__=yn?yn(null):{},this.size=0},On.prototype.delete=function(e){var t=this.has(e)&&delete this.__data__[e];return this.size-=t?1:0,t},On.prototype.get=function(e){var t=this.__data__;if(yn){var n=t[e];return"__lodash_hash_undefined__"===n?void 0:n}return Ee.call(t,e)?t[e]:void 0},On.prototype.has=function(e){var t=this.__data__;return yn?void 0!==t[e]:Ee.call(t,e)},On.prototype.set=function(e,t){var n=this.__data__;return this.size+=this.has(e)?0:1,n[e]=yn&&void 0===t?"__lodash_hash_undefined__":t,this},Fn.prototype.clear=function(){this.__data__=[],this.size=0},Fn.prototype.delete=function(e){var t=this.__data__,n=$n(t,e);return!(n<0)&&(n==t.length-1?t.pop():Ge.call(t,n,1),--this.size,!0)},Fn.prototype.get=function(e){var t=this.__data__,n=$n(t,e);return n<0?void 0:t[n][1]},Fn.prototype.has=function(e){return $n(this.__data__,e)>-1},Fn.prototype.set=function(e,t){var n=this.__data__,r=$n(n,e);return r<0?(++this.size,n.push([e,t])):n[r][1]=t,this},Ln.prototype.clear=function(){this.size=0,this.__data__={hash:new On,map:new(dn||Fn),string:new On}},Ln.prototype.delete=function(e){var t=Zi(this,e).delete(e);return this.size-=t?1:0,t},Ln.prototype.get=function(e){return Zi(this,e).get(e)},Ln.prototype.has=function(e){return Zi(this,e).has(e)},Ln.prototype.set=function(e,t){var n=Zi(this,e),r=n.size;return n.set(e,t),this.size+=n.size==r?0:1,this},Pn.prototype.add=Pn.prototype.push=function(e){return this.__data__.set(e,"__lodash_hash_undefined__"),this},Pn.prototype.has=function(e){return this.__data__.has(e)},qn.prototype.clear=function(){this.__data__=new Fn,this.size=0},qn.prototype.delete=function(e){var t=this.__data__,n=t.delete(e);return this.size=t.size,n},qn.prototype.get=function(e){return this.__data__.get(e)},qn.prototype.has=function(e){return this.__data__.has(e)},qn.prototype.set=function(e,t){var n=this.__data__;if(n instanceof Fn){var r=n.__data__;if(!dn||r.length<199)return r.push([e,t]),this.size=++n.size,this;n=this.__data__=new Ln(r)}return n.set(e,t),this.size=n.size,this};var tr=xi(lr),nr=xi(cr,!0);function rr(e,t){var n=!0;return tr(e,(function(e,r,i){return n=!!t(e,r,i)})),n}function ir(e,t,n){for(var r=-1,i=e.length;++r<i;){var o=e[r],s=t(o);if(null!=s&&(void 0===a?s==s&&!Zs(s):n(s,a)))var a=s,u=o}return u}function or(e,t){var n=[];return tr(e,(function(e,r,i){t(e,r,i)&&n.push(e)})),n}function sr(e,t,n,r,i){var o=-1,s=e.length;for(n||(n=so),i||(i=[]);++o<s;){var a=e[o];t>0&&n(a)?t>1?sr(a,t-1,n,r,i):dt(i,a):r||(i[i.length]=a)}return i}var ar=Ei(),ur=Ei(!0);function lr(e,t){return e&&ar(e,t,wa)}function cr(e,t){return e&&ur(e,t,wa)}function fr(e,t){return ct(t,(function(t){return Ws(e[t])}))}function hr(e,t){for(var n=0,r=(t=ui(t,e)).length;null!=e&&n<r;)e=e[Ao(t[n++])];return n&&n==r?e:void 0}function pr(e,t,n){var r=t(e);return Os(e)?r:dt(r,n(e))}function dr(e){return null==e?void 0===e?"[object Undefined]":"[object Null]":Ze&&Ze in de(e)?function(e){var t=Ee.call(e,Ze),n=e[Ze];try{e[Ze]=void 0;var r=!0}catch(e){}var i=Ce.call(e);r&&(t?e[Ze]=n:delete e[Ze]);return i}(e):function(e){return Ce.call(e)}(e)}function gr(e,t){return e>t}function vr(e,t){return null!=e&&Ee.call(e,t)}function mr(e,t){return null!=e&&t in de(e)}function yr(e,t,n){for(var i=n?ht:ft,o=e[0].length,s=e.length,a=s,u=r(s),l=1/0,c=[];a--;){var f=e[a];a&&t&&(f=pt(f,Nt(t))),l=un(f.length,l),u[a]=!n&&(t||o>=120&&f.length>=120)?new Pn(a&&f):void 0}f=e[0];var h=-1,p=u[0];e:for(;++h<o&&c.length<l;){var d=f[h],g=t?t(d):d;if(d=n||0!==d?d:0,!(p?jt(p,g):i(c,g,n))){for(a=s;--a;){var v=u[a];if(!(v?jt(v,g):i(e[a],g,n)))continue e}p&&p.push(g),c.push(d)}}return c}function _r(e,t,n){var r=null==(e=mo(e,t=ui(t,e)))?e:e[Ao(Mo(t))];return null==r?void 0:ot(r,e,n)}function br(e){return $s(e)&&dr(e)==u}function wr(e,t,n,r,i){return e===t||(null==e||null==t||!$s(e)&&!$s(t)?e!=e&&t!=t:function(e,t,n,r,i,o){var s=Os(e),a=Os(t),p=s?l:ro(e),d=a?l:ro(t),x=(p=p==u?m:p)==m,S=(d=d==u?m:d)==m,C=p==d;if(C&&qs(e)){if(!qs(t))return!1;s=!0,x=!1}if(C&&!x)return o||(o=new qn),s||Qs(e)?Ui(e,t,n,r,i,o):function(e,t,n,r,i,o,s){switch(n){case T:if(e.byteLength!=t.byteLength||e.byteOffset!=t.byteOffset)return!1;e=e.buffer,t=t.buffer;case E:return!(e.byteLength!=t.byteLength||!o(new Oe(e),new Oe(t)));case c:case f:case v:return ks(+e,+t);case h:return e.name==t.name&&e.message==t.message;case y:case b:return e==t+"";case g:var a=Wt;case _:var u=1&r;if(a||(a=Ut),e.size!=t.size&&!u)return!1;var l=s.get(e);if(l)return l==t;r|=2,s.set(e,t);var p=Ui(a(e),a(t),r,i,o,s);return s.delete(e),p;case w:if(An)return An.call(e)==An.call(t)}return!1}(e,t,p,n,r,i,o);if(!(1&n)){var A=x&&Ee.call(e,"__wrapped__"),D=S&&Ee.call(t,"__wrapped__");if(A||D){var z=A?e.value():e,k=D?t.value():t;return o||(o=new qn),i(z,k,n,r,o)}}if(!C)return!1;return o||(o=new qn),function(e,t,n,r,i,o){var s=1&n,a=Xi(e),u=a.length,l=Xi(t).length;if(u!=l&&!s)return!1;var c=u;for(;c--;){var f=a[c];if(!(s?f in t:Ee.call(t,f)))return!1}var h=o.get(e),p=o.get(t);if(h&&p)return h==t&&p==e;var d=!0;o.set(e,t),o.set(t,e);var g=s;for(;++c<u;){f=a[c];var v=e[f],m=t[f];if(r)var y=s?r(m,v,f,t,e,o):r(v,m,f,e,t,o);if(!(void 0===y?v===m||i(v,m,n,r,o):y)){d=!1;break}g||(g="constructor"==f)}if(d&&!g){var _=e.constructor,b=t.constructor;_==b||!("constructor"in e)||!("constructor"in t)||"function"==typeof _&&_ instanceof _&&"function"==typeof b&&b instanceof b||(d=!1)}return o.delete(e),o.delete(t),d}(e,t,n,r,i,o)}(e,t,n,r,wr,i))}function xr(e,t,n,r){var i=n.length,o=i,s=!r;if(null==e)return!o;for(e=de(e);i--;){var a=n[i];if(s&&a[2]?a[1]!==e[a[0]]:!(a[0]in e))return!1}for(;++i<o;){var u=(a=n[i])[0],l=e[u],c=a[1];if(s&&a[2]){if(void 0===l&&!(u in e))return!1}else{var f=new qn;if(r)var h=r(l,c,u,e,t,f);if(!(void 0===h?wr(c,l,3,r,f):h))return!1}}return!0}function Er(e){return!(!Us(e)||(t=e,Se&&Se in t))&&(Ws(e)?ze:se).test(Do(e));var t}function Tr(e){return"function"==typeof e?e:null==e?Va:"object"==typeof e?Os(e)?kr(e[0],e[1]):zr(e):nu(e)}function Sr(e){if(!ho(e))return sn(e);var t=[];for(var n in de(e))Ee.call(e,n)&&"constructor"!=n&&t.push(n);return t}function Cr(e){if(!Us(e))return function(e){var t=[];if(null!=e)for(var n in de(e))t.push(n);return t}(e);var t=ho(e),n=[];for(var r in e)("constructor"!=r||!t&&Ee.call(e,r))&&n.push(r);return n}function Ar(e,t){return e<t}function Dr(e,t){var n=-1,i=Ls(e)?r(e.length):[];return tr(e,(function(e,r,o){i[++n]=t(e,r,o)})),i}function zr(e){var t=Qi(e);return 1==t.length&&t[0][2]?go(t[0][0],t[0][1]):function(n){return n===e||xr(n,e,t)}}function kr(e,t){return lo(e)&&po(t)?go(Ao(e),t):function(n){var r=va(n,e);return void 0===r&&r===t?ma(n,e):wr(t,r,3)}}function Nr(e,t,n,r,i){e!==t&&ar(t,(function(o,s){if(i||(i=new qn),Us(o))!function(e,t,n,r,i,o,s){var a=_o(e,n),u=_o(t,n),l=s.get(u);if(l)return void Hn(e,n,l);var c=o?o(a,u,n+"",e,t,s):void 0,f=void 0===c;if(f){var h=Os(u),p=!h&&qs(u),d=!h&&!p&&Qs(u);c=u,h||p||d?Os(a)?c=a:Ps(a)?c=yi(a):p?(f=!1,c=hi(u,!0)):d?(f=!1,c=di(u,!0)):c=[]:Gs(u)||js(u)?(c=a,js(a)?c=aa(a):Us(a)&&!Ws(a)||(c=oo(u))):f=!1}f&&(s.set(u,c),i(c,u,r,o,s),s.delete(u));Hn(e,n,c)}(e,t,s,n,Nr,r,i);else{var a=r?r(_o(e,s),o,s+"",e,t,i):void 0;void 0===a&&(a=o),Hn(e,s,a)}}),xa)}function Rr(e,t){var n=e.length;if(n)return ao(t+=t<0?n:0,n)?e[t]:void 0}function jr(e,t,n){t=t.length?pt(t,(function(e){return Os(e)?function(t){return hr(t,1===e.length?e[0]:e)}:e})):[Va];var r=-1;return t=pt(t,Nt(Ji())),function(e,t){var n=e.length;for(e.sort(t);n--;)e[n]=e[n].value;return e}(Dr(e,(function(e,n,i){return{criteria:pt(t,(function(t){return t(e)})),index:++r,value:e}})),(function(e,t){return function(e,t,n){var r=-1,i=e.criteria,o=t.criteria,s=i.length,a=n.length;for(;++r<s;){var u=gi(i[r],o[r]);if(u){if(r>=a)return u;var l=n[r];return u*("desc"==l?-1:1)}}return e.index-t.index}(e,t,n)}))}function Or(e,t,n){for(var r=-1,i=t.length,o={};++r<i;){var s=t[r],a=hr(e,s);n(a,s)&&Br(o,ui(s,e),a)}return o}function Fr(e,t,n,r){var i=r?xt:wt,o=-1,s=t.length,a=e;for(e===t&&(t=yi(t)),n&&(a=pt(e,Nt(n)));++o<s;)for(var u=0,l=t[o],c=n?n(l):l;(u=i(a,c,u,r))>-1;)a!==e&&Ge.call(a,u,1),Ge.call(e,u,1);return e}function Lr(e,t){for(var n=e?t.length:0,r=n-1;n--;){var i=t[n];if(n==r||i!==o){var o=i;ao(i)?Ge.call(e,i,1):ei(e,i)}}return e}function Pr(e,t){return e+en(fn()*(t-e+1))}function qr(e,t){var n="";if(!e||t<1||t>9007199254740991)return n;do{t%2&&(n+=e),(t=en(t/2))&&(e+=e)}while(t);return n}function Ir(e,t){return xo(vo(e,t,Va),e+"")}function Mr(e){return Mn(ka(e))}function Wr(e,t){var n=ka(e);return So(n,Yn(t,0,n.length))}function Br(e,t,n,r){if(!Us(e))return e;for(var i=-1,o=(t=ui(t,e)).length,s=o-1,a=e;null!=a&&++i<o;){var u=Ao(t[i]),l=n;if("__proto__"===u||"constructor"===u||"prototype"===u)return e;if(i!=s){var c=a[u];void 0===(l=r?r(c,u,a):void 0)&&(l=Us(c)?c:ao(t[i+1])?[]:{})}Un(a,u,l),a=a[u]}return e}var Hr=_n?function(e,t){return _n.set(e,t),e}:Va,Ur=yt?function(e,t){return yt(e,"toString",{configurable:!0,enumerable:!1,value:Ua(t),writable:!0})}:Va;function $r(e){return So(ka(e))}function Xr(e,t,n){var i=-1,o=e.length;t<0&&(t=-t>o?0:o+t),(n=n>o?o:n)<0&&(n+=o),o=t>n?0:n-t>>>0,t>>>=0;for(var s=r(o);++i<o;)s[i]=e[i+t];return s}function Vr(e,t){var n;return tr(e,(function(e,r,i){return!(n=t(e,r,i))})),!!n}function Gr(e,t,n){var r=0,i=null==e?r:e.length;if("number"==typeof t&&t==t&&i<=2147483647){for(;r<i;){var o=r+i>>>1,s=e[o];null!==s&&!Zs(s)&&(n?s<=t:s<t)?r=o+1:i=o}return i}return Kr(e,t,Va,n)}function Kr(e,t,n,r){var i=0,o=null==e?0:e.length;if(0===o)return 0;for(var s=(t=n(t))!=t,a=null===t,u=Zs(t),l=void 0===t;i<o;){var c=en((i+o)/2),f=n(e[c]),h=void 0!==f,p=null===f,d=f==f,g=Zs(f);if(s)var v=r||d;else v=l?d&&(r||h):a?d&&h&&(r||!p):u?d&&h&&!p&&(r||!g):!p&&!g&&(r?f<=t:f<t);v?i=c+1:o=c}return un(o,4294967294)}function Yr(e,t){for(var n=-1,r=e.length,i=0,o=[];++n<r;){var s=e[n],a=t?t(s):s;if(!n||!ks(a,u)){var u=a;o[i++]=0===s?0:s}}return o}function Jr(e){return"number"==typeof e?e:Zs(e)?NaN:+e}function Zr(e){if("string"==typeof e)return e;if(Os(e))return pt(e,Zr)+"";if(Zs(e))return Dn?Dn.call(e):"";var t=e+"";return"0"==t&&1/e==-1/0?"-0":t}function Qr(e,t,n){var r=-1,i=ft,o=e.length,s=!0,a=[],u=a;if(n)s=!1,i=ht;else if(o>=200){var l=t?null:qi(e);if(l)return Ut(l);s=!1,i=jt,u=new Pn}else u=t?[]:a;e:for(;++r<o;){var c=e[r],f=t?t(c):c;if(c=n||0!==c?c:0,s&&f==f){for(var h=u.length;h--;)if(u[h]===f)continue e;t&&u.push(f),a.push(c)}else i(u,f,n)||(u!==a&&u.push(f),a.push(c))}return a}function ei(e,t){return null==(e=mo(e,t=ui(t,e)))||delete e[Ao(Mo(t))]}function ti(e,t,n,r){return Br(e,t,n(hr(e,t)),r)}function ni(e,t,n,r){for(var i=e.length,o=r?i:-1;(r?o--:++o<i)&&t(e[o],o,e););return n?Xr(e,r?0:o,r?o+1:i):Xr(e,r?o+1:0,r?i:o)}function ri(e,t){var n=e;return n instanceof jn&&(n=n.value()),gt(t,(function(e,t){return t.func.apply(t.thisArg,dt([e],t.args))}),n)}function ii(e,t,n){var i=e.length;if(i<2)return i?Qr(e[0]):[];for(var o=-1,s=r(i);++o<i;)for(var a=e[o],u=-1;++u<i;)u!=o&&(s[o]=er(s[o]||a,e[u],t,n));return Qr(sr(s,1),t,n)}function oi(e,t,n){for(var r=-1,i=e.length,o=t.length,s={};++r<i;){var a=r<o?t[r]:void 0;n(s,e[r],a)}return s}function si(e){return Ps(e)?e:[]}function ai(e){return"function"==typeof e?e:Va}function ui(e,t){return Os(e)?e:lo(e,t)?[e]:Co(ua(e))}var li=Ir;function ci(e,t,n){var r=e.length;return n=void 0===n?r:n,!t&&n>=r?e:Xr(e,t,n)}var fi=Ct||function(e){return Ve.clearTimeout(e)};function hi(e,t){if(t)return e.slice();var n=e.length,r=Le?Le(n):new e.constructor(n);return e.copy(r),r}function pi(e){var t=new e.constructor(e.byteLength);return new Oe(t).set(new Oe(e)),t}function di(e,t){var n=t?pi(e.buffer):e.buffer;return new e.constructor(n,e.byteOffset,e.length)}function gi(e,t){if(e!==t){var n=void 0!==e,r=null===e,i=e==e,o=Zs(e),s=void 0!==t,a=null===t,u=t==t,l=Zs(t);if(!a&&!l&&!o&&e>t||o&&s&&u&&!a&&!l||r&&s&&u||!n&&u||!i)return 1;if(!r&&!o&&!l&&e<t||l&&n&&i&&!r&&!o||a&&n&&i||!s&&i||!u)return-1}return 0}function vi(e,t,n,i){for(var o=-1,s=e.length,a=n.length,u=-1,l=t.length,c=an(s-a,0),f=r(l+c),h=!i;++u<l;)f[u]=t[u];for(;++o<a;)(h||o<s)&&(f[n[o]]=e[o]);for(;c--;)f[u++]=e[o++];return f}function mi(e,t,n,i){for(var o=-1,s=e.length,a=-1,u=n.length,l=-1,c=t.length,f=an(s-u,0),h=r(f+c),p=!i;++o<f;)h[o]=e[o];for(var d=o;++l<c;)h[d+l]=t[l];for(;++a<u;)(p||o<s)&&(h[d+n[a]]=e[o++]);return h}function yi(e,t){var n=-1,i=e.length;for(t||(t=r(i));++n<i;)t[n]=e[n];return t}function _i(e,t,n,r){var i=!n;n||(n={});for(var o=-1,s=t.length;++o<s;){var a=t[o],u=r?r(n[a],e[a],a,n,e):void 0;void 0===u&&(u=e[a]),i?Gn(n,a,u):Un(n,a,u)}return n}function bi(e,t){return function(n,r){var i=Os(n)?st:Xn,o=t?t():{};return i(n,e,Ji(r,2),o)}}function wi(e){return Ir((function(t,n){var r=-1,i=n.length,o=i>1?n[i-1]:void 0,s=i>2?n[2]:void 0;for(o=e.length>3&&"function"==typeof o?(i--,o):void 0,s&&uo(n[0],n[1],s)&&(o=i<3?void 0:o,i=1),t=de(t);++r<i;){var a=n[r];a&&e(t,a,r,o)}return t}))}function xi(e,t){return function(n,r){if(null==n)return n;if(!Ls(n))return e(n,r);for(var i=n.length,o=t?i:-1,s=de(n);(t?o--:++o<i)&&!1!==r(s[o],o,s););return n}}function Ei(e){return function(t,n,r){for(var i=-1,o=de(t),s=r(t),a=s.length;a--;){var u=s[e?a:++i];if(!1===n(o[u],u,o))break}return t}}function Ti(e){return function(t){var n=Mt(t=ua(t))?Vt(t):void 0,r=n?n[0]:t.charAt(0),i=n?ci(n,1).join(""):t.slice(1);return r[e]()+i}}function Si(e){return function(t){return gt(Wa(ja(t).replace(Re,"")),e,"")}}function Ci(e){return function(){var t=arguments;switch(t.length){case 0:return new e;case 1:return new e(t[0]);case 2:return new e(t[0],t[1]);case 3:return new e(t[0],t[1],t[2]);case 4:return new e(t[0],t[1],t[2],t[3]);case 5:return new e(t[0],t[1],t[2],t[3],t[4]);case 6:return new e(t[0],t[1],t[2],t[3],t[4],t[5]);case 7:return new e(t[0],t[1],t[2],t[3],t[4],t[5],t[6])}var n=kn(e.prototype),r=e.apply(n,t);return Us(r)?r:n}}function Ai(e){return function(t,n,r){var i=de(t);if(!Ls(t)){var o=Ji(n,3);t=wa(t),n=function(e){return o(i[e],e,i)}}var s=e(t,n,r);return s>-1?i[o?t[s]:s]:void 0}}function Di(e){return $i((function(t){var n=t.length,r=n,i=Rn.prototype.thru;for(e&&t.reverse();r--;){var s=t[r];if("function"!=typeof s)throw new me(o);if(i&&!a&&"wrapper"==Ki(s))var a=new Rn([],!0)}for(r=a?r:n;++r<n;){var u=Ki(s=t[r]),l="wrapper"==u?Gi(s):void 0;a=l&&co(l[0])&&424==l[1]&&!l[4].length&&1==l[9]?a[Ki(l[0])].apply(a,l[3]):1==s.length&&co(s)?a[u]():a.thru(s)}return function(){var e=arguments,r=e[0];if(a&&1==e.length&&Os(r))return a.plant(r).value();for(var i=0,o=n?t[i].apply(this,e):r;++i<n;)o=t[i].call(this,o);return o}}))}function zi(e,t,n,i,o,s,a,u,l,c){var f=128&t,h=1&t,p=2&t,d=24&t,g=512&t,v=p?void 0:Ci(e);return function m(){for(var y=arguments.length,_=r(y),b=y;b--;)_[b]=arguments[b];if(d)var w=Yi(m),x=Lt(_,w);if(i&&(_=vi(_,i,o,d)),s&&(_=mi(_,s,a,d)),y-=x,d&&y<c){var E=Ht(_,w);return Li(e,t,zi,m.placeholder,n,_,E,u,l,c-y)}var T=h?n:this,S=p?T[e]:e;return y=_.length,u?_=yo(_,u):g&&y>1&&_.reverse(),f&&l<y&&(_.length=l),this&&this!==Ve&&this instanceof m&&(S=v||Ci(S)),S.apply(T,_)}}function ki(e,t){return function(n,r){return function(e,t,n,r){return lr(e,(function(e,i,o){t(r,n(e),i,o)})),r}(n,e,t(r),{})}}function Ni(e,t){return function(n,r){var i;if(void 0===n&&void 0===r)return t;if(void 0!==n&&(i=n),void 0!==r){if(void 0===i)return r;"string"==typeof n||"string"==typeof r?(n=Zr(n),r=Zr(r)):(n=Jr(n),r=Jr(r)),i=e(n,r)}return i}}function Ri(e){return $i((function(t){return t=pt(t,Nt(Ji())),Ir((function(n){var r=this;return e(t,(function(e){return ot(e,r,n)}))}))}))}function ji(e,t){var n=(t=void 0===t?" ":Zr(t)).length;if(n<2)return n?qr(t,e):t;var r=qr(t,Qt(e/Xt(t)));return Mt(t)?ci(Vt(r),0,e).join(""):r.slice(0,e)}function Oi(e){return function(t,n,i){return i&&"number"!=typeof i&&uo(t,n,i)&&(n=i=void 0),t=ra(t),void 0===n?(n=t,t=0):n=ra(n),function(e,t,n,i){for(var o=-1,s=an(Qt((t-e)/(n||1)),0),a=r(s);s--;)a[i?s:++o]=e,e+=n;return a}(t,n,i=void 0===i?t<n?1:-1:ra(i),e)}}function Fi(e){return function(t,n){return"string"==typeof t&&"string"==typeof n||(t=sa(t),n=sa(n)),e(t,n)}}function Li(e,t,n,r,i,o,s,a,u,l){var c=8&t;t|=c?32:64,4&(t&=~(c?64:32))||(t&=-4);var f=[e,t,i,c?o:void 0,c?s:void 0,c?void 0:o,c?void 0:s,a,u,l],h=n.apply(void 0,f);return co(e)&&bo(h,f),h.placeholder=r,Eo(h,e,t)}function Pi(e){var t=pe[e];return function(e,n){if(e=sa(e),(n=null==n?0:un(ia(n),292))&&rn(e)){var r=(ua(e)+"e").split("e");return+((r=(ua(t(r[0]+"e"+(+r[1]+n)))+"e").split("e"))[0]+"e"+(+r[1]-n))}return t(e)}}var qi=vn&&1/Ut(new vn([,-0]))[1]==1/0?function(e){return new vn(e)}:Za;function Ii(e){return function(t){var n=ro(t);return n==g?Wt(t):n==_?$t(t):function(e,t){return pt(t,(function(t){return[t,e[t]]}))}(t,e(t))}}function Mi(e,t,n,i,a,u,l,c){var f=2&t;if(!f&&"function"!=typeof e)throw new me(o);var h=i?i.length:0;if(h||(t&=-97,i=a=void 0),l=void 0===l?l:an(ia(l),0),c=void 0===c?c:ia(c),h-=a?a.length:0,64&t){var p=i,d=a;i=a=void 0}var g=f?void 0:Gi(e),v=[e,t,n,i,a,p,d,u,l,c];if(g&&function(e,t){var n=e[1],r=t[1],i=n|r,o=i<131,a=128==r&&8==n||128==r&&256==n&&e[7].length<=t[8]||384==r&&t[7].length<=t[8]&&8==n;if(!o&&!a)return e;1&r&&(e[2]=t[2],i|=1&n?0:4);var u=t[3];if(u){var l=e[3];e[3]=l?vi(l,u,t[4]):u,e[4]=l?Ht(e[3],s):t[4]}(u=t[5])&&(l=e[5],e[5]=l?mi(l,u,t[6]):u,e[6]=l?Ht(e[5],s):t[6]);(u=t[7])&&(e[7]=u);128&r&&(e[8]=null==e[8]?t[8]:un(e[8],t[8]));null==e[9]&&(e[9]=t[9]);e[0]=t[0],e[1]=i}(v,g),e=v[0],t=v[1],n=v[2],i=v[3],a=v[4],!(c=v[9]=void 0===v[9]?f?0:e.length:an(v[9]-h,0))&&24&t&&(t&=-25),t&&1!=t)m=8==t||16==t?function(e,t,n){var i=Ci(e);return function o(){for(var s=arguments.length,a=r(s),u=s,l=Yi(o);u--;)a[u]=arguments[u];var c=s<3&&a[0]!==l&&a[s-1]!==l?[]:Ht(a,l);if((s-=c.length)<n)return Li(e,t,zi,o.placeholder,void 0,a,c,void 0,void 0,n-s);var f=this&&this!==Ve&&this instanceof o?i:e;return ot(f,this,a)}}(e,t,c):32!=t&&33!=t||a.length?zi.apply(void 0,v):function(e,t,n,i){var o=1&t,s=Ci(e);return function t(){for(var a=-1,u=arguments.length,l=-1,c=i.length,f=r(c+u),h=this&&this!==Ve&&this instanceof t?s:e;++l<c;)f[l]=i[l];for(;u--;)f[l++]=arguments[++a];return ot(h,o?n:this,f)}}(e,t,n,i);else var m=function(e,t,n){var r=1&t,i=Ci(e);return function t(){var o=this&&this!==Ve&&this instanceof t?i:e;return o.apply(r?n:this,arguments)}}(e,t,n);return Eo((g?Hr:bo)(m,v),e,t)}function Wi(e,t,n,r){return void 0===e||ks(e,be[n])&&!Ee.call(r,n)?t:e}function Bi(e,t,n,r,i,o){return Us(e)&&Us(t)&&(o.set(t,e),Nr(e,t,void 0,Bi,o),o.delete(t)),e}function Hi(e){return Gs(e)?void 0:e}function Ui(e,t,n,r,i,o){var s=1&n,a=e.length,u=t.length;if(a!=u&&!(s&&u>a))return!1;var l=o.get(e),c=o.get(t);if(l&&c)return l==t&&c==e;var f=-1,h=!0,p=2&n?new Pn:void 0;for(o.set(e,t),o.set(t,e);++f<a;){var d=e[f],g=t[f];if(r)var v=s?r(g,d,f,t,e,o):r(d,g,f,e,t,o);if(void 0!==v){if(v)continue;h=!1;break}if(p){if(!mt(t,(function(e,t){if(!jt(p,t)&&(d===e||i(d,e,n,r,o)))return p.push(t)}))){h=!1;break}}else if(d!==g&&!i(d,g,n,r,o)){h=!1;break}}return o.delete(e),o.delete(t),h}function $i(e){return xo(vo(e,void 0,Fo),e+"")}function Xi(e){return pr(e,wa,to)}function Vi(e){return pr(e,xa,no)}var Gi=_n?function(e){return _n.get(e)}:Za;function Ki(e){for(var t=e.name+"",n=bn[t],r=Ee.call(bn,t)?n.length:0;r--;){var i=n[r],o=i.func;if(null==o||o==e)return i.name}return t}function Yi(e){return(Ee.call(zn,"placeholder")?zn:e).placeholder}function Ji(){var e=zn.iteratee||Ga;return e=e===Ga?Tr:e,arguments.length?e(arguments[0],arguments[1]):e}function Zi(e,t){var n,r,i=e.__data__;return("string"==(r=typeof(n=t))||"number"==r||"symbol"==r||"boolean"==r?"__proto__"!==n:null===n)?i["string"==typeof t?"string":"hash"]:i.map}function Qi(e){for(var t=wa(e),n=t.length;n--;){var r=t[n],i=e[r];t[n]=[r,i,po(i)]}return t}function eo(e,t){var n=function(e,t){return null==e?void 0:e[t]}(e,t);return Er(n)?n:void 0}var to=tn?function(e){return null==e?[]:(e=de(e),ct(tn(e),(function(t){return Xe.call(e,t)})))}:ou,no=tn?function(e){for(var t=[];e;)dt(t,to(e)),e=Be(e);return t}:ou,ro=dr;function io(e,t,n){for(var r=-1,i=(t=ui(t,e)).length,o=!1;++r<i;){var s=Ao(t[r]);if(!(o=null!=e&&n(e,s)))break;e=e[s]}return o||++r!=i?o:!!(i=null==e?0:e.length)&&Hs(i)&&ao(s,i)&&(Os(e)||js(e))}function oo(e){return"function"!=typeof e.constructor||ho(e)?{}:kn(Be(e))}function so(e){return Os(e)||js(e)||!!(Ke&&e&&e[Ke])}function ao(e,t){var n=typeof e;return!!(t=null==t?9007199254740991:t)&&("number"==n||"symbol"!=n&&ue.test(e))&&e>-1&&e%1==0&&e<t}function uo(e,t,n){if(!Us(n))return!1;var r=typeof t;return!!("number"==r?Ls(n)&&ao(t,n.length):"string"==r&&t in n)&&ks(n[t],e)}function lo(e,t){if(Os(e))return!1;var n=typeof e;return!("number"!=n&&"symbol"!=n&&"boolean"!=n&&null!=e&&!Zs(e))||(U.test(e)||!H.test(e)||null!=t&&e in de(t))}function co(e){var t=Ki(e),n=zn[t];if("function"!=typeof n||!(t in jn.prototype))return!1;if(e===n)return!0;var r=Gi(n);return!!r&&e===r[0]}(pn&&ro(new pn(new ArrayBuffer(1)))!=T||dn&&ro(new dn)!=g||gn&&"[object Promise]"!=ro(gn.resolve())||vn&&ro(new vn)!=_||mn&&ro(new mn)!=x)&&(ro=function(e){var t=dr(e),n=t==m?e.constructor:void 0,r=n?Do(n):"";if(r)switch(r){case wn:return T;case xn:return g;case En:return"[object Promise]";case Tn:return _;case Sn:return x}return t});var fo=we?Ws:su;function ho(e){var t=e&&e.constructor;return e===("function"==typeof t&&t.prototype||be)}function po(e){return e==e&&!Us(e)}function go(e,t){return function(n){return null!=n&&(n[e]===t&&(void 0!==t||e in de(n)))}}function vo(e,t,n){return t=an(void 0===t?e.length-1:t,0),function(){for(var i=arguments,o=-1,s=an(i.length-t,0),a=r(s);++o<s;)a[o]=i[t+o];o=-1;for(var u=r(t+1);++o<t;)u[o]=i[o];return u[t]=n(a),ot(e,this,u)}}function mo(e,t){return t.length<2?e:hr(e,Xr(t,0,-1))}function yo(e,t){for(var n=e.length,r=un(t.length,n),i=yi(e);r--;){var o=t[r];e[r]=ao(o,n)?i[o]:void 0}return e}function _o(e,t){if(("constructor"!==t||"function"!=typeof e[t])&&"__proto__"!=t)return e[t]}var bo=To(Hr),wo=Zt||function(e,t){return Ve.setTimeout(e,t)},xo=To(Ur);function Eo(e,t,n){var r=t+"";return xo(e,function(e,t){var n=t.length;if(!n)return e;var r=n-1;return t[r]=(n>1?"& ":"")+t[r],t=t.join(n>2?", ":" "),e.replace(Y,"{\n/* [wrapped with "+t+"] */\n")}(r,function(e,t){return at(a,(function(n){var r="_."+n[0];t&n[1]&&!ft(e,r)&&e.push(r)})),e.sort()}(function(e){var t=e.match(J);return t?t[1].split(Z):[]}(r),n)))}function To(e){var t=0,n=0;return function(){var r=ln(),i=16-(r-n);if(n=r,i>0){if(++t>=800)return arguments[0]}else t=0;return e.apply(void 0,arguments)}}function So(e,t){var n=-1,r=e.length,i=r-1;for(t=void 0===t?r:t;++n<t;){var o=Pr(n,i),s=e[o];e[o]=e[n],e[n]=s}return e.length=t,e}var Co=function(e){var t=Ts(e,(function(e){return 500===n.size&&n.clear(),e})),n=t.cache;return t}((function(e){var t=[];return 46===e.charCodeAt(0)&&t.push(""),e.replace($,(function(e,n,r,i){t.push(r?i.replace(te,"$1"):n||e)})),t}));function Ao(e){if("string"==typeof e||Zs(e))return e;var t=e+"";return"0"==t&&1/e==-1/0?"-0":t}function Do(e){if(null!=e){try{return xe.call(e)}catch(e){}try{return e+""}catch(e){}}return""}function zo(e){if(e instanceof jn)return e.clone();var t=new Rn(e.__wrapped__,e.__chain__);return t.__actions__=yi(e.__actions__),t.__index__=e.__index__,t.__values__=e.__values__,t}var ko=Ir((function(e,t){return Ps(e)?er(e,sr(t,1,Ps,!0)):[]})),No=Ir((function(e,t){var n=Mo(t);return Ps(n)&&(n=void 0),Ps(e)?er(e,sr(t,1,Ps,!0),Ji(n,2)):[]})),Ro=Ir((function(e,t){var n=Mo(t);return Ps(n)&&(n=void 0),Ps(e)?er(e,sr(t,1,Ps,!0),void 0,n):[]}));function jo(e,t,n){var r=null==e?0:e.length;if(!r)return-1;var i=null==n?0:ia(n);return i<0&&(i=an(r+i,0)),bt(e,Ji(t,3),i)}function Oo(e,t,n){var r=null==e?0:e.length;if(!r)return-1;var i=r-1;return void 0!==n&&(i=ia(n),i=n<0?an(r+i,0):un(i,r-1)),bt(e,Ji(t,3),i,!0)}function Fo(e){return(null==e?0:e.length)?sr(e,1):[]}function Lo(e){return e&&e.length?e[0]:void 0}var Po=Ir((function(e){var t=pt(e,si);return t.length&&t[0]===e[0]?yr(t):[]})),qo=Ir((function(e){var t=Mo(e),n=pt(e,si);return t===Mo(n)?t=void 0:n.pop(),n.length&&n[0]===e[0]?yr(n,Ji(t,2)):[]})),Io=Ir((function(e){var t=Mo(e),n=pt(e,si);return(t="function"==typeof t?t:void 0)&&n.pop(),n.length&&n[0]===e[0]?yr(n,void 0,t):[]}));function Mo(e){var t=null==e?0:e.length;return t?e[t-1]:void 0}var Wo=Ir(Bo);function Bo(e,t){return e&&e.length&&t&&t.length?Fr(e,t):e}var Ho=$i((function(e,t){var n=null==e?0:e.length,r=Kn(e,t);return Lr(e,pt(t,(function(e){return ao(e,n)?+e:e})).sort(gi)),r}));function Uo(e){return null==e?e:hn.call(e)}var $o=Ir((function(e){return Qr(sr(e,1,Ps,!0))})),Xo=Ir((function(e){var t=Mo(e);return Ps(t)&&(t=void 0),Qr(sr(e,1,Ps,!0),Ji(t,2))})),Vo=Ir((function(e){var t=Mo(e);return t="function"==typeof t?t:void 0,Qr(sr(e,1,Ps,!0),void 0,t)}));function Go(e){if(!e||!e.length)return[];var t=0;return e=ct(e,(function(e){if(Ps(e))return t=an(e.length,t),!0})),zt(t,(function(t){return pt(e,St(t))}))}function Ko(e,t){if(!e||!e.length)return[];var n=Go(e);return null==t?n:pt(n,(function(e){return ot(t,void 0,e)}))}var Yo=Ir((function(e,t){return Ps(e)?er(e,t):[]})),Jo=Ir((function(e){return ii(ct(e,Ps))})),Zo=Ir((function(e){var t=Mo(e);return Ps(t)&&(t=void 0),ii(ct(e,Ps),Ji(t,2))})),Qo=Ir((function(e){var t=Mo(e);return t="function"==typeof t?t:void 0,ii(ct(e,Ps),void 0,t)})),es=Ir(Go);var ts=Ir((function(e){var t=e.length,n=t>1?e[t-1]:void 0;return n="function"==typeof n?(e.pop(),n):void 0,Ko(e,n)}));function ns(e){var t=zn(e);return t.__chain__=!0,t}function rs(e,t){return t(e)}var is=$i((function(e){var t=e.length,n=t?e[0]:0,r=this.__wrapped__,i=function(t){return Kn(t,e)};return!(t>1||this.__actions__.length)&&r instanceof jn&&ao(n)?((r=r.slice(n,+n+(t?1:0))).__actions__.push({func:rs,args:[i],thisArg:void 0}),new Rn(r,this.__chain__).thru((function(e){return t&&!e.length&&e.push(void 0),e}))):this.thru(i)}));var os=bi((function(e,t,n){Ee.call(e,n)?++e[n]:Gn(e,n,1)}));var ss=Ai(jo),as=Ai(Oo);function us(e,t){return(Os(e)?at:tr)(e,Ji(t,3))}function ls(e,t){return(Os(e)?ut:nr)(e,Ji(t,3))}var cs=bi((function(e,t,n){Ee.call(e,n)?e[n].push(t):Gn(e,n,[t])}));var fs=Ir((function(e,t,n){var i=-1,o="function"==typeof t,s=Ls(e)?r(e.length):[];return tr(e,(function(e){s[++i]=o?ot(t,e,n):_r(e,t,n)})),s})),hs=bi((function(e,t,n){Gn(e,n,t)}));function ps(e,t){return(Os(e)?pt:Dr)(e,Ji(t,3))}var ds=bi((function(e,t,n){e[n?0:1].push(t)}),(function(){return[[],[]]}));var gs=Ir((function(e,t){if(null==e)return[];var n=t.length;return n>1&&uo(e,t[0],t[1])?t=[]:n>2&&uo(t[0],t[1],t[2])&&(t=[t[0]]),jr(e,sr(t,1),[])})),vs=Jt||function(){return Ve.Date.now()};function ms(e,t,n){return t=n?void 0:t,Mi(e,128,void 0,void 0,void 0,void 0,t=e&&null==t?e.length:t)}function ys(e,t){var n;if("function"!=typeof t)throw new me(o);return e=ia(e),function(){return--e>0&&(n=t.apply(this,arguments)),e<=1&&(t=void 0),n}}var _s=Ir((function(e,t,n){var r=1;if(n.length){var i=Ht(n,Yi(_s));r|=32}return Mi(e,r,t,n,i)})),bs=Ir((function(e,t,n){var r=3;if(n.length){var i=Ht(n,Yi(bs));r|=32}return Mi(t,r,e,n,i)}));function ws(e,t,n){var r,i,s,a,u,l,c=0,f=!1,h=!1,p=!0;if("function"!=typeof e)throw new me(o);function d(t){var n=r,o=i;return r=i=void 0,c=t,a=e.apply(o,n)}function g(e){return c=e,u=wo(m,t),f?d(e):a}function v(e){var n=e-l;return void 0===l||n>=t||n<0||h&&e-c>=s}function m(){var e=vs();if(v(e))return y(e);u=wo(m,function(e){var n=t-(e-l);return h?un(n,s-(e-c)):n}(e))}function y(e){return u=void 0,p&&r?d(e):(r=i=void 0,a)}function _(){var e=vs(),n=v(e);if(r=arguments,i=this,l=e,n){if(void 0===u)return g(l);if(h)return fi(u),u=wo(m,t),d(l)}return void 0===u&&(u=wo(m,t)),a}return t=sa(t)||0,Us(n)&&(f=!!n.leading,s=(h="maxWait"in n)?an(sa(n.maxWait)||0,t):s,p="trailing"in n?!!n.trailing:p),_.cancel=function(){void 0!==u&&fi(u),c=0,r=l=i=u=void 0},_.flush=function(){return void 0===u?a:y(vs())},_}var xs=Ir((function(e,t){return Qn(e,1,t)})),Es=Ir((function(e,t,n){return Qn(e,sa(t)||0,n)}));function Ts(e,t){if("function"!=typeof e||null!=t&&"function"!=typeof t)throw new me(o);var n=function(){var r=arguments,i=t?t.apply(this,r):r[0],o=n.cache;if(o.has(i))return o.get(i);var s=e.apply(this,r);return n.cache=o.set(i,s)||o,s};return n.cache=new(Ts.Cache||Ln),n}function Ss(e){if("function"!=typeof e)throw new me(o);return function(){var t=arguments;switch(t.length){case 0:return!e.call(this);case 1:return!e.call(this,t[0]);case 2:return!e.call(this,t[0],t[1]);case 3:return!e.call(this,t[0],t[1],t[2])}return!e.apply(this,t)}}Ts.Cache=Ln;var Cs=li((function(e,t){var n=(t=1==t.length&&Os(t[0])?pt(t[0],Nt(Ji())):pt(sr(t,1),Nt(Ji()))).length;return Ir((function(r){for(var i=-1,o=un(r.length,n);++i<o;)r[i]=t[i].call(this,r[i]);return ot(e,this,r)}))})),As=Ir((function(e,t){return Mi(e,32,void 0,t,Ht(t,Yi(As)))})),Ds=Ir((function(e,t){return Mi(e,64,void 0,t,Ht(t,Yi(Ds)))})),zs=$i((function(e,t){return Mi(e,256,void 0,void 0,void 0,t)}));function ks(e,t){return e===t||e!=e&&t!=t}var Ns=Fi(gr),Rs=Fi((function(e,t){return e>=t})),js=br(function(){return arguments}())?br:function(e){return $s(e)&&Ee.call(e,"callee")&&!Xe.call(e,"callee")},Os=r.isArray,Fs=Qe?Nt(Qe):function(e){return $s(e)&&dr(e)==E};function Ls(e){return null!=e&&Hs(e.length)&&!Ws(e)}function Ps(e){return $s(e)&&Ls(e)}var qs=nn||su,Is=et?Nt(et):function(e){return $s(e)&&dr(e)==f};function Ms(e){if(!$s(e))return!1;var t=dr(e);return t==h||"[object DOMException]"==t||"string"==typeof e.message&&"string"==typeof e.name&&!Gs(e)}function Ws(e){if(!Us(e))return!1;var t=dr(e);return t==p||t==d||"[object AsyncFunction]"==t||"[object Proxy]"==t}function Bs(e){return"number"==typeof e&&e==ia(e)}function Hs(e){return"number"==typeof e&&e>-1&&e%1==0&&e<=9007199254740991}function Us(e){var t=typeof e;return null!=e&&("object"==t||"function"==t)}function $s(e){return null!=e&&"object"==typeof e}var Xs=tt?Nt(tt):function(e){return $s(e)&&ro(e)==g};function Vs(e){return"number"==typeof e||$s(e)&&dr(e)==v}function Gs(e){if(!$s(e)||dr(e)!=m)return!1;var t=Be(e);if(null===t)return!0;var n=Ee.call(t,"constructor")&&t.constructor;return"function"==typeof n&&n instanceof n&&xe.call(n)==Ae}var Ks=nt?Nt(nt):function(e){return $s(e)&&dr(e)==y};var Ys=rt?Nt(rt):function(e){return $s(e)&&ro(e)==_};function Js(e){return"string"==typeof e||!Os(e)&&$s(e)&&dr(e)==b}function Zs(e){return"symbol"==typeof e||$s(e)&&dr(e)==w}var Qs=it?Nt(it):function(e){return $s(e)&&Hs(e.length)&&!!Me[dr(e)]};var ea=Fi(Ar),ta=Fi((function(e,t){return e<=t}));function na(e){if(!e)return[];if(Ls(e))return Js(e)?Vt(e):yi(e);if(Je&&e[Je])return function(e){for(var t,n=[];!(t=e.next()).done;)n.push(t.value);return n}(e[Je]());var t=ro(e);return(t==g?Wt:t==_?Ut:ka)(e)}function ra(e){return e?(e=sa(e))===1/0||e===-1/0?17976931348623157e292*(e<0?-1:1):e==e?e:0:0===e?e:0}function ia(e){var t=ra(e),n=t%1;return t==t?n?t-n:t:0}function oa(e){return e?Yn(ia(e),0,4294967295):0}function sa(e){if("number"==typeof e)return e;if(Zs(e))return NaN;if(Us(e)){var t="function"==typeof e.valueOf?e.valueOf():e;e=Us(t)?t+"":t}if("string"!=typeof e)return 0===e?e:+e;e=kt(e);var n=oe.test(e);return n||ae.test(e)?Ue(e.slice(2),n?2:8):ie.test(e)?NaN:+e}function aa(e){return _i(e,xa(e))}function ua(e){return null==e?"":Zr(e)}var la=wi((function(e,t){if(ho(t)||Ls(t))_i(t,wa(t),e);else for(var n in t)Ee.call(t,n)&&Un(e,n,t[n])})),ca=wi((function(e,t){_i(t,xa(t),e)})),fa=wi((function(e,t,n,r){_i(t,xa(t),e,r)})),ha=wi((function(e,t,n,r){_i(t,wa(t),e,r)})),pa=$i(Kn);var da=Ir((function(e,t){e=de(e);var n=-1,r=t.length,i=r>2?t[2]:void 0;for(i&&uo(t[0],t[1],i)&&(r=1);++n<r;)for(var o=t[n],s=xa(o),a=-1,u=s.length;++a<u;){var l=s[a],c=e[l];(void 0===c||ks(c,be[l])&&!Ee.call(e,l))&&(e[l]=o[l])}return e})),ga=Ir((function(e){return e.push(void 0,Bi),ot(Ta,void 0,e)}));function va(e,t,n){var r=null==e?void 0:hr(e,t);return void 0===r?n:r}function ma(e,t){return null!=e&&io(e,t,mr)}var ya=ki((function(e,t,n){null!=t&&"function"!=typeof t.toString&&(t=Ce.call(t)),e[t]=n}),Ua(Va)),_a=ki((function(e,t,n){null!=t&&"function"!=typeof t.toString&&(t=Ce.call(t)),Ee.call(e,t)?e[t].push(n):e[t]=[n]}),Ji),ba=Ir(_r);function wa(e){return Ls(e)?In(e):Sr(e)}function xa(e){return Ls(e)?In(e,!0):Cr(e)}var Ea=wi((function(e,t,n){Nr(e,t,n)})),Ta=wi((function(e,t,n,r){Nr(e,t,n,r)})),Sa=$i((function(e,t){var n={};if(null==e)return n;var r=!1;t=pt(t,(function(t){return t=ui(t,e),r||(r=t.length>1),t})),_i(e,Vi(e),n),r&&(n=Jn(n,7,Hi));for(var i=t.length;i--;)ei(n,t[i]);return n}));var Ca=$i((function(e,t){return null==e?{}:function(e,t){return Or(e,t,(function(t,n){return ma(e,n)}))}(e,t)}));function Aa(e,t){if(null==e)return{};var n=pt(Vi(e),(function(e){return[e]}));return t=Ji(t),Or(e,n,(function(e,n){return t(e,n[0])}))}var Da=Ii(wa),za=Ii(xa);function ka(e){return null==e?[]:Rt(e,wa(e))}var Na=Si((function(e,t,n){return t=t.toLowerCase(),e+(n?Ra(t):t)}));function Ra(e){return Ma(ua(e).toLowerCase())}function ja(e){return(e=ua(e))&&e.replace(le,Pt).replace(je,"")}var Oa=Si((function(e,t,n){return e+(n?"-":"")+t.toLowerCase()})),Fa=Si((function(e,t,n){return e+(n?" ":"")+t.toLowerCase()})),La=Ti("toLowerCase");var Pa=Si((function(e,t,n){return e+(n?"_":"")+t.toLowerCase()}));var qa=Si((function(e,t,n){return e+(n?" ":"")+Ma(t)}));var Ia=Si((function(e,t,n){return e+(n?" ":"")+t.toUpperCase()})),Ma=Ti("toUpperCase");function Wa(e,t,n){return e=ua(e),void 0===(t=n?void 0:t)?function(e){return Pe.test(e)}(e)?function(e){return e.match(Fe)||[]}(e):function(e){return e.match(Q)||[]}(e):e.match(t)||[]}var Ba=Ir((function(e,t){try{return ot(e,void 0,t)}catch(e){return Ms(e)?e:new K(e)}})),Ha=$i((function(e,t){return at(t,(function(t){t=Ao(t),Gn(e,t,_s(e[t],e))})),e}));function Ua(e){return function(){return e}}var $a=Di(),Xa=Di(!0);function Va(e){return e}function Ga(e){return Tr("function"==typeof e?e:Jn(e,1))}var Ka=Ir((function(e,t){return function(n){return _r(n,e,t)}})),Ya=Ir((function(e,t){return function(n){return _r(e,n,t)}}));function Ja(e,t,n){var r=wa(t),i=fr(t,r);null!=n||Us(t)&&(i.length||!r.length)||(n=t,t=e,e=this,i=fr(t,wa(t)));var o=!(Us(n)&&"chain"in n&&!n.chain),s=Ws(e);return at(i,(function(n){var r=t[n];e[n]=r,s&&(e.prototype[n]=function(){var t=this.__chain__;if(o||t){var n=e(this.__wrapped__),i=n.__actions__=yi(this.__actions__);return i.push({func:r,args:arguments,thisArg:e}),n.__chain__=t,n}return r.apply(e,dt([this.value()],arguments))})})),e}function Za(){}var Qa=Ri(pt),eu=Ri(lt),tu=Ri(mt);function nu(e){return lo(e)?St(Ao(e)):function(e){return function(t){return hr(t,e)}}(e)}var ru=Oi(),iu=Oi(!0);function ou(){return[]}function su(){return!1}var au=Ni((function(e,t){return e+t}),0),uu=Pi("ceil"),lu=Ni((function(e,t){return e/t}),1),cu=Pi("floor");var fu,hu=Ni((function(e,t){return e*t}),1),pu=Pi("round"),du=Ni((function(e,t){return e-t}),0);return zn.after=function(e,t){if("function"!=typeof t)throw new me(o);return e=ia(e),function(){if(--e<1)return t.apply(this,arguments)}},zn.ary=ms,zn.assign=la,zn.assignIn=ca,zn.assignInWith=fa,zn.assignWith=ha,zn.at=pa,zn.before=ys,zn.bind=_s,zn.bindAll=Ha,zn.bindKey=bs,zn.castArray=function(){if(!arguments.length)return[];var e=arguments[0];return Os(e)?e:[e]},zn.chain=ns,zn.chunk=function(e,t,n){t=(n?uo(e,t,n):void 0===t)?1:an(ia(t),0);var i=null==e?0:e.length;if(!i||t<1)return[];for(var o=0,s=0,a=r(Qt(i/t));o<i;)a[s++]=Xr(e,o,o+=t);return a},zn.compact=function(e){for(var t=-1,n=null==e?0:e.length,r=0,i=[];++t<n;){var o=e[t];o&&(i[r++]=o)}return i},zn.concat=function(){var e=arguments.length;if(!e)return[];for(var t=r(e-1),n=arguments[0],i=e;i--;)t[i-1]=arguments[i];return dt(Os(n)?yi(n):[n],sr(t,1))},zn.cond=function(e){var t=null==e?0:e.length,n=Ji();return e=t?pt(e,(function(e){if("function"!=typeof e[1])throw new me(o);return[n(e[0]),e[1]]})):[],Ir((function(n){for(var r=-1;++r<t;){var i=e[r];if(ot(i[0],this,n))return ot(i[1],this,n)}}))},zn.conforms=function(e){return function(e){var t=wa(e);return function(n){return Zn(n,e,t)}}(Jn(e,1))},zn.constant=Ua,zn.countBy=os,zn.create=function(e,t){var n=kn(e);return null==t?n:Vn(n,t)},zn.curry=function e(t,n,r){var i=Mi(t,8,void 0,void 0,void 0,void 0,void 0,n=r?void 0:n);return i.placeholder=e.placeholder,i},zn.curryRight=function e(t,n,r){var i=Mi(t,16,void 0,void 0,void 0,void 0,void 0,n=r?void 0:n);return i.placeholder=e.placeholder,i},zn.debounce=ws,zn.defaults=da,zn.defaultsDeep=ga,zn.defer=xs,zn.delay=Es,zn.difference=ko,zn.differenceBy=No,zn.differenceWith=Ro,zn.drop=function(e,t,n){var r=null==e?0:e.length;return r?Xr(e,(t=n||void 0===t?1:ia(t))<0?0:t,r):[]},zn.dropRight=function(e,t,n){var r=null==e?0:e.length;return r?Xr(e,0,(t=r-(t=n||void 0===t?1:ia(t)))<0?0:t):[]},zn.dropRightWhile=function(e,t){return e&&e.length?ni(e,Ji(t,3),!0,!0):[]},zn.dropWhile=function(e,t){return e&&e.length?ni(e,Ji(t,3),!0):[]},zn.fill=function(e,t,n,r){var i=null==e?0:e.length;return i?(n&&"number"!=typeof n&&uo(e,t,n)&&(n=0,r=i),function(e,t,n,r){var i=e.length;for((n=ia(n))<0&&(n=-n>i?0:i+n),(r=void 0===r||r>i?i:ia(r))<0&&(r+=i),r=n>r?0:oa(r);n<r;)e[n++]=t;return e}(e,t,n,r)):[]},zn.filter=function(e,t){return(Os(e)?ct:or)(e,Ji(t,3))},zn.flatMap=function(e,t){return sr(ps(e,t),1)},zn.flatMapDeep=function(e,t){return sr(ps(e,t),1/0)},zn.flatMapDepth=function(e,t,n){return n=void 0===n?1:ia(n),sr(ps(e,t),n)},zn.flatten=Fo,zn.flattenDeep=function(e){return(null==e?0:e.length)?sr(e,1/0):[]},zn.flattenDepth=function(e,t){return(null==e?0:e.length)?sr(e,t=void 0===t?1:ia(t)):[]},zn.flip=function(e){return Mi(e,512)},zn.flow=$a,zn.flowRight=Xa,zn.fromPairs=function(e){for(var t=-1,n=null==e?0:e.length,r={};++t<n;){var i=e[t];r[i[0]]=i[1]}return r},zn.functions=function(e){return null==e?[]:fr(e,wa(e))},zn.functionsIn=function(e){return null==e?[]:fr(e,xa(e))},zn.groupBy=cs,zn.initial=function(e){return(null==e?0:e.length)?Xr(e,0,-1):[]},zn.intersection=Po,zn.intersectionBy=qo,zn.intersectionWith=Io,zn.invert=ya,zn.invertBy=_a,zn.invokeMap=fs,zn.iteratee=Ga,zn.keyBy=hs,zn.keys=wa,zn.keysIn=xa,zn.map=ps,zn.mapKeys=function(e,t){var n={};return t=Ji(t,3),lr(e,(function(e,r,i){Gn(n,t(e,r,i),e)})),n},zn.mapValues=function(e,t){var n={};return t=Ji(t,3),lr(e,(function(e,r,i){Gn(n,r,t(e,r,i))})),n},zn.matches=function(e){return zr(Jn(e,1))},zn.matchesProperty=function(e,t){return kr(e,Jn(t,1))},zn.memoize=Ts,zn.merge=Ea,zn.mergeWith=Ta,zn.method=Ka,zn.methodOf=Ya,zn.mixin=Ja,zn.negate=Ss,zn.nthArg=function(e){return e=ia(e),Ir((function(t){return Rr(t,e)}))},zn.omit=Sa,zn.omitBy=function(e,t){return Aa(e,Ss(Ji(t)))},zn.once=function(e){return ys(2,e)},zn.orderBy=function(e,t,n,r){return null==e?[]:(Os(t)||(t=null==t?[]:[t]),Os(n=r?void 0:n)||(n=null==n?[]:[n]),jr(e,t,n))},zn.over=Qa,zn.overArgs=Cs,zn.overEvery=eu,zn.overSome=tu,zn.partial=As,zn.partialRight=Ds,zn.partition=ds,zn.pick=Ca,zn.pickBy=Aa,zn.property=nu,zn.propertyOf=function(e){return function(t){return null==e?void 0:hr(e,t)}},zn.pull=Wo,zn.pullAll=Bo,zn.pullAllBy=function(e,t,n){return e&&e.length&&t&&t.length?Fr(e,t,Ji(n,2)):e},zn.pullAllWith=function(e,t,n){return e&&e.length&&t&&t.length?Fr(e,t,void 0,n):e},zn.pullAt=Ho,zn.range=ru,zn.rangeRight=iu,zn.rearg=zs,zn.reject=function(e,t){return(Os(e)?ct:or)(e,Ss(Ji(t,3)))},zn.remove=function(e,t){var n=[];if(!e||!e.length)return n;var r=-1,i=[],o=e.length;for(t=Ji(t,3);++r<o;){var s=e[r];t(s,r,e)&&(n.push(s),i.push(r))}return Lr(e,i),n},zn.rest=function(e,t){if("function"!=typeof e)throw new me(o);return Ir(e,t=void 0===t?t:ia(t))},zn.reverse=Uo,zn.sampleSize=function(e,t,n){return t=(n?uo(e,t,n):void 0===t)?1:ia(t),(Os(e)?Wn:Wr)(e,t)},zn.set=function(e,t,n){return null==e?e:Br(e,t,n)},zn.setWith=function(e,t,n,r){return r="function"==typeof r?r:void 0,null==e?e:Br(e,t,n,r)},zn.shuffle=function(e){return(Os(e)?Bn:$r)(e)},zn.slice=function(e,t,n){var r=null==e?0:e.length;return r?(n&&"number"!=typeof n&&uo(e,t,n)?(t=0,n=r):(t=null==t?0:ia(t),n=void 0===n?r:ia(n)),Xr(e,t,n)):[]},zn.sortBy=gs,zn.sortedUniq=function(e){return e&&e.length?Yr(e):[]},zn.sortedUniqBy=function(e,t){return e&&e.length?Yr(e,Ji(t,2)):[]},zn.split=function(e,t,n){return n&&"number"!=typeof n&&uo(e,t,n)&&(t=n=void 0),(n=void 0===n?4294967295:n>>>0)?(e=ua(e))&&("string"==typeof t||null!=t&&!Ks(t))&&!(t=Zr(t))&&Mt(e)?ci(Vt(e),0,n):e.split(t,n):[]},zn.spread=function(e,t){if("function"!=typeof e)throw new me(o);return t=null==t?0:an(ia(t),0),Ir((function(n){var r=n[t],i=ci(n,0,t);return r&&dt(i,r),ot(e,this,i)}))},zn.tail=function(e){var t=null==e?0:e.length;return t?Xr(e,1,t):[]},zn.take=function(e,t,n){return e&&e.length?Xr(e,0,(t=n||void 0===t?1:ia(t))<0?0:t):[]},zn.takeRight=function(e,t,n){var r=null==e?0:e.length;return r?Xr(e,(t=r-(t=n||void 0===t?1:ia(t)))<0?0:t,r):[]},zn.takeRightWhile=function(e,t){return e&&e.length?ni(e,Ji(t,3),!1,!0):[]},zn.takeWhile=function(e,t){return e&&e.length?ni(e,Ji(t,3)):[]},zn.tap=function(e,t){return t(e),e},zn.throttle=function(e,t,n){var r=!0,i=!0;if("function"!=typeof e)throw new me(o);return Us(n)&&(r="leading"in n?!!n.leading:r,i="trailing"in n?!!n.trailing:i),ws(e,t,{leading:r,maxWait:t,trailing:i})},zn.thru=rs,zn.toArray=na,zn.toPairs=Da,zn.toPairsIn=za,zn.toPath=function(e){return Os(e)?pt(e,Ao):Zs(e)?[e]:yi(Co(ua(e)))},zn.toPlainObject=aa,zn.transform=function(e,t,n){var r=Os(e),i=r||qs(e)||Qs(e);if(t=Ji(t,4),null==n){var o=e&&e.constructor;n=i?r?new o:[]:Us(e)&&Ws(o)?kn(Be(e)):{}}return(i?at:lr)(e,(function(e,r,i){return t(n,e,r,i)})),n},zn.unary=function(e){return ms(e,1)},zn.union=$o,zn.unionBy=Xo,zn.unionWith=Vo,zn.uniq=function(e){return e&&e.length?Qr(e):[]},zn.uniqBy=function(e,t){return e&&e.length?Qr(e,Ji(t,2)):[]},zn.uniqWith=function(e,t){return t="function"==typeof t?t:void 0,e&&e.length?Qr(e,void 0,t):[]},zn.unset=function(e,t){return null==e||ei(e,t)},zn.unzip=Go,zn.unzipWith=Ko,zn.update=function(e,t,n){return null==e?e:ti(e,t,ai(n))},zn.updateWith=function(e,t,n,r){return r="function"==typeof r?r:void 0,null==e?e:ti(e,t,ai(n),r)},zn.values=ka,zn.valuesIn=function(e){return null==e?[]:Rt(e,xa(e))},zn.without=Yo,zn.words=Wa,zn.wrap=function(e,t){return As(ai(t),e)},zn.xor=Jo,zn.xorBy=Zo,zn.xorWith=Qo,zn.zip=es,zn.zipObject=function(e,t){return oi(e||[],t||[],Un)},zn.zipObjectDeep=function(e,t){return oi(e||[],t||[],Br)},zn.zipWith=ts,zn.entries=Da,zn.entriesIn=za,zn.extend=ca,zn.extendWith=fa,Ja(zn,zn),zn.add=au,zn.attempt=Ba,zn.camelCase=Na,zn.capitalize=Ra,zn.ceil=uu,zn.clamp=function(e,t,n){return void 0===n&&(n=t,t=void 0),void 0!==n&&(n=(n=sa(n))==n?n:0),void 0!==t&&(t=(t=sa(t))==t?t:0),Yn(sa(e),t,n)},zn.clone=function(e){return Jn(e,4)},zn.cloneDeep=function(e){return Jn(e,5)},zn.cloneDeepWith=function(e,t){return Jn(e,5,t="function"==typeof t?t:void 0)},zn.cloneWith=function(e,t){return Jn(e,4,t="function"==typeof t?t:void 0)},zn.conformsTo=function(e,t){return null==t||Zn(e,t,wa(t))},zn.deburr=ja,zn.defaultTo=function(e,t){return null==e||e!=e?t:e},zn.divide=lu,zn.endsWith=function(e,t,n){e=ua(e),t=Zr(t);var r=e.length,i=n=void 0===n?r:Yn(ia(n),0,r);return(n-=t.length)>=0&&e.slice(n,i)==t},zn.eq=ks,zn.escape=function(e){return(e=ua(e))&&I.test(e)?e.replace(P,qt):e},zn.escapeRegExp=function(e){return(e=ua(e))&&V.test(e)?e.replace(X,"\\$&"):e},zn.every=function(e,t,n){var r=Os(e)?lt:rr;return n&&uo(e,t,n)&&(t=void 0),r(e,Ji(t,3))},zn.find=ss,zn.findIndex=jo,zn.findKey=function(e,t){return _t(e,Ji(t,3),lr)},zn.findLast=as,zn.findLastIndex=Oo,zn.findLastKey=function(e,t){return _t(e,Ji(t,3),cr)},zn.floor=cu,zn.forEach=us,zn.forEachRight=ls,zn.forIn=function(e,t){return null==e?e:ar(e,Ji(t,3),xa)},zn.forInRight=function(e,t){return null==e?e:ur(e,Ji(t,3),xa)},zn.forOwn=function(e,t){return e&&lr(e,Ji(t,3))},zn.forOwnRight=function(e,t){return e&&cr(e,Ji(t,3))},zn.get=va,zn.gt=Ns,zn.gte=Rs,zn.has=function(e,t){return null!=e&&io(e,t,vr)},zn.hasIn=ma,zn.head=Lo,zn.identity=Va,zn.includes=function(e,t,n,r){e=Ls(e)?e:ka(e),n=n&&!r?ia(n):0;var i=e.length;return n<0&&(n=an(i+n,0)),Js(e)?n<=i&&e.indexOf(t,n)>-1:!!i&&wt(e,t,n)>-1},zn.indexOf=function(e,t,n){var r=null==e?0:e.length;if(!r)return-1;var i=null==n?0:ia(n);return i<0&&(i=an(r+i,0)),wt(e,t,i)},zn.inRange=function(e,t,n){return t=ra(t),void 0===n?(n=t,t=0):n=ra(n),function(e,t,n){return e>=un(t,n)&&e<an(t,n)}(e=sa(e),t,n)},zn.invoke=ba,zn.isArguments=js,zn.isArray=Os,zn.isArrayBuffer=Fs,zn.isArrayLike=Ls,zn.isArrayLikeObject=Ps,zn.isBoolean=function(e){return!0===e||!1===e||$s(e)&&dr(e)==c},zn.isBuffer=qs,zn.isDate=Is,zn.isElement=function(e){return $s(e)&&1===e.nodeType&&!Gs(e)},zn.isEmpty=function(e){if(null==e)return!0;if(Ls(e)&&(Os(e)||"string"==typeof e||"function"==typeof e.splice||qs(e)||Qs(e)||js(e)))return!e.length;var t=ro(e);if(t==g||t==_)return!e.size;if(ho(e))return!Sr(e).length;for(var n in e)if(Ee.call(e,n))return!1;return!0},zn.isEqual=function(e,t){return wr(e,t)},zn.isEqualWith=function(e,t,n){var r=(n="function"==typeof n?n:void 0)?n(e,t):void 0;return void 0===r?wr(e,t,void 0,n):!!r},zn.isError=Ms,zn.isFinite=function(e){return"number"==typeof e&&rn(e)},zn.isFunction=Ws,zn.isInteger=Bs,zn.isLength=Hs,zn.isMap=Xs,zn.isMatch=function(e,t){return e===t||xr(e,t,Qi(t))},zn.isMatchWith=function(e,t,n){return n="function"==typeof n?n:void 0,xr(e,t,Qi(t),n)},zn.isNaN=function(e){return Vs(e)&&e!=+e},zn.isNative=function(e){if(fo(e))throw new K("Unsupported core-js use. Try https://npms.io/search?q=ponyfill.");return Er(e)},zn.isNil=function(e){return null==e},zn.isNull=function(e){return null===e},zn.isNumber=Vs,zn.isObject=Us,zn.isObjectLike=$s,zn.isPlainObject=Gs,zn.isRegExp=Ks,zn.isSafeInteger=function(e){return Bs(e)&&e>=-9007199254740991&&e<=9007199254740991},zn.isSet=Ys,zn.isString=Js,zn.isSymbol=Zs,zn.isTypedArray=Qs,zn.isUndefined=function(e){return void 0===e},zn.isWeakMap=function(e){return $s(e)&&ro(e)==x},zn.isWeakSet=function(e){return $s(e)&&"[object WeakSet]"==dr(e)},zn.join=function(e,t){return null==e?"":on.call(e,t)},zn.kebabCase=Oa,zn.last=Mo,zn.lastIndexOf=function(e,t,n){var r=null==e?0:e.length;if(!r)return-1;var i=r;return void 0!==n&&(i=(i=ia(n))<0?an(r+i,0):un(i,r-1)),t==t?function(e,t,n){for(var r=n+1;r--;)if(e[r]===t)return r;return r}(e,t,i):bt(e,Et,i,!0)},zn.lowerCase=Fa,zn.lowerFirst=La,zn.lt=ea,zn.lte=ta,zn.max=function(e){return e&&e.length?ir(e,Va,gr):void 0},zn.maxBy=function(e,t){return e&&e.length?ir(e,Ji(t,2),gr):void 0},zn.mean=function(e){return Tt(e,Va)},zn.meanBy=function(e,t){return Tt(e,Ji(t,2))},zn.min=function(e){return e&&e.length?ir(e,Va,Ar):void 0},zn.minBy=function(e,t){return e&&e.length?ir(e,Ji(t,2),Ar):void 0},zn.stubArray=ou,zn.stubFalse=su,zn.stubObject=function(){return{}},zn.stubString=function(){return""},zn.stubTrue=function(){return!0},zn.multiply=hu,zn.nth=function(e,t){return e&&e.length?Rr(e,ia(t)):void 0},zn.noConflict=function(){return Ve._===this&&(Ve._=De),this},zn.noop=Za,zn.now=vs,zn.pad=function(e,t,n){e=ua(e);var r=(t=ia(t))?Xt(e):0;if(!t||r>=t)return e;var i=(t-r)/2;return ji(en(i),n)+e+ji(Qt(i),n)},zn.padEnd=function(e,t,n){e=ua(e);var r=(t=ia(t))?Xt(e):0;return t&&r<t?e+ji(t-r,n):e},zn.padStart=function(e,t,n){e=ua(e);var r=(t=ia(t))?Xt(e):0;return t&&r<t?ji(t-r,n)+e:e},zn.parseInt=function(e,t,n){return n||null==t?t=0:t&&(t=+t),cn(ua(e).replace(G,""),t||0)},zn.random=function(e,t,n){if(n&&"boolean"!=typeof n&&uo(e,t,n)&&(t=n=void 0),void 0===n&&("boolean"==typeof t?(n=t,t=void 0):"boolean"==typeof e&&(n=e,e=void 0)),void 0===e&&void 0===t?(e=0,t=1):(e=ra(e),void 0===t?(t=e,e=0):t=ra(t)),e>t){var r=e;e=t,t=r}if(n||e%1||t%1){var i=fn();return un(e+i*(t-e+He("1e-"+((i+"").length-1))),t)}return Pr(e,t)},zn.reduce=function(e,t,n){var r=Os(e)?gt:At,i=arguments.length<3;return r(e,Ji(t,4),n,i,tr)},zn.reduceRight=function(e,t,n){var r=Os(e)?vt:At,i=arguments.length<3;return r(e,Ji(t,4),n,i,nr)},zn.repeat=function(e,t,n){return t=(n?uo(e,t,n):void 0===t)?1:ia(t),qr(ua(e),t)},zn.replace=function(){var e=arguments,t=ua(e[0]);return e.length<3?t:t.replace(e[1],e[2])},zn.result=function(e,t,n){var r=-1,i=(t=ui(t,e)).length;for(i||(i=1,e=void 0);++r<i;){var o=null==e?void 0:e[Ao(t[r])];void 0===o&&(r=i,o=n),e=Ws(o)?o.call(e):o}return e},zn.round=pu,zn.runInContext=e,zn.sample=function(e){return(Os(e)?Mn:Mr)(e)},zn.size=function(e){if(null==e)return 0;if(Ls(e))return Js(e)?Xt(e):e.length;var t=ro(e);return t==g||t==_?e.size:Sr(e).length},zn.snakeCase=Pa,zn.some=function(e,t,n){var r=Os(e)?mt:Vr;return n&&uo(e,t,n)&&(t=void 0),r(e,Ji(t,3))},zn.sortedIndex=function(e,t){return Gr(e,t)},zn.sortedIndexBy=function(e,t,n){return Kr(e,t,Ji(n,2))},zn.sortedIndexOf=function(e,t){var n=null==e?0:e.length;if(n){var r=Gr(e,t);if(r<n&&ks(e[r],t))return r}return-1},zn.sortedLastIndex=function(e,t){return Gr(e,t,!0)},zn.sortedLastIndexBy=function(e,t,n){return Kr(e,t,Ji(n,2),!0)},zn.sortedLastIndexOf=function(e,t){if(null==e?0:e.length){var n=Gr(e,t,!0)-1;if(ks(e[n],t))return n}return-1},zn.startCase=qa,zn.startsWith=function(e,t,n){return e=ua(e),n=null==n?0:Yn(ia(n),0,e.length),t=Zr(t),e.slice(n,n+t.length)==t},zn.subtract=du,zn.sum=function(e){return e&&e.length?Dt(e,Va):0},zn.sumBy=function(e,t){return e&&e.length?Dt(e,Ji(t,2)):0},zn.template=function(e,t,n){var r=zn.templateSettings;n&&uo(e,t,n)&&(t=void 0),e=ua(e),t=fa({},t,r,Wi);var i,o,s=fa({},t.imports,r.imports,Wi),a=wa(s),u=Rt(s,a),l=0,c=t.interpolate||ce,f="__p += '",h=ge((t.escape||ce).source+"|"+c.source+"|"+(c===B?ne:ce).source+"|"+(t.evaluate||ce).source+"|$","g"),p="//# sourceURL="+(Ee.call(t,"sourceURL")?(t.sourceURL+"").replace(/\s/g," "):"lodash.templateSources["+ ++Ie+"]")+"\n";e.replace(h,(function(t,n,r,s,a,u){return r||(r=s),f+=e.slice(l,u).replace(fe,It),n&&(i=!0,f+="' +\n__e("+n+") +\n'"),a&&(o=!0,f+="';\n"+a+";\n__p += '"),r&&(f+="' +\n((__t = ("+r+")) == null ? '' : __t) +\n'"),l=u+t.length,t})),f+="';\n";var d=Ee.call(t,"variable")&&t.variable;if(d){if(ee.test(d))throw new K("Invalid `variable` option passed into `_.template`")}else f="with (obj) {\n"+f+"\n}\n";f=(o?f.replace(j,""):f).replace(O,"$1").replace(F,"$1;"),f="function("+(d||"obj")+") {\n"+(d?"":"obj || (obj = {});\n")+"var __t, __p = ''"+(i?", __e = _.escape":"")+(o?", __j = Array.prototype.join;\nfunction print() { __p += __j.call(arguments, '') }\n":";\n")+f+"return __p\n}";var g=Ba((function(){return he(a,p+"return "+f).apply(void 0,u)}));if(g.source=f,Ms(g))throw g;return g},zn.times=function(e,t){if((e=ia(e))<1||e>9007199254740991)return[];var n=4294967295,r=un(e,4294967295);e-=4294967295;for(var i=zt(r,t=Ji(t));++n<e;)t(n);return i},zn.toFinite=ra,zn.toInteger=ia,zn.toLength=oa,zn.toLower=function(e){return ua(e).toLowerCase()},zn.toNumber=sa,zn.toSafeInteger=function(e){return e?Yn(ia(e),-9007199254740991,9007199254740991):0===e?e:0},zn.toString=ua,zn.toUpper=function(e){return ua(e).toUpperCase()},zn.trim=function(e,t,n){if((e=ua(e))&&(n||void 0===t))return kt(e);if(!e||!(t=Zr(t)))return e;var r=Vt(e),i=Vt(t);return ci(r,Ot(r,i),Ft(r,i)+1).join("")},zn.trimEnd=function(e,t,n){if((e=ua(e))&&(n||void 0===t))return e.slice(0,Gt(e)+1);if(!e||!(t=Zr(t)))return e;var r=Vt(e);return ci(r,0,Ft(r,Vt(t))+1).join("")},zn.trimStart=function(e,t,n){if((e=ua(e))&&(n||void 0===t))return e.replace(G,"");if(!e||!(t=Zr(t)))return e;var r=Vt(e);return ci(r,Ot(r,Vt(t))).join("")},zn.truncate=function(e,t){var n=30,r="...";if(Us(t)){var i="separator"in t?t.separator:i;n="length"in t?ia(t.length):n,r="omission"in t?Zr(t.omission):r}var o=(e=ua(e)).length;if(Mt(e)){var s=Vt(e);o=s.length}if(n>=o)return e;var a=n-Xt(r);if(a<1)return r;var u=s?ci(s,0,a).join(""):e.slice(0,a);if(void 0===i)return u+r;if(s&&(a+=u.length-a),Ks(i)){if(e.slice(a).search(i)){var l,c=u;for(i.global||(i=ge(i.source,ua(re.exec(i))+"g")),i.lastIndex=0;l=i.exec(c);)var f=l.index;u=u.slice(0,void 0===f?a:f)}}else if(e.indexOf(Zr(i),a)!=a){var h=u.lastIndexOf(i);h>-1&&(u=u.slice(0,h))}return u+r},zn.unescape=function(e){return(e=ua(e))&&q.test(e)?e.replace(L,Kt):e},zn.uniqueId=function(e){var t=++Te;return ua(e)+t},zn.upperCase=Ia,zn.upperFirst=Ma,zn.each=us,zn.eachRight=ls,zn.first=Lo,Ja(zn,(fu={},lr(zn,(function(e,t){Ee.call(zn.prototype,t)||(fu[t]=e)})),fu),{chain:!1}),zn.VERSION="4.17.21",at(["bind","bindKey","curry","curryRight","partial","partialRight"],(function(e){zn[e].placeholder=zn})),at(["drop","take"],(function(e,t){jn.prototype[e]=function(n){n=void 0===n?1:an(ia(n),0);var r=this.__filtered__&&!t?new jn(this):this.clone();return r.__filtered__?r.__takeCount__=un(n,r.__takeCount__):r.__views__.push({size:un(n,4294967295),type:e+(r.__dir__<0?"Right":"")}),r},jn.prototype[e+"Right"]=function(t){return this.reverse()[e](t).reverse()}})),at(["filter","map","takeWhile"],(function(e,t){var n=t+1,r=1==n||3==n;jn.prototype[e]=function(e){var t=this.clone();return t.__iteratees__.push({iteratee:Ji(e,3),type:n}),t.__filtered__=t.__filtered__||r,t}})),at(["head","last"],(function(e,t){var n="take"+(t?"Right":"");jn.prototype[e]=function(){return this[n](1).value()[0]}})),at(["initial","tail"],(function(e,t){var n="drop"+(t?"":"Right");jn.prototype[e]=function(){return this.__filtered__?new jn(this):this[n](1)}})),jn.prototype.compact=function(){return this.filter(Va)},jn.prototype.find=function(e){return this.filter(e).head()},jn.prototype.findLast=function(e){return this.reverse().find(e)},jn.prototype.invokeMap=Ir((function(e,t){return"function"==typeof e?new jn(this):this.map((function(n){return _r(n,e,t)}))})),jn.prototype.reject=function(e){return this.filter(Ss(Ji(e)))},jn.prototype.slice=function(e,t){e=ia(e);var n=this;return n.__filtered__&&(e>0||t<0)?new jn(n):(e<0?n=n.takeRight(-e):e&&(n=n.drop(e)),void 0!==t&&(n=(t=ia(t))<0?n.dropRight(-t):n.take(t-e)),n)},jn.prototype.takeRightWhile=function(e){return this.reverse().takeWhile(e).reverse()},jn.prototype.toArray=function(){return this.take(4294967295)},lr(jn.prototype,(function(e,t){var n=/^(?:filter|find|map|reject)|While$/.test(t),r=/^(?:head|last)$/.test(t),i=zn[r?"take"+("last"==t?"Right":""):t],o=r||/^find/.test(t);i&&(zn.prototype[t]=function(){var t=this.__wrapped__,s=r?[1]:arguments,a=t instanceof jn,u=s[0],l=a||Os(t),c=function(e){var t=i.apply(zn,dt([e],s));return r&&f?t[0]:t};l&&n&&"function"==typeof u&&1!=u.length&&(a=l=!1);var f=this.__chain__,h=!!this.__actions__.length,p=o&&!f,d=a&&!h;if(!o&&l){t=d?t:new jn(this);var g=e.apply(t,s);return g.__actions__.push({func:rs,args:[c],thisArg:void 0}),new Rn(g,f)}return p&&d?e.apply(this,s):(g=this.thru(c),p?r?g.value()[0]:g.value():g)})})),at(["pop","push","shift","sort","splice","unshift"],(function(e){var t=ye[e],n=/^(?:push|sort|unshift)$/.test(e)?"tap":"thru",r=/^(?:pop|shift)$/.test(e);zn.prototype[e]=function(){var e=arguments;if(r&&!this.__chain__){var i=this.value();return t.apply(Os(i)?i:[],e)}return this[n]((function(n){return t.apply(Os(n)?n:[],e)}))}})),lr(jn.prototype,(function(e,t){var n=zn[t];if(n){var r=n.name+"";Ee.call(bn,r)||(bn[r]=[]),bn[r].push({name:t,func:n})}})),bn[zi(void 0,2).name]=[{name:"wrapper",func:void 0}],jn.prototype.clone=function(){var e=new jn(this.__wrapped__);return e.__actions__=yi(this.__actions__),e.__dir__=this.__dir__,e.__filtered__=this.__filtered__,e.__iteratees__=yi(this.__iteratees__),e.__takeCount__=this.__takeCount__,e.__views__=yi(this.__views__),e},jn.prototype.reverse=function(){if(this.__filtered__){var e=new jn(this);e.__dir__=-1,e.__filtered__=!0}else(e=this.clone()).__dir__*=-1;return e},jn.prototype.value=function(){var e=this.__wrapped__.value(),t=this.__dir__,n=Os(e),r=t<0,i=n?e.length:0,o=function(e,t,n){var r=-1,i=n.length;for(;++r<i;){var o=n[r],s=o.size;switch(o.type){case"drop":e+=s;break;case"dropRight":t-=s;break;case"take":t=un(t,e+s);break;case"takeRight":e=an(e,t-s)}}return{start:e,end:t}}(0,i,this.__views__),s=o.start,a=o.end,u=a-s,l=r?a:s-1,c=this.__iteratees__,f=c.length,h=0,p=un(u,this.__takeCount__);if(!n||!r&&i==u&&p==u)return ri(e,this.__actions__);var d=[];e:for(;u--&&h<p;){for(var g=-1,v=e[l+=t];++g<f;){var m=c[g],y=m.iteratee,_=m.type,b=y(v);if(2==_)v=b;else if(!b){if(1==_)continue e;break e}}d[h++]=v}return d},zn.prototype.at=is,zn.prototype.chain=function(){return ns(this)},zn.prototype.commit=function(){return new Rn(this.value(),this.__chain__)},zn.prototype.next=function(){void 0===this.__values__&&(this.__values__=na(this.value()));var e=this.__index__>=this.__values__.length;return{done:e,value:e?void 0:this.__values__[this.__index__++]}},zn.prototype.plant=function(e){for(var t,n=this;n instanceof Nn;){var r=zo(n);r.__index__=0,r.__values__=void 0,t?i.__wrapped__=r:t=r;var i=r;n=n.__wrapped__}return i.__wrapped__=e,t},zn.prototype.reverse=function(){var e=this.__wrapped__;if(e instanceof jn){var t=e;return this.__actions__.length&&(t=new jn(this)),(t=t.reverse()).__actions__.push({func:rs,args:[Uo],thisArg:void 0}),new Rn(t,this.__chain__)}return this.thru(Uo)},zn.prototype.toJSON=zn.prototype.valueOf=zn.prototype.value=function(){return ri(this.__wrapped__,this.__actions__)},zn.prototype.first=zn.prototype.head,Je&&(zn.prototype[Je]=function(){return this}),zn}();Ve._=Yt,void 0===(i=function(){return Yt}.call(t,n,t,r))||(r.exports=i)}).call(this)}).call(this,n(8),n(9)(e))},function(e,t){var n;n=function(){return this}();try{n=n||new Function("return this")()}catch(e){"object"==typeof window&&(n=window)}e.exports=n},function(e,t){e.exports=function(e){return e.webpackPolyfill||(e.deprecate=function(){},e.paths=[],e.children||(e.children=[]),Object.defineProperty(e,"loaded",{enumerable:!0,get:function(){return e.l}}),Object.defineProperty(e,"id",{enumerable:!0,get:function(){return e.i}}),e.webpackPolyfill=1),e}},function(e,t,n){var r=n(0),i=n(2),o=n(11),s=n(15),a=function(){this.events=s,this.gloo=o,this._is_loop_running=!1,this._canvases=[]};a.prototype.init=function(e,t){var n;n=r(e);var o=new i(n);return o.deactivate_context_menu(),this.events.init(o),this.gloo.init(o,t),this.register(o),o},a.prototype.register=function(e){this._canvases.push(e)},a.prototype.unregister=function(e){var t=this._canvases.indexOf(e);t>-1&&this._canvases.splice(t,1)},a.prototype.start_event_loop=function(){if(!this._is_loop_running){window.requestAnimFrame=window.requestAnimationFrame||window.webkitRequestAnimationFrame||window.mozRequestAnimationFrame||function(e){window.setTimeout(e,1e3/60)};var e=this;!function t(){e._request_id=requestAnimFrame(t);try{for(var n=0;n<e._canvases.length;n++)e._canvases[n].event_tick()}catch(t){throw e.stop_event_loop(),t}}(),this._is_loop_running=!0,console.debug("Event loop started.")}},a.prototype.stop_event_loop=function(){window.cancelAnimationFrame(this._request_id),this._is_loop_running=!1,console.debug("Event loop stopped.")},e.exports=new a},function(e,t,n){var r=n(12);var i=function(){this.glir=r};i.prototype.init=function(e,t){!function(e,t){var n=e.$el.get(0);e.gl=n.getContext("webgl",t)||n.getContext("experimental-webgl",t);var r=e.gl.getExtension("OES_standard_derivatives")||e.gl.getExtension("MOZ_OES_standard_derivatives")||e.gl.getExtension("WEBKIT_OES_standard_derivatives");null===r&&console.warn("Extension 'OES_standard_derivatives' is not supported in this browser. Some features may not work as expected"),null===(r=e.gl.getExtension("OES_element_index_uint")||e.gl.getExtension("MOZ_OES_element_index_uint")||e.gl.getExtension("WEBKIT_OES_element_index_uint"))&&console.warn("Extension 'OES_element_index_uint' is not supported in this browser. Some features may not work as expected"),null===(r=e.gl.getExtension("OES_texture_float"))&&console.warn("Extension 'OES_texture_float' is not supported in this browser. Some features may not work as expected"),null===(r=e.gl.getExtension("OES_texture_float_linear"))&&console.warn("Extension 'OES_texture_float_linear' is not supported in this browser. Some features may not work as expected")}(e,t),this.glir.init(e)},e.exports=new i},function(e,t,n){var r=n(2),i=n(13),o=n(14),s=i.debug,a=o.to_array_buffer;function u(e,t,n,r,i,o){var s=function(e){return g[e]}(r),a=s[0],u=s[1];_vbo_info=e._ns[n];var l=_vbo_info.handle;e.gl.enableVertexAttribArray(t),e.gl.bindBuffer(e.gl.ARRAY_BUFFER,l),e.gl.vertexAttribPointer(t,u,e.gl[a],!1,i,o)}function l(e,t){e.gl.disableVertexAttribArray(t)}function c(e,t,n,r){"JUST_DELETED"!==t&&(e.gl.activeTexture(e.gl.TEXTURE0+r),e.gl.bindTexture(e.gl.TEXTURE_2D,t))}function f(e,t,n,r){e.gl.activeTexture(e.gl.TEXTURE0+r),e.gl.bindTexture(e.gl.TEXTURE_2D,null)}function h(e,t,n,r,i,o,s,a,u,l){if(e.gl.bindTexture(n,t),e.gl.pixelStorei(e.gl.UNPACK_ALIGNMENT,1),null===s)e.gl.texImage2D(n,0,r,i,o,0,r,e.gl.UNSIGNED_BYTE,s);else if(s.getContext)e.gl.texImage2D(n,0,e.gl.RGBA,e.gl.RGBA,e.gl.UNSIGNED_BYTE,s);else if(s.canvas)e.gl.texImage2D(n,0,e.gl.RGBA,e.gl.RGBA,e.gl.UNSIGNED_BYTE,s.canvas);else{var c;if(c=l==e.gl.FLOAT?new Float32Array(s):new Uint8Array(s),a&&u&&(u[0]!==o||u[1]!==i)){var f=function(e){for(var t=[8,4,2,1],n=0;n<t.length;n++)if(e%t[n]==0)return t[n]}(u[u.length-2]*u[u.length-1]);e.gl.pixelStorei(e.gl.UNPACK_ALIGNMENT,f),e.gl.texSubImage2D(n,0,a[1],a[0],u[1],u[0],r,l,c)}else e.gl.pixelStorei(e.gl.UNPACK_ALIGNMENT,1),e.gl.texImage2D(n,0,r,i,o,0,r,l,c)}}function p(e,t,n,r,i,o){e.gl.bindBuffer(n,t),o?e.gl.bufferSubData(n,r,i):e.gl.bufferData(n,i,e.gl.STATIC_DRAW)}var d={float32:"FLOAT",uint8:"UNSIGNED_BYTE"};var g={float:["FLOAT",1],vec2:["FLOAT",2],vec3:["FLOAT",3],vec4:["FLOAT",4],int:["INT",1],ivec2:["INT",2],ivec3:["INT",3],ivec4:["INT",4]};var v={float:"uniform1fv",vec2:"uniform2fv",vec3:"uniform3fv",vec4:"uniform4fv",int:"uniform1iv",ivec2:"uniform2iv",ivec3:"uniform3iv",ivec4:"uniform4iv",mat2:"uniformMatrix2fv",mat3:"uniformMatrix3fv",mat4:"uniformMatrix4fv"};var m={VertexBuffer:"ARRAY_BUFFER",IndexBuffer:"ELEMENT_ARRAY_BUFFER",Texture2D:"TEXTURE_2D"};function y(e){return m[e]}var _={color:["COLOR_ATTACHMENT0","RGBA4"],depth:["DEPTH_ATTACHMENT","DEPTH_COMPONENT16"],stencil:["STENCIL_ATTACHMENT","STENCIL_INDEX8"]};function b(e,t){for(var n=t.split("|"),r=0,i=0;i<n.length;i++){var o=n[i].toUpperCase().trim();r|=e.gl[o]}return r}function w(e,t){var n=e._ns[t],r=e.env.fb_stack;0==r.length&&r.push(null),r[r.length-1]!==n.handle?(s("Binding frame buffer {0}.".format(t)),e.gl.bindFramebuffer(e.gl.FRAMEBUFFER,n.handle),r.push(n.handle)):s("Frame buffer already active {0}".format(t))}function x(e,t){var n=e._ns[t],r=e.env.fb_stack;for(0==r.length&&r.push(null);r[r.length-1]===n.handle;)r.pop();s("Binding previous frame buffer"),e.gl.bindFramebuffer(e.gl.FRAMEBUFFER,r[r.length-1])}function E(e){e.env={fb_stack:[]}}function T(){this._queue=[]}T.prototype.clear=function(){this._queue=[]},T.prototype.append=function(e){this._queue.push(e)},T.prototype.append_multi=function(e){for(var t=0;t<e.length;t++)this._queue.push(e[t])},T.prototype.get=function(){return this._queue},Object.defineProperty(T.prototype,"length",{get:function(){return this._queue.length}}),r.prototype.set_deferred=function(e){this._deferred=e},r.prototype.execute_pending_commands=function(){var e=this.glir_queue.get(),t=-1;if(0==e.length)return 0;for(var n=0;n<e.length;n++)if("SWAP"===e[n][0]){t=n;break}for(n=0;n<=t;n++)this.command(e[n],!1);return t>=0&&(s("Processed {0} events.".format(t+1)),this.glir_queue._queue=this.glir_queue._queue.slice(t+1)),t+1},r.prototype.command=function(e,t){void 0===t&&(t=this._deferred);var n=e[0].toLowerCase();t?this.glir_queue.append(e):this.glir[n](this,e.slice(1))};var S=function(){};S.prototype.init=function(e){e._ns={},e._deferred=!0,E(e),e.glir_queue=new T,e.glir=this},S.prototype.current=function(e,t){E(e),e.gl.bindFramebuffer(e.gl.FRAMEBUFFER,null)},S.prototype.swap=function(e,t){},S.prototype.create=function(e,t){var n=t[0],r=t[1];"VertexBuffer"==r?(s("Creating vertex buffer '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createBuffer(),size:0}):"IndexBuffer"==r?(s("Creating index buffer '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createBuffer(),size:0}):"FrameBuffer"==r?(s("Creating frame buffer '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createFramebuffer(),size:0,validated:!1}):"RenderBuffer"==r?(s("Creating render buffer '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createRenderbuffer(),size:0}):"Texture2D"==r?(s("Creating texture '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createTexture(),size:0,shape:[]}):"Program"==r?(s("Creating program '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createProgram(),attributes:{},_handles_cache:{},textures:{},texture_uniforms:{},shaders:{}}):"VertexShader"==r?(s("Creating VertexShader '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createShader(e.gl.VERTEX_SHADER)}):"FragmentShader"==r&&(s("Creating FragmentShader '{0}'.".format(n)),e._ns[n]={object_type:r,handle:e.gl.createShader(e.gl.FRAGMENT_SHADER)})},S.prototype.delete=function(e,t){var n=t[0],r=e._ns[n].object_type,i=e._ns[n].handle;e._ns[n].handle="JUST_DELETED","VertexBuffer"==r?(s("Deleting vertex buffer '{0}'.".format(n)),e.gl.deleteBuffer(i)):"IndexBuffer"==r?(s("Deleting index buffer '{0}'.".format(n)),e.gl.deleteBuffer(i)):"FrameBuffer"==r?(s("Deleting frame buffer '{0}'.".format(n)),e.gl.deleteFramebuffer(i)):"RenderBuffer"==r?(s("Deleting render buffer '{0}'.".format(n)),e.gl.deleteRenderbuffer(i)):"Texture2D"==r?(s("Deleting texture '{0}'.".format(n)),e.gl.deleteTexture(i)):"Program"==r?(s("Deleting program '{0}'.".format(n)),e.gl.deleteProgram(i)):r.indexOf("Shader")>=0&&(s("Deleting shader '{0}'.".format(n)),e.gl.deleteShader(i))},S.prototype.size=function(e,t){var n,r=t[0],i=t[1],o=t[2],a=e._ns[r],u=a.handle,l=a.object_type,c=e.gl[y(l)];l.indexOf("Texture")>=0?(a.format=o.toUpperCase(),s("Setting texture size to {1} for '{0}'.".format(r,i))):"RenderBuffer"==l?(e.gl.bindRenderbuffer(e.gl.RENDERBUFFER,u),a.format=e.gl[(n=o,_[n][1])],e.gl.renderbufferStorage(e.gl.RENDERBUFFER,a.format,i[1],i[0]),e.gl.bindRenderbuffer(e.gl.RENDERBUFFER,null)):(s("Setting buffer size to {1} for '{0}'.".format(r,i)),p(e,u,c,0,i,!1)),a.size=i},S.prototype.data=function(e,t){var n,r=t[0],i=t[1],o=t[2],u=e._ns[r],l=u.object_type,c=u.handle,f=e.gl[y(l)],g=a(o);if(l.indexOf("Shader")>=0)!function(e,t,n){"string"!=typeof n&&(n=String.fromCharCode.apply(null,new Uint8Array(n))),n=n.replace(/\\n/g,"\n"),e.gl.shaderSource(t,n),e.gl.compileShader(t),!!e.gl.getShaderParameter(t,e.gl.COMPILE_STATUS)||console.error(e.gl.getShaderInfoLog(t))}(e,c,g);else if(l.indexOf("Texture")>=0){var v=u.size,m=v[0],_=v[1],b=e.gl[u.format];s("Setting texture data for '{0}'.".format(r));var w=e.gl[(n=o.dtype,d[n])];h(e,c,f,b,_,m,g,i,o.shape,w),u.shape=v}else s("Setting buffer data for '{0}'.".format(r)),p(e,c,f,i,g,u.size>0),u.size=g.byteLength},S.prototype.attribute=function(e,t){var n=t[0],r=t[1],i=t[2],o=t[3][0],a=t[3][1],u=t[3][2],l=e._ns[n],c=l.handle;s("Creating attribute '{0}' for program '{1}'.".format(r,n));var f=function(e,t,n){return e.gl.getAttribLocation(t,n)}(e,c,r);l.attributes[r]={handle:f,type:i,vbo_id:o,stride:a,offset:u}},S.prototype.uniform=function(e,t){var n,r,i,o=t[0],u=t[1],l=t[2],c=t[3],f=e._ns[o],h=f.handle;void 0===f._handles_cache[u]&&(s("Creating uniform '{0}' for program '{1}'.".format(u,o)),n=e.gl.getUniformLocation(h,u),r=function(e){return v[e]}(l),f._handles_cache[u]=[n,r]),e.gl.useProgram(h),s("Setting uniform '{0}' to '{1}' with {2} elements.".format(u,c,c.length)),function(e,t,n,r){array=a(r),n.indexOf("Matrix")>0?e.gl[n](t,!1,array):e.gl[n](t,array)}(e,n=(i=e._ns[o]._handles_cache[u])[0],r=i[1],c)},S.prototype.texture=function(e,t){var n=t[0],r=t[1],i=t[2],o=e._ns[n],a=o.handle,u=e._ns[i].handle;if("JUST_DELETED"===u)return s("Removing texture '{0}' from program '{1}'".format(i,n)),void delete o.textures[i];s("Initializing texture '{0}' for program '{1}'.".format(i,n)),o.texture_uniforms.hasOwnProperty(r)&&(s("Removing previously assigned texture for '{0}'".format(r)),delete o.textures[o.texture_uniforms[r]]);var l=e.gl.getUniformLocation(a,r);o.texture_uniforms[r]=i,e._ns[n].textures[i]={sampler_name:r,sampler_handle:l,number:-1,handle:u}},S.prototype.interpolation=function(e,t){var n=t[0],r=t[1].toUpperCase(),i=t[2].toUpperCase(),o=e._ns[n].handle,s=e.gl.TEXTURE_2D;e.gl.bindTexture(s,o),e.gl.texParameteri(s,e.gl.TEXTURE_MIN_FILTER,e.gl[r]),e.gl.texParameteri(s,e.gl.TEXTURE_MAG_FILTER,e.gl[i]),e.gl.bindTexture(s,null)},S.prototype.wrapping=function(e,t){var n=t[0],r=t[1],i=e._ns[n].handle,o=e.gl.TEXTURE_2D;e.gl.bindTexture(o,i),e.gl.texParameteri(o,e.gl.TEXTURE_WRAP_S,e.gl[r[0].toUpperCase()]),e.gl.texParameteri(o,e.gl.TEXTURE_WRAP_T,e.gl[r[1].toUpperCase()]),e.gl.bindTexture(o,null)},S.prototype.draw=function(e,t){var n,r,i,o,a=t[0],h=t[1].toUpperCase(),p=t[2],d=e._ns[a].handle,g=e._ns[a].attributes,v=e._ns[a].textures,m=0;for(r in e.gl.useProgram(d),g){var y=g[r];s("Activating attribute '{0}' for program '{1}'.".format(r,a)),u(e,y.handle,y.vbo_id,y.type,y.stride,y.offset)}for(i in v)n=v[i],"JUST_DELETED"!==e._ns[i].handle?(n.number=m,m+=1,s("Activating texture '{0}' for program '{1}' as number '{2}'.".format(i,a,n.number)),c(e,n.handle,n.sampler_handle,n.number),e.gl.uniform1i(n.sampler_handle,n.number)):(s("Ignoring texture '{0}' from program '{1}'".format(i,a)),n.handle="JUST_DELETED");if(2==p.length){var _=p[0];o=p[1],s("Rendering program '{0}' with {1}.".format(a,h)),e.gl.drawArrays(e.gl[h],_,o)}else if(3==p.length){var b=p[0],w=p[1];o=p[2];var x=e._ns[b].handle;s("Rendering program '{0}' with {1} and index buffer '{2}' of type '{3}'.".format(a,h,b,w)),e.gl.bindBuffer(e.gl.ELEMENT_ARRAY_BUFFER,x),e.gl.drawElements(e.gl[h],o,e.gl[w],0)}for(r in g)s("Deactivating attribute '{0}' for program '{1}'.".format(r,a)),l(e,g[r].handle);var E={};for(i in v)n=v[i],s("Deactivating texture '{0}' for program '{1}'.".format(i,a)),f(e,n.handle,n.sampler_handle,n.number),"JUST_DELETED"!=e._ns[i].handle&&(E[i]=n);e._ns[a].textures=E},S.prototype.attach=function(e,t){var n=t[0],r=e._ns[n],i=r.object_type,o=r.handle;if("Program"==i){var a=t[1],u=e._ns[a].handle;return r.shaders[a]=u,void e.gl.attachShader(o,u)}var l,c,f=t[2],p=e.gl[(l=t[1],_[l][0])];w(e,n),0==f?(s("Attaching RenderBuffer object {0} to framebuffer {1}".format(f,n)),e.gl.framebufferRenderbuffer(e.gl.FRAMEBUFFER,p,e.gl.RENDERBUFFER,null)):(c=e._ns[f],s("Attaching {0} object {1} to framebuffer {2} for {3}".format(c.object_type,f,n,t[1])),"RenderBuffer"==c.object_type?(e.gl.bindRenderbuffer(e.gl.RENDERBUFFER,c.handle),e.gl.framebufferRenderbuffer(e.gl.FRAMEBUFFER,p,e.gl.RENDERBUFFER,c.handle),e.gl.bindRenderbuffer(e.gl.RENDERBUFFER,null)):"Texture2D"==c.object_type&&(0==c.shape.length&&(s("Setting empty texture data to unset texture before attaching to framebuffer"),h(e,c.handle,e.gl.TEXTURE_2D,e.gl[c.format],c.size[1],c.size[0],null)),e.gl.bindTexture(e.gl.TEXTURE_2D,c.handle),e.gl.framebufferTexture2D(e.gl.FRAMEBUFFER,p,e.gl.TEXTURE_2D,c.handle,0),e.gl.bindTexture(e.gl.TEXTURE_2D,null))),e._ns[n].validated=!1,x(e,n)},S.prototype.link=function(e,t){var n,r=t[0],i=e._ns[r],o=i.handle,s=i.shaders;for(n in e.gl.linkProgram(o),e.gl.getProgramParameter(o,e.gl.LINK_STATUS)||console.warn("Could not initialise shaders on program '{0}'.".format(o)),s)e.gl.detachShader(o,s[n]);i.shaders={},i._handles_cache={}},S.prototype.framebuffer=function(e,t){var n=t[0],r=t[1],i=e._ns[n];r?(s("Binding framebuffer {0}".format(n)),w(e,n),i.validated||(i.validated=!0,function(e){var t=e.gl.checkFramebufferStatus(e.gl.FRAMEBUFFER);if(t!=e.gl.FRAMEBUFFER_COMPLETE)throw t==e.gl.FRAMEBUFFER_INCOMPLETE_ATTACHMENT?"FrameBuffer attachments are incomplete.":t==e.gl.FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT?"No valid attachments in the FrameBuffer.":t==e.gl.FRAMEBUFFER_INCOMPLETE_DIMENSIONS?"attachments do not have the same width and height.":t==e.gl.FRAMEBUFFER_UNSUPPORTED?"Combination of internal formats used by attachments is not supported.":"Unknown framebuffer error"+t}(e))):(s("Unbinding framebuffer {0}".format(n)),x(e,n))},S.prototype.func=function(e,t){var n=t[0];s("Calling {0}({1}).".format(n,t.slice(1)));for(var r=1;r<t.length;r++)"string"==typeof t[r]&&(t[r]=b(e,t[r]));var i=e.gl[n],o=t.slice(1);i.apply(e.gl,o)},e.exports=new S},function(e,t){String.prototype.format||(String.prototype.format=function(){var e=arguments;return this.replace(/{(\d+)}/g,(function(t,n){return void 0!==e[n]?e[n]:t}))}),void 0===String.prototype.trim&&(String.prototype.trim=function(){return String(this).replace(/^\s+|\s+$/g,"")}),Array.prototype.equals=function(e){if(!e)return!1;if(this.length!=e.length)return!1;for(var t=0,n=this.length;t<n;t++)if(this[t]instanceof Array&&e[t]instanceof Array){if(!this[t].equals(e[t]))return!1}else if(this[t]!=e[t])return!1;return!0},"function"!=typeof String.prototype.startsWith&&(String.prototype.startsWith=function(e){return this.slice(0,e.length)==e}),window.VISPY_DEBUG=!1,e.exports={debug:function(e){window.VISPY_DEBUG&&console.debug(e)}}},function(e,t){var n={float32:Float32Array,int8:Int8Array,int16:Int16Array,int32:Int32Array,uint8:Uint8Array,uint16:Uint16Array,uint32:Uint32Array};e.exports={to_array_buffer:function(e){var t=e.storage_type;if(null==t)return e;var r=e.data_type,i=e.buffer;return"javascript_array"==t?n[r](i):"javascript_typed_array"==t||"array_buffer"==t?i:"binary"==t?i.buffer:"base64"==t?function(e){for(var t=window.atob(e),n=t.length,r=new Uint8Array(n),i=0;i<n;i++){var o=t.charCodeAt(i);r[i]=o}return r.buffer}(i):void 0}}},function(e,t,n){var r=n(2),i=n(16);function o(e){var t=[];return e.altKey&&t.push("alt"),e.ctrlKey&&t.push("ctrl"),e.metaKey&&t.push("meta"),e.shiftKey&&t.push("shift"),t}n(0),n(17);var s={8:"BACKSPACE",9:"TAB",13:"ENTER",16:"SHIFT",17:"CONTROL",18:"ALT",27:"ESCAPE",32:"SPACE",33:"PAGEUP",34:"PAGEDOWN",35:"END",36:"HOME",37:"LEFT",38:"UP",39:"RIGHT",40:"DOWN",45:"INSERT",46:"DELETE",91:"META",92:"META",96:"0",97:"1",98:"2",99:"3",100:"4",101:"5",102:"6",103:"7",104:"8",105:"9",106:"*",107:"+",109:"-",110:".",111:"/",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",186:";",187:"=",188:",",189:"-",190:".",191:"/",192:"`",219:"[",220:"\\",221:"]",222:"'"};function a(e){var t=function(e){return window.event?e.keyCode:e.which?e.which:void 0}(e),n=s[t];return null==n&&(n=function(e){return String.fromCharCode(e).toUpperCase().trim()}(t)),n}var u={0:1,2:2,1:3};function l(e,t,n){var r;r=e._eventinfo.is_button_pressed?u[t.button]:null;var i=function(e,t){var n=e.getBoundingClientRect();return[t.clientX-n.left,t.clientY-n.top]}(e.$el.get(0),t),s=o(t),a=e._eventinfo.press_event;e._eventinfo.last_event;return{type:n,pos:i,button:r,is_dragging:null!=a,modifiers:s,delta:null,press_event:a,last_event:null}}function c(e,t,n){var r=o(t);e._eventinfo.last_event;return{type:n,modifiers:r,key_code:a(t),last_event:null}}function f(e){null==e&&(e=100),this._queue=[],this.maxlen=e}r.prototype._mouse_press=function(e){},r.prototype._mouse_release=function(e){},r.prototype._mouse_move=function(e){},r.prototype._mouse_wheel=function(e){},r.prototype._mouse_click=function(e){},r.prototype._mouse_dblclick=function(e){},r.prototype._key_press=function(e){},r.prototype._key_release=function(e){},r.prototype._initialize=function(e){},r.prototype._resize=function(e){},r.prototype._paint=function(e){},r.prototype._event_tick=function(e){},r.prototype.on_mouse_press=function(e){this._mouse_press=e},r.prototype.on_mouse_release=function(e){this._mouse_release=e},r.prototype.on_mouse_move=function(e){this._mouse_move=e},r.prototype.on_mouse_wheel=function(e){this._mouse_wheel=e},r.prototype.on_mouse_dblclick=function(e){this._mouse_dblclick=e},r.prototype.on_key_press=function(e){this._key_press=e},r.prototype.on_key_release=function(e){this._key_release=e},r.prototype.on_initialize=function(e){this._initialize=e},r.prototype.on_resize=function(e){this._resize=e},r.prototype.on_paint=function(e){this._paint=e},r.prototype.on_event_tick=function(e){this._event_tick=e},r.prototype.initialize=function(){var e={type:"initialize"};this._set_size(),this._initialize(e)},r.prototype._set_size=function(e){return null==e&&(e=[this.$el.width(),this.$el.height()]),this.size=e,this.width=e[0],this.height=e[1],e},r.prototype.paint=function(){var e={type:"paint"};this.event_queue.append(e)},r.prototype.update=r.prototype.paint,r.prototype.resize=function(e){var t=function(e,t){return{type:"resize",size:t}}(0,e=this._set_size(e));this.gl.canvas.width=e[0],this.gl.canvas.height=e[1],this.event_queue.append(t),this._resize(t)},r.prototype.event_tick=function(){if(this._event_tick(),this.execute_pending_commands()>0){var e={type:"paint"};this._paint(e)}},r.prototype.is_fullscreen=function(){return i.enabled&i.isFullscreen},r.prototype.toggle_fullscreen=function(){i.enabled&&(i.isFullscreen?(i.exit(),this.resize(this._size)):(this.$el.width("100%").height("100%"),this._size=[this.$el.width(),this.$el.height()],i.request(this.$el[0]),this.resize([screen.width,screen.height])))},r.prototype.deactivate_context_menu=function(){document.oncontextmenu=function(){return!1}},r.prototype.resizable=function(){var e=this;this.$el.resizable({resize:function(t,n){e.resize([n.size.width,n.size.height])}})},f.prototype.clear=function(){this._queue=[]},f.prototype.append=function(e,t){var n,r,i,o=!0;if(null==t&&(t=!0),t){var s=this._queue[this._queue.length-1];if(null!=s){var a=(r=s,(i=(n=e).type)==r.type&&"mouse_move"==i&&n.button==r.button&n.is_dragging==r.is_dragging&n.modifiers.equals(r.modifiers)?["pos"]:[]);if(a.length>0){for(var u=0;u<a.length;u++){var l=a[u];this._queue[this._queue.length-1][l]=e[l]}o=!1}}}o&&this._queue.push(e),this._queue.length>this.maxlen&&(this._queue.shift(),this._queue[0].last_event=null)},f.prototype.get=function(){return this._queue},Object.defineProperty(f.prototype,"length",{get:function(){return this._queue.length}});var h=function(){};h.prototype.init=function(e){!function(e){e.$el.on("resize",(function(t){e.resize([t.width(),t.height()])})),e.event_queue=new f,e._eventinfo={type:null,pos:null,button:null,is_dragging:null,key:null,modifiers:[],press_event:null,last_event:null,delta:null},e._eventinfo.is_button_pressed=0,e.$el.on("mousemove",(function(t){var n=l(e,t,"mouse_move");e._mouse_move(n),e.event_queue.append(n)})),e.$el.on("mousedown",(function(t){++e._eventinfo.is_button_pressed;var n=l(e,t,"mouse_press");e._mouse_press(n),e._eventinfo.press_event=n,e.event_queue.append(n)})),e.$el.on("mouseup",(function(t){--e._eventinfo.is_button_pressed;var n=l(e,t,"mouse_release");e._mouse_release(n),e._eventinfo.press_event=null,e.event_queue.append(n)})),e.$el.on("click",(function(t){e._eventinfo.press_event=null})),e.$el.on("dblclick",(function(t){e._eventinfo.press_event=null})),e.$el.on("wheel",(function(t){var n=l(e,t,"mouse_wheel");n.delta=[.01*t.originalEvent.deltaX,.01*-t.originalEvent.deltaY],e._mouse_wheel(n),e.event_queue.append(n),t.preventDefault(),t.stopPropagation()})),e.$el.on("keydown",(function(t){var n=c(e,t,"key_press");e._key_press(n),e.event_queue.append(n)})),e.$el.on("keyup",(function(t){var n=c(e,t,"key_release");e._key_release(n),e.event_queue.append(n)})),e.$el.on("mouseout",(function(e){}))}(e)},e.exports=new h},function(e,t,n){
/*!
* screenfull
* v4.2.1 - 2019-07-27
* (c) Sindre Sorhus; MIT License
*/
!function(){"use strict";var t="undefined"!=typeof window&&void 0!==window.document?window.document:{},n=e.exports,r="undefined"!=typeof Element&&"ALLOW_KEYBOARD_INPUT"in Element,i=function(){for(var e,n=[["requestFullscreen","exitFullscreen","fullscreenElement","fullscreenEnabled","fullscreenchange","fullscreenerror"],["webkitRequestFullscreen","webkitExitFullscreen","webkitFullscreenElement","webkitFullscreenEnabled","webkitfullscreenchange","webkitfullscreenerror"],["webkitRequestFullScreen","webkitCancelFullScreen","webkitCurrentFullScreenElement","webkitCancelFullScreen","webkitfullscreenchange","webkitfullscreenerror"],["mozRequestFullScreen","mozCancelFullScreen","mozFullScreenElement","mozFullScreenEnabled","mozfullscreenchange","mozfullscreenerror"],["msRequestFullscreen","msExitFullscreen","msFullscreenElement","msFullscreenEnabled","MSFullscreenChange","MSFullscreenError"]],r=0,i=n.length,o={};r<i;r++)if((e=n[r])&&e[1]in t){for(r=0;r<e.length;r++)o[n[0][r]]=e[r];return o}return!1}(),o={change:i.fullscreenchange,error:i.fullscreenerror},s={request:function(e){return new Promise(function(n,o){var s,a=i.requestFullscreen,u=function(){this.off("change",u),n()}.bind(this);this.on("change",u),e=e||t.documentElement,s=/ Version\/5\.1(?:\.\d+)? Safari\//.test(navigator.userAgent)?e[a]():e[a](r?Element.ALLOW_KEYBOARD_INPUT:{}),Promise.resolve(s).catch(o)}.bind(this))},exit:function(){return new Promise(function(e){if(this.isFullscreen){var n=function(){this.off("change",n),e()}.bind(this);t[i.exitFullscreen](),this.on("change",n)}else e()}.bind(this))},toggle:function(e){return this.isFullscreen?this.exit():this.request(e)},onchange:function(e){this.on("change",e)},onerror:function(e){this.on("error",e)},on:function(e,n){var r=o[e];r&&t.addEventListener(r,n,!1)},off:function(e,n){var r=o[e];r&&t.removeEventListener(r,n,!1)},raw:i};i?(Object.defineProperties(s,{isFullscreen:{get:function(){return Boolean(t[i.fullscreenElement])}},element:{enumerable:!0,get:function(){return t[i.fullscreenElement]}},enabled:{enumerable:!0,get:function(){return Boolean(t[i.fullscreenEnabled])}}}),n?(e.exports=s,e.exports.default=s):window.screenfull=s):n?e.exports=!1:window.screenfull=!1}()},function(e,t,n){var r,i,o;
/*!
 * jQuery UI Resizable 1.12.1
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */i=[n(0),n(18),n(20),n(21),n(1),n(3)],void 0===(o="function"==typeof(r=function(e){return e.widget("ui.resizable",e.ui.mouse,{version:"1.12.1",widgetEventPrefix:"resize",options:{alsoResize:!1,animate:!1,animateDuration:"slow",animateEasing:"swing",aspectRatio:!1,autoHide:!1,classes:{"ui-resizable-se":"ui-icon ui-icon-gripsmall-diagonal-se"},containment:!1,ghost:!1,grid:!1,handles:"e,s,se",helper:!1,maxHeight:null,maxWidth:null,minHeight:10,minWidth:10,zIndex:90,resize:null,start:null,stop:null},_num:function(e){return parseFloat(e)||0},_isNumber:function(e){return!isNaN(parseFloat(e))},_hasScroll:function(t,n){if("hidden"===e(t).css("overflow"))return!1;var r,i=n&&"left"===n?"scrollLeft":"scrollTop";return t[i]>0||(t[i]=1,r=t[i]>0,t[i]=0,r)},_create:function(){var t,n=this.options,r=this;this._addClass("ui-resizable"),e.extend(this,{_aspectRatio:!!n.aspectRatio,aspectRatio:n.aspectRatio,originalElement:this.element,_proportionallyResizeElements:[],_helper:n.helper||n.ghost||n.animate?n.helper||"ui-resizable-helper":null}),this.element[0].nodeName.match(/^(canvas|textarea|input|select|button|img)$/i)&&(this.element.wrap(e("<div class='ui-wrapper' style='overflow: hidden;'></div>").css({position:this.element.css("position"),width:this.element.outerWidth(),height:this.element.outerHeight(),top:this.element.css("top"),left:this.element.css("left")})),this.element=this.element.parent().data("ui-resizable",this.element.resizable("instance")),this.elementIsWrapper=!0,t={marginTop:this.originalElement.css("marginTop"),marginRight:this.originalElement.css("marginRight"),marginBottom:this.originalElement.css("marginBottom"),marginLeft:this.originalElement.css("marginLeft")},this.element.css(t),this.originalElement.css("margin",0),this.originalResizeStyle=this.originalElement.css("resize"),this.originalElement.css("resize","none"),this._proportionallyResizeElements.push(this.originalElement.css({position:"static",zoom:1,display:"block"})),this.originalElement.css(t),this._proportionallyResize()),this._setupHandles(),n.autoHide&&e(this.element).on("mouseenter",(function(){n.disabled||(r._removeClass("ui-resizable-autohide"),r._handles.show())})).on("mouseleave",(function(){n.disabled||r.resizing||(r._addClass("ui-resizable-autohide"),r._handles.hide())})),this._mouseInit()},_destroy:function(){this._mouseDestroy();var t,n=function(t){e(t).removeData("resizable").removeData("ui-resizable").off(".resizable").find(".ui-resizable-handle").remove()};return this.elementIsWrapper&&(n(this.element),t=this.element,this.originalElement.css({position:t.css("position"),width:t.outerWidth(),height:t.outerHeight(),top:t.css("top"),left:t.css("left")}).insertAfter(t),t.remove()),this.originalElement.css("resize",this.originalResizeStyle),n(this.originalElement),this},_setOption:function(e,t){switch(this._super(e,t),e){case"handles":this._removeHandles(),this._setupHandles()}},_setupHandles:function(){var t,n,r,i,o,s=this.options,a=this;if(this.handles=s.handles||(e(".ui-resizable-handle",this.element).length?{n:".ui-resizable-n",e:".ui-resizable-e",s:".ui-resizable-s",w:".ui-resizable-w",se:".ui-resizable-se",sw:".ui-resizable-sw",ne:".ui-resizable-ne",nw:".ui-resizable-nw"}:"e,s,se"),this._handles=e(),this.handles.constructor===String)for("all"===this.handles&&(this.handles="n,e,s,w,se,sw,ne,nw"),r=this.handles.split(","),this.handles={},n=0;n<r.length;n++)i="ui-resizable-"+(t=e.trim(r[n])),o=e("<div>"),this._addClass(o,"ui-resizable-handle "+i),o.css({zIndex:s.zIndex}),this.handles[t]=".ui-resizable-"+t,this.element.append(o);this._renderAxis=function(t){var n,r,i,o;for(n in t=t||this.element,this.handles)this.handles[n].constructor===String?this.handles[n]=this.element.children(this.handles[n]).first().show():(this.handles[n].jquery||this.handles[n].nodeType)&&(this.handles[n]=e(this.handles[n]),this._on(this.handles[n],{mousedown:a._mouseDown})),this.elementIsWrapper&&this.originalElement[0].nodeName.match(/^(textarea|input|select|button)$/i)&&(r=e(this.handles[n],this.element),o=/sw|ne|nw|se|n|s/.test(n)?r.outerHeight():r.outerWidth(),i=["padding",/ne|nw|n/.test(n)?"Top":/se|sw|s/.test(n)?"Bottom":/^e$/.test(n)?"Right":"Left"].join(""),t.css(i,o),this._proportionallyResize()),this._handles=this._handles.add(this.handles[n])},this._renderAxis(this.element),this._handles=this._handles.add(this.element.find(".ui-resizable-handle")),this._handles.disableSelection(),this._handles.on("mouseover",(function(){a.resizing||(this.className&&(o=this.className.match(/ui-resizable-(se|sw|ne|nw|n|e|s|w)/i)),a.axis=o&&o[1]?o[1]:"se")})),s.autoHide&&(this._handles.hide(),this._addClass("ui-resizable-autohide"))},_removeHandles:function(){this._handles.remove()},_mouseCapture:function(t){var n,r,i=!1;for(n in this.handles)((r=e(this.handles[n])[0])===t.target||e.contains(r,t.target))&&(i=!0);return!this.options.disabled&&i},_mouseStart:function(t){var n,r,i,o=this.options,s=this.element;return this.resizing=!0,this._renderProxy(),n=this._num(this.helper.css("left")),r=this._num(this.helper.css("top")),o.containment&&(n+=e(o.containment).scrollLeft()||0,r+=e(o.containment).scrollTop()||0),this.offset=this.helper.offset(),this.position={left:n,top:r},this.size=this._helper?{width:this.helper.width(),height:this.helper.height()}:{width:s.width(),height:s.height()},this.originalSize=this._helper?{width:s.outerWidth(),height:s.outerHeight()}:{width:s.width(),height:s.height()},this.sizeDiff={width:s.outerWidth()-s.width(),height:s.outerHeight()-s.height()},this.originalPosition={left:n,top:r},this.originalMousePosition={left:t.pageX,top:t.pageY},this.aspectRatio="number"==typeof o.aspectRatio?o.aspectRatio:this.originalSize.width/this.originalSize.height||1,i=e(".ui-resizable-"+this.axis).css("cursor"),e("body").css("cursor","auto"===i?this.axis+"-resize":i),this._addClass("ui-resizable-resizing"),this._propagate("start",t),!0},_mouseDrag:function(t){var n,r,i=this.originalMousePosition,o=this.axis,s=t.pageX-i.left||0,a=t.pageY-i.top||0,u=this._change[o];return this._updatePrevProperties(),!!u&&(n=u.apply(this,[t,s,a]),this._updateVirtualBoundaries(t.shiftKey),(this._aspectRatio||t.shiftKey)&&(n=this._updateRatio(n,t)),n=this._respectSize(n,t),this._updateCache(n),this._propagate("resize",t),r=this._applyChanges(),!this._helper&&this._proportionallyResizeElements.length&&this._proportionallyResize(),e.isEmptyObject(r)||(this._updatePrevProperties(),this._trigger("resize",t,this.ui()),this._applyChanges()),!1)},_mouseStop:function(t){this.resizing=!1;var n,r,i,o,s,a,u,l=this.options;return this._helper&&(i=(r=(n=this._proportionallyResizeElements).length&&/textarea/i.test(n[0].nodeName))&&this._hasScroll(n[0],"left")?0:this.sizeDiff.height,o=r?0:this.sizeDiff.width,s={width:this.helper.width()-o,height:this.helper.height()-i},a=parseFloat(this.element.css("left"))+(this.position.left-this.originalPosition.left)||null,u=parseFloat(this.element.css("top"))+(this.position.top-this.originalPosition.top)||null,l.animate||this.element.css(e.extend(s,{top:u,left:a})),this.helper.height(this.size.height),this.helper.width(this.size.width),this._helper&&!l.animate&&this._proportionallyResize()),e("body").css("cursor","auto"),this._removeClass("ui-resizable-resizing"),this._propagate("stop",t),this._helper&&this.helper.remove(),!1},_updatePrevProperties:function(){this.prevPosition={top:this.position.top,left:this.position.left},this.prevSize={width:this.size.width,height:this.size.height}},_applyChanges:function(){var e={};return this.position.top!==this.prevPosition.top&&(e.top=this.position.top+"px"),this.position.left!==this.prevPosition.left&&(e.left=this.position.left+"px"),this.size.width!==this.prevSize.width&&(e.width=this.size.width+"px"),this.size.height!==this.prevSize.height&&(e.height=this.size.height+"px"),this.helper.css(e),e},_updateVirtualBoundaries:function(e){var t,n,r,i,o,s=this.options;o={minWidth:this._isNumber(s.minWidth)?s.minWidth:0,maxWidth:this._isNumber(s.maxWidth)?s.maxWidth:1/0,minHeight:this._isNumber(s.minHeight)?s.minHeight:0,maxHeight:this._isNumber(s.maxHeight)?s.maxHeight:1/0},(this._aspectRatio||e)&&(t=o.minHeight*this.aspectRatio,r=o.minWidth/this.aspectRatio,n=o.maxHeight*this.aspectRatio,i=o.maxWidth/this.aspectRatio,t>o.minWidth&&(o.minWidth=t),r>o.minHeight&&(o.minHeight=r),n<o.maxWidth&&(o.maxWidth=n),i<o.maxHeight&&(o.maxHeight=i)),this._vBoundaries=o},_updateCache:function(e){this.offset=this.helper.offset(),this._isNumber(e.left)&&(this.position.left=e.left),this._isNumber(e.top)&&(this.position.top=e.top),this._isNumber(e.height)&&(this.size.height=e.height),this._isNumber(e.width)&&(this.size.width=e.width)},_updateRatio:function(e){var t=this.position,n=this.size,r=this.axis;return this._isNumber(e.height)?e.width=e.height*this.aspectRatio:this._isNumber(e.width)&&(e.height=e.width/this.aspectRatio),"sw"===r&&(e.left=t.left+(n.width-e.width),e.top=null),"nw"===r&&(e.top=t.top+(n.height-e.height),e.left=t.left+(n.width-e.width)),e},_respectSize:function(e){var t=this._vBoundaries,n=this.axis,r=this._isNumber(e.width)&&t.maxWidth&&t.maxWidth<e.width,i=this._isNumber(e.height)&&t.maxHeight&&t.maxHeight<e.height,o=this._isNumber(e.width)&&t.minWidth&&t.minWidth>e.width,s=this._isNumber(e.height)&&t.minHeight&&t.minHeight>e.height,a=this.originalPosition.left+this.originalSize.width,u=this.originalPosition.top+this.originalSize.height,l=/sw|nw|w/.test(n),c=/nw|ne|n/.test(n);return o&&(e.width=t.minWidth),s&&(e.height=t.minHeight),r&&(e.width=t.maxWidth),i&&(e.height=t.maxHeight),o&&l&&(e.left=a-t.minWidth),r&&l&&(e.left=a-t.maxWidth),s&&c&&(e.top=u-t.minHeight),i&&c&&(e.top=u-t.maxHeight),e.width||e.height||e.left||!e.top?e.width||e.height||e.top||!e.left||(e.left=null):e.top=null,e},_getPaddingPlusBorderDimensions:function(e){for(var t=0,n=[],r=[e.css("borderTopWidth"),e.css("borderRightWidth"),e.css("borderBottomWidth"),e.css("borderLeftWidth")],i=[e.css("paddingTop"),e.css("paddingRight"),e.css("paddingBottom"),e.css("paddingLeft")];t<4;t++)n[t]=parseFloat(r[t])||0,n[t]+=parseFloat(i[t])||0;return{height:n[0]+n[2],width:n[1]+n[3]}},_proportionallyResize:function(){if(this._proportionallyResizeElements.length)for(var e,t=0,n=this.helper||this.element;t<this._proportionallyResizeElements.length;t++)e=this._proportionallyResizeElements[t],this.outerDimensions||(this.outerDimensions=this._getPaddingPlusBorderDimensions(e)),e.css({height:n.height()-this.outerDimensions.height||0,width:n.width()-this.outerDimensions.width||0})},_renderProxy:function(){var t=this.element,n=this.options;this.elementOffset=t.offset(),this._helper?(this.helper=this.helper||e("<div style='overflow:hidden;'></div>"),this._addClass(this.helper,this._helper),this.helper.css({width:this.element.outerWidth(),height:this.element.outerHeight(),position:"absolute",left:this.elementOffset.left+"px",top:this.elementOffset.top+"px",zIndex:++n.zIndex}),this.helper.appendTo("body").disableSelection()):this.helper=this.element},_change:{e:function(e,t){return{width:this.originalSize.width+t}},w:function(e,t){var n=this.originalSize;return{left:this.originalPosition.left+t,width:n.width-t}},n:function(e,t,n){var r=this.originalSize;return{top:this.originalPosition.top+n,height:r.height-n}},s:function(e,t,n){return{height:this.originalSize.height+n}},se:function(t,n,r){return e.extend(this._change.s.apply(this,arguments),this._change.e.apply(this,[t,n,r]))},sw:function(t,n,r){return e.extend(this._change.s.apply(this,arguments),this._change.w.apply(this,[t,n,r]))},ne:function(t,n,r){return e.extend(this._change.n.apply(this,arguments),this._change.e.apply(this,[t,n,r]))},nw:function(t,n,r){return e.extend(this._change.n.apply(this,arguments),this._change.w.apply(this,[t,n,r]))}},_propagate:function(t,n){e.ui.plugin.call(this,t,[n,this.ui()]),"resize"!==t&&this._trigger(t,n,this.ui())},plugins:{},ui:function(){return{originalElement:this.originalElement,element:this.element,helper:this.helper,position:this.position,size:this.size,originalSize:this.originalSize,originalPosition:this.originalPosition}}}),e.ui.plugin.add("resizable","animate",{stop:function(t){var n=e(this).resizable("instance"),r=n.options,i=n._proportionallyResizeElements,o=i.length&&/textarea/i.test(i[0].nodeName),s=o&&n._hasScroll(i[0],"left")?0:n.sizeDiff.height,a=o?0:n.sizeDiff.width,u={width:n.size.width-a,height:n.size.height-s},l=parseFloat(n.element.css("left"))+(n.position.left-n.originalPosition.left)||null,c=parseFloat(n.element.css("top"))+(n.position.top-n.originalPosition.top)||null;n.element.animate(e.extend(u,c&&l?{top:c,left:l}:{}),{duration:r.animateDuration,easing:r.animateEasing,step:function(){var r={width:parseFloat(n.element.css("width")),height:parseFloat(n.element.css("height")),top:parseFloat(n.element.css("top")),left:parseFloat(n.element.css("left"))};i&&i.length&&e(i[0]).css({width:r.width,height:r.height}),n._updateCache(r),n._propagate("resize",t)}})}}),e.ui.plugin.add("resizable","containment",{start:function(){var t,n,r,i,o,s,a,u=e(this).resizable("instance"),l=u.options,c=u.element,f=l.containment,h=f instanceof e?f.get(0):/parent/.test(f)?c.parent().get(0):f;h&&(u.containerElement=e(h),/document/.test(f)||f===document?(u.containerOffset={left:0,top:0},u.containerPosition={left:0,top:0},u.parentData={element:e(document),left:0,top:0,width:e(document).width(),height:e(document).height()||document.body.parentNode.scrollHeight}):(t=e(h),n=[],e(["Top","Right","Left","Bottom"]).each((function(e,r){n[e]=u._num(t.css("padding"+r))})),u.containerOffset=t.offset(),u.containerPosition=t.position(),u.containerSize={height:t.innerHeight()-n[3],width:t.innerWidth()-n[1]},r=u.containerOffset,i=u.containerSize.height,o=u.containerSize.width,s=u._hasScroll(h,"left")?h.scrollWidth:o,a=u._hasScroll(h)?h.scrollHeight:i,u.parentData={element:h,left:r.left,top:r.top,width:s,height:a}))},resize:function(t){var n,r,i,o,s=e(this).resizable("instance"),a=s.options,u=s.containerOffset,l=s.position,c=s._aspectRatio||t.shiftKey,f={top:0,left:0},h=s.containerElement,p=!0;h[0]!==document&&/static/.test(h.css("position"))&&(f=u),l.left<(s._helper?u.left:0)&&(s.size.width=s.size.width+(s._helper?s.position.left-u.left:s.position.left-f.left),c&&(s.size.height=s.size.width/s.aspectRatio,p=!1),s.position.left=a.helper?u.left:0),l.top<(s._helper?u.top:0)&&(s.size.height=s.size.height+(s._helper?s.position.top-u.top:s.position.top),c&&(s.size.width=s.size.height*s.aspectRatio,p=!1),s.position.top=s._helper?u.top:0),i=s.containerElement.get(0)===s.element.parent().get(0),o=/relative|absolute/.test(s.containerElement.css("position")),i&&o?(s.offset.left=s.parentData.left+s.position.left,s.offset.top=s.parentData.top+s.position.top):(s.offset.left=s.element.offset().left,s.offset.top=s.element.offset().top),n=Math.abs(s.sizeDiff.width+(s._helper?s.offset.left-f.left:s.offset.left-u.left)),r=Math.abs(s.sizeDiff.height+(s._helper?s.offset.top-f.top:s.offset.top-u.top)),n+s.size.width>=s.parentData.width&&(s.size.width=s.parentData.width-n,c&&(s.size.height=s.size.width/s.aspectRatio,p=!1)),r+s.size.height>=s.parentData.height&&(s.size.height=s.parentData.height-r,c&&(s.size.width=s.size.height*s.aspectRatio,p=!1)),p||(s.position.left=s.prevPosition.left,s.position.top=s.prevPosition.top,s.size.width=s.prevSize.width,s.size.height=s.prevSize.height)},stop:function(){var t=e(this).resizable("instance"),n=t.options,r=t.containerOffset,i=t.containerPosition,o=t.containerElement,s=e(t.helper),a=s.offset(),u=s.outerWidth()-t.sizeDiff.width,l=s.outerHeight()-t.sizeDiff.height;t._helper&&!n.animate&&/relative/.test(o.css("position"))&&e(this).css({left:a.left-i.left-r.left,width:u,height:l}),t._helper&&!n.animate&&/static/.test(o.css("position"))&&e(this).css({left:a.left-i.left-r.left,width:u,height:l})}}),e.ui.plugin.add("resizable","alsoResize",{start:function(){var t=e(this).resizable("instance").options;e(t.alsoResize).each((function(){var t=e(this);t.data("ui-resizable-alsoresize",{width:parseFloat(t.width()),height:parseFloat(t.height()),left:parseFloat(t.css("left")),top:parseFloat(t.css("top"))})}))},resize:function(t,n){var r=e(this).resizable("instance"),i=r.options,o=r.originalSize,s=r.originalPosition,a={height:r.size.height-o.height||0,width:r.size.width-o.width||0,top:r.position.top-s.top||0,left:r.position.left-s.left||0};e(i.alsoResize).each((function(){var t=e(this),r=e(this).data("ui-resizable-alsoresize"),i={},o=t.parents(n.originalElement[0]).length?["width","height"]:["width","height","top","left"];e.each(o,(function(e,t){var n=(r[t]||0)+(a[t]||0);n&&n>=0&&(i[t]=n||null)})),t.css(i)}))},stop:function(){e(this).removeData("ui-resizable-alsoresize")}}),e.ui.plugin.add("resizable","ghost",{start:function(){var t=e(this).resizable("instance"),n=t.size;t.ghost=t.originalElement.clone(),t.ghost.css({opacity:.25,display:"block",position:"relative",height:n.height,width:n.width,margin:0,left:0,top:0}),t._addClass(t.ghost,"ui-resizable-ghost"),!1!==e.uiBackCompat&&"string"==typeof t.options.ghost&&t.ghost.addClass(this.options.ghost),t.ghost.appendTo(t.helper)},resize:function(){var t=e(this).resizable("instance");t.ghost&&t.ghost.css({position:"relative",height:t.size.height,width:t.size.width})},stop:function(){var t=e(this).resizable("instance");t.ghost&&t.helper&&t.helper.get(0).removeChild(t.ghost.get(0))}}),e.ui.plugin.add("resizable","grid",{resize:function(){var t,n=e(this).resizable("instance"),r=n.options,i=n.size,o=n.originalSize,s=n.originalPosition,a=n.axis,u="number"==typeof r.grid?[r.grid,r.grid]:r.grid,l=u[0]||1,c=u[1]||1,f=Math.round((i.width-o.width)/l)*l,h=Math.round((i.height-o.height)/c)*c,p=o.width+f,d=o.height+h,g=r.maxWidth&&r.maxWidth<p,v=r.maxHeight&&r.maxHeight<d,m=r.minWidth&&r.minWidth>p,y=r.minHeight&&r.minHeight>d;r.grid=u,m&&(p+=l),y&&(d+=c),g&&(p-=l),v&&(d-=c),/^(se|s|e)$/.test(a)?(n.size.width=p,n.size.height=d):/^(ne)$/.test(a)?(n.size.width=p,n.size.height=d,n.position.top=s.top-h):/^(sw)$/.test(a)?(n.size.width=p,n.size.height=d,n.position.left=s.left-f):((d-c<=0||p-l<=0)&&(t=n._getPaddingPlusBorderDimensions(this)),d-c>0?(n.size.height=d,n.position.top=s.top-h):(d=c-t.height,n.size.height=d,n.position.top=s.top+o.height-d),p-l>0?(n.size.width=p,n.position.left=s.left-f):(p=l-t.width,n.size.width=p,n.position.left=s.left+o.width-p))}}),e.ui.resizable})?r.apply(t,i):r)||(e.exports=o)},function(e,t,n){var r,i,o;
/*!
 * jQuery UI Mouse 1.12.1
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */i=[n(0),n(19),n(1),n(3)],void 0===(o="function"==typeof(r=function(e){var t=!1;return e(document).on("mouseup",(function(){t=!1})),e.widget("ui.mouse",{version:"1.12.1",options:{cancel:"input, textarea, button, select, option",distance:1,delay:0},_mouseInit:function(){var t=this;this.element.on("mousedown."+this.widgetName,(function(e){return t._mouseDown(e)})).on("click."+this.widgetName,(function(n){if(!0===e.data(n.target,t.widgetName+".preventClickEvent"))return e.removeData(n.target,t.widgetName+".preventClickEvent"),n.stopImmediatePropagation(),!1})),this.started=!1},_mouseDestroy:function(){this.element.off("."+this.widgetName),this._mouseMoveDelegate&&this.document.off("mousemove."+this.widgetName,this._mouseMoveDelegate).off("mouseup."+this.widgetName,this._mouseUpDelegate)},_mouseDown:function(n){if(!t){this._mouseMoved=!1,this._mouseStarted&&this._mouseUp(n),this._mouseDownEvent=n;var r=this,i=1===n.which,o=!("string"!=typeof this.options.cancel||!n.target.nodeName)&&e(n.target).closest(this.options.cancel).length;return!(i&&!o&&this._mouseCapture(n)&&(this.mouseDelayMet=!this.options.delay,this.mouseDelayMet||(this._mouseDelayTimer=setTimeout((function(){r.mouseDelayMet=!0}),this.options.delay)),this._mouseDistanceMet(n)&&this._mouseDelayMet(n)&&(this._mouseStarted=!1!==this._mouseStart(n),!this._mouseStarted)?(n.preventDefault(),0):(!0===e.data(n.target,this.widgetName+".preventClickEvent")&&e.removeData(n.target,this.widgetName+".preventClickEvent"),this._mouseMoveDelegate=function(e){return r._mouseMove(e)},this._mouseUpDelegate=function(e){return r._mouseUp(e)},this.document.on("mousemove."+this.widgetName,this._mouseMoveDelegate).on("mouseup."+this.widgetName,this._mouseUpDelegate),n.preventDefault(),t=!0,0)))}},_mouseMove:function(t){if(this._mouseMoved){if(e.ui.ie&&(!document.documentMode||document.documentMode<9)&&!t.button)return this._mouseUp(t);if(!t.which)if(t.originalEvent.altKey||t.originalEvent.ctrlKey||t.originalEvent.metaKey||t.originalEvent.shiftKey)this.ignoreMissingWhich=!0;else if(!this.ignoreMissingWhich)return this._mouseUp(t)}return(t.which||t.button)&&(this._mouseMoved=!0),this._mouseStarted?(this._mouseDrag(t),t.preventDefault()):(this._mouseDistanceMet(t)&&this._mouseDelayMet(t)&&(this._mouseStarted=!1!==this._mouseStart(this._mouseDownEvent,t),this._mouseStarted?this._mouseDrag(t):this._mouseUp(t)),!this._mouseStarted)},_mouseUp:function(n){this.document.off("mousemove."+this.widgetName,this._mouseMoveDelegate).off("mouseup."+this.widgetName,this._mouseUpDelegate),this._mouseStarted&&(this._mouseStarted=!1,n.target===this._mouseDownEvent.target&&e.data(n.target,this.widgetName+".preventClickEvent",!0),this._mouseStop(n)),this._mouseDelayTimer&&(clearTimeout(this._mouseDelayTimer),delete this._mouseDelayTimer),this.ignoreMissingWhich=!1,t=!1,n.preventDefault()},_mouseDistanceMet:function(e){return Math.max(Math.abs(this._mouseDownEvent.pageX-e.pageX),Math.abs(this._mouseDownEvent.pageY-e.pageY))>=this.options.distance},_mouseDelayMet:function(){return this.mouseDelayMet},_mouseStart:function(){},_mouseDrag:function(){},_mouseStop:function(){},_mouseCapture:function(){return!0}})})?r.apply(t,i):r)||(e.exports=o)},function(e,t,n){var r,i,o;i=[n(0),n(1)],void 0===(o="function"==typeof(r=function(e){return e.ui.ie=!!/msie [\w.]+/.exec(navigator.userAgent.toLowerCase())})?r.apply(t,i):r)||(e.exports=o)},function(e,t,n){var r,i,o;
/*!
 * jQuery UI Disable Selection 1.12.1
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 */i=[n(0),n(1)],void 0===(o="function"==typeof(r=function(e){return e.fn.extend({disableSelection:(t="onselectstart"in document.createElement("div")?"selectstart":"mousedown",function(){return this.on(t+".ui-disableSelection",(function(e){e.preventDefault()}))}),enableSelection:function(){return this.off(".ui-disableSelection")}});var t})?r.apply(t,i):r)||(e.exports=o)},function(e,t,n){var r,i,o;i=[n(0),n(1)],void 0===(o="function"==typeof(r=function(e){return e.ui.plugin={add:function(t,n,r){var i,o=e.ui[t].prototype;for(i in r)o.plugins[i]=o.plugins[i]||[],o.plugins[i].push([n,r[i]])},call:function(e,t,n,r){var i,o=e.plugins[t];if(o&&(r||e.element[0].parentNode&&11!==e.element[0].parentNode.nodeType))for(i=0;i<o.length;i++)e.options[o[i][0]]&&o[i][1].apply(e.element,n)}}})?r.apply(t,i):r)||(e.exports=o)},function(e){e.exports=JSON.parse('{"name":"vispy","version":"0.3.0","description":"A Custom Jupyter Widget Library using WebGL for the VisPy Python Library","author":"Vispy Development Team","license":"BSD-3-Clause","main":"lib/index.js","bugs":{"url":"https://github.com/vispy/vispy.js/issues"},"repository":{"type":"git","url":"https://github.com/vispy/vispy.git"},"homepage":"https://github.com/vispy/vispy","keywords":["jupyter","widgets","ipython","ipywidgets","jupyterlab-extension","vispy","webgl","opengl"],"files":["lib/**/*.js","dist/*.js"],"scripts":{"clean":"rimraf dist/","prepare":"webpack --mode=production","test":"echo \\"Error: no test specified\\" && exit 1","watch":"npm-run-all -p watch:*","watch:nbextension":"webpack --watch --mode=development"},"devDependencies":{"npm-run-all":"^4.1.5","npm-watch":"^0.6.0","raw-loader":"^2.0.0","rimraf":"^2.7.1","source-map-loader":"^0.2.4","webpack":"^4.43.0","webpack-cli":"^3.3.12"},"dependencies":{"@jupyter-widgets/base":"^1 || ^2 || ^3","jquery":"^3.5.1","jquery-ui":"^1.12.1","lodash":"^4.17.19","screenfull":"^4.2.1"},"jupyterlab":{"extension":"lib/labplugin"}}')}])}));
//# sourceMappingURL=index.js.map